@@ -729,6 +729,31 @@ static int test_caps_regression(char *node, char *service, uint64_t flags,
 	return 0;
 }
 
+/* Check if at least one of FI_LOCAL_COMM and FI_REMOTE_COMM is set */
+static int test_comm_caps(char *node, char *service, uint64_t flags,
+		struct fi_info *hints, struct fi_info **info)
+{
+	uint64_t test_caps[2] = { FI_LOCAL_COMM, FI_REMOTE_COMM };
+	int ret, i;
+
+	for (i = 0; i < 2; i++) {
+		hints->caps &= ~(FI_LOCAL_COMM | FI_REMOTE_COMM);
+		hints->domain_attr->caps &= ~(FI_LOCAL_COMM | FI_REMOTE_COMM);
+		hints->caps |= test_caps[i];
+		hints->domain_attr->caps |= test_caps[i];
+		ret = fi_getinfo(FT_FIVERSION, node, service, flags, hints, info);
+		if (ret == 0) {
+			if (((*info)->caps & test_caps[i]) == test_caps[i] &&
+			    ((*info)->domain_attr->caps & test_caps[i]) == test_caps[i])
+				return 0;
+
+			fi_freeinfo(*info);
+		}
+	}
+
+	FT_DEBUG("Neither FI_LOCAL_COMM nor FI_REMOTE_COMM is set\n");
+	return -FI_EIO;
+}
 
 /*
  * getinfo test
@@ -936,6 +961,8 @@ getinfo_test(caps, 3, "Test domain capabilities", NULL, NULL, 0,
 	     hints, NULL, validate_domain_caps, NULL, 0)
 getinfo_test(caps, 4, "Test for capability bit regression",
 	     NULL, NULL, 0, hints, NULL, test_caps_regression, NULL, 0)
+getinfo_test(caps, 5, "Test if either FI_LOCAL_COMM or FI_REMOTE_COMM is set",
+	     NULL, NULL, 0, hints, NULL, test_comm_caps, NULL, 0)
 
 
 static void usage(char *name)
@@ -1020,6 +1047,7 @@ int main(int argc, char **argv)
 		TEST_ENTRY_GETINFO(caps2),
 		TEST_ENTRY_GETINFO(caps3),
 		TEST_ENTRY_GETINFO(caps4),
+		TEST_ENTRY_GETINFO(caps5),
 		{ NULL, "" }
 	};
 
