@@ -64,42 +64,102 @@
  * Functions to manipulate the expected queue in mq_ep.
  */
 
+#ifdef LEARN_HASH_SELECTOR
+// hash the req and then append to specified hash table in given q
+PSMI_ALWAYS_INLINE(
+void mq_qq_append_which_hash(psm2_mq_t mq,
+				struct mqq q[NUM_HASH_CONFIGS][NUM_HASH_BUCKETS],
+				int table, psm2_mq_req_t req))
+{
+	unsigned hashval = hash_src_tag_sel(&mq->table_sel[table],
+										req->req_data.peer, &req->req_data.tag)
+							% NUM_HASH_BUCKETS;
+	mq_qq_append_which(q, table, hashval, req);
+}
+
+// find the table with matching table_sel. 
+// If none found and have room for a new table, add a new empty table.
+// if none found and didn't add a table, return NUM_HASH_CONFIGS and caller
+// will need to use the simple linear list
+PSMI_ALWAYS_INLINE(
+int mq_qq_find_table(psm2_mq_t mq, psm2_epaddr_t src, psm2_mq_tag_t *tagsel))
+{
+	int t;
+	uint8_t srcsel = src != PSM2_MQ_ANY_ADDR;
+
+	// chose to have a loop instead of unrolling loop due to
+	// possible better CPU instruction cache hit rate
+	// search in the order the tables got created, 1st created may be used more
+	for (t = NUM_HASH_CONFIGS-1; t >= mq->min_table; t--) {
+		if (table_sel_cmp(&mq->table_sel[t], srcsel, tagsel))
+			return t;
+	}
+	if (mq->min_table > 0) {
+		// add a table
+		psmi_assert(mq->min_table <= NUM_HASH_CONFIGS);
+		_HFI_VDBG("add table_sel, srcsel=%u tagsel=%08x.%08x.%08x\n",
+					srcsel, tagsel->tag[0], tagsel->tag[1], tagsel->tag[2]);
+		t = --(mq->min_table);
+		UPDATE_SUBQUEUE_COUNT(mq);
+		build_table_sel(&mq->table_sel[t], srcsel, tagsel);
+		return t;
+	}
+	// out of tables, must use simple linear list
+	if (! mq->search_linear_expected) {
+		mq->search_linear_expected = 1;
+		UPDATE_SUBQUEUE_COUNT(mq);
+	}
+	return NUM_HASH_CONFIGS;
+}
+#endif /* LEARN_HASH_SELECTOR */
+
 /*
  * Once the linked lists cross the size limit, this function will enable tag
  * hashing and disable the non-hashing fastpath. We need to go back and insert
  * reqs into the hash tables where the hashing searches will look for them.
+ *
+ * When called, either the expected or unexpected queue has crossed the
+ * mq->hash_thresh.  However it's possible the other queue is empty.
  */
+#ifdef LEARN_HASH_SELECTOR
+/* If this is the first time we are transitioning out of nohash mode, and
+ * the expected queue is currently empty, we will only adjust flags,
+ * but won't actually have any table_sel values to use to build hashes until a
+ * subsequent psm3_mq_* call provides a src/tagsel. At which point, that call
+ * will then trigger the addition of a corresponding table_sel.
+ */
+#endif
 void
 psm3_mq_fastpath_disable(psm2_mq_t mq)
 {
-	psm2_mq_req_t *curp, cur;
-	struct mqq *qp;
+	psm2_mq_req_t next, cur;
+	int t;
+#ifndef LEARN_HASH_SELECTOR
 	unsigned hashvals[NUM_HASH_CONFIGS];
-	int t = PSM2_ANYTAG_ANYSRC;
+#endif
+
+	_HFI_VDBG("enable tag hashing: thresh %u unexp %u exp %u\n",
+			mq->hash_thresh, mq->unexpected_list_len, mq->expected_list_len);
 
 	mq->nohash_fastpath = 0;
-	/* Everything in the unexpected_q needs to be duplicated into
-	   each of the (three) unexpected hash tables. */
-	qp = &mq->unexpected_q;
-	for (curp = &qp->first; (cur = *curp) != NULL; curp = &cur->next[t]) {
-		mq->unexpected_hash_len++;
-		hashvals[PSM2_TAG_SRC] =
-			hash_64(cur->req_data.tag.tag64) % NUM_HASH_BUCKETS;
-		hashvals[PSM2_TAG_ANYSRC] =
-			hash_32(cur->req_data.tag.tag[0]) % NUM_HASH_BUCKETS;
-		hashvals[PSM2_ANYTAG_SRC] =
-			hash_32(cur->req_data.tag.tag[1]) % NUM_HASH_BUCKETS;
-		for (t = PSM2_TAG_SRC; t < PSM2_ANYTAG_ANYSRC; t++)
-			mq_qq_append_which(mq->unexpected_htab,
-					   t, hashvals[t], cur);
-	}
+#ifndef LEARN_HASH_SELECTOR
+	mq->min_table = PSM2_TAG_SRC;
+	psmi_assert(PSM2_ANYTAG_ANYSRC == NUM_HASH_CONFIGS);
+#endif
 
 	/* Everything in the expected_q needs to be moved into the
-	   (single) correct expected hash table. */
-	qp = &mq->expected_q;
-	for (curp = &qp->first; (cur = *curp) != NULL; /*curp = &cur->next*/) {
+	   (single) correct expected hash table.
+	   This will also establish the table_sel's we will use
+	   for unexepected queue below */
+	for (cur = mq->expected_q.first; cur != NULL; cur = next) {
 		/* must read next ptr before remove */
-		curp = &cur->next[PSM2_ANYTAG_ANYSRC];
+		next = cur->next[NUM_HASH_CONFIGS];
+#ifdef LEARN_HASH_SELECTOR
+		t = mq_qq_find_table(mq, cur->req_data.peer, &cur->req_data.tagsel);
+		if_pf (t == NUM_HASH_CONFIGS)
+			continue; /* req must stay in simple linear list */
+		mq_qq_append_which_hash(mq, mq->expected_htab, t, cur);
+#else
 		if ((cur->req_data.tagsel.tag[0] == 0xFFFFFFFF) &&
 		    (cur->req_data.tagsel.tag[1] == 0xFFFFFFFF)) {
 			/* hash tag0 and tag1 */
@@ -119,11 +179,45 @@ psm3_mq_fastpath_disable(psm2_mq_t mq)
 					   t, hashvals[t], cur);
 		} else
 			continue; /* else, req must stay in ANY ANY */
-
+#endif
 		mq->expected_list_len--;
 		mq->expected_hash_len++;
-		mq_qq_remove_which(cur, PSM2_ANYTAG_ANYSRC);
+		mq_qq_remove_which(cur, NUM_HASH_CONFIGS);
 	}
+
+	/* Everything in the unexpected_q needs to be duplicated into
+	   each of the unexpected hash tables. */
+#ifdef LEARN_HASH_SELECTOR
+	if (mq->min_table < NUM_HASH_CONFIGS) {
+		for (cur = mq->unexpected_q.first; cur != NULL;
+			 cur = cur->next[NUM_HASH_CONFIGS]) {
+			mq->unexpected_hash_len++;
+			for (t = mq->min_table; t < NUM_HASH_CONFIGS; t++) {
+				mq_qq_append_which_hash(mq, mq->unexpected_htab, t, cur);
+			}
+		}
+		psmi_assert(mq->unexpected_hash_len == mq->unexpected_list_len);
+	}
+#else
+	for (cur = mq->unexpected_q.first; cur != NULL;
+		 cur = cur->next[NUM_HASH_CONFIGS]) {
+		mq->unexpected_hash_len++;
+		hashvals[PSM2_TAG_SRC] =
+			hash_64(cur->req_data.tag.tag64) % NUM_HASH_BUCKETS;
+		hashvals[PSM2_TAG_ANYSRC] =
+			hash_32(cur->req_data.tag.tag[0]) % NUM_HASH_BUCKETS;
+		hashvals[PSM2_ANYTAG_SRC] =
+			hash_32(cur->req_data.tag.tag[1]) % NUM_HASH_BUCKETS;
+		for (t = PSM2_TAG_SRC; t < PSM2_ANYTAG_ANYSRC; t++)
+			mq_qq_append_which(mq->unexpected_htab,
+					   t, hashvals[t], cur);
+	}
+	psmi_assert(mq->unexpected_hash_len == mq->unexpected_list_len);
+#endif
+
+	UPDATE_SUBQUEUE_COUNT(mq);
+	UPDATE_EXP_HASH_COUNT(mq);
+	UPDATE_UNEXP_HASH_COUNT(mq);
 }
 
 /* easy threshold to re-enable: if |hash| == 0 && |list| < X
@@ -136,11 +230,81 @@ void psm3_mq_fastpath_try_reenable(psm2_mq_t mq)
 	      mq->unexpected_hash_len == 0 &&
 	      mq->expected_hash_len == 0 &&
 	      mq->unexpected_list_len == 0 &&
-	      mq->expected_list_len == 0){
+	      mq->expected_list_len == 0 &&
+	      mq->hash_thresh > 0) {
+		// all unexpected on list and hashes, so total unexp is list (== hash)
+		// all expected on list or 1 hash, so total exp is sum
+		_HFI_VDBG("re-disable tag hashing: thresh %u unexp %u exp %u\n",
+				mq->hash_thresh, mq->unexpected_list_len,
+				mq->expected_list_len + mq->expected_hash_len);
 		mq->nohash_fastpath = 1;
+#if ! defined(LEARN_HASH_SELECTOR) || ! defined(RETAIN_PAST_TABLE_SEL)
+		mq->min_table = NUM_HASH_CONFIGS;
+#ifdef LEARN_HASH_SELECTOR
+		mq->search_linear_expected = 0;
+#endif
+#endif
 	}
 }
 
+#ifdef LEARN_HASH_SELECTOR
+// populate a new unexpected hash table and while walking unexpected list
+// also search for a matching req
+// This assumes the real cost is memory references to walk the list
+// as opposed to the cost of tag_cmp.
+// If tag_cmp is the higher cost, it would be faster to search the appropriate
+// unexpected hash table when done building it because the hash would
+// reduce number of entries to tag_cmp.
+//
+// It would be possible to defer the cost of adding and populating the new
+// unexpected hash table until we know the unexpected search will fail.
+// However doing it now should allow future unexpected queue searches to
+// be more efficient, which will be a win, especially if the searches
+// need to go deeper into the unexpected queue.
+PSMI_ALWAYS_INLINE(
+psm2_mq_req_t populate_and_search(psm2_mq_t mq, int table, int srcsel,
+								psm2_epaddr_t src,
+								psm2_mq_tag_t *tagsel, psm2_mq_tag_t *tag))
+{
+#ifdef PSM_DEBUG
+	unsigned cnt = 0;
+#endif
+	psm2_mq_req_t match = NULL;
+	psm2_mq_req_t cur;
+	unsigned k = 0;
+
+	// walk the linear unexpected list to populate hash table and search
+	for (cur = mq->unexpected_q.first; cur != NULL;
+		 cur = cur->next[NUM_HASH_CONFIGS], k++) {
+#ifdef PSM_DEBUG
+		cnt++;
+#endif
+		mq_qq_append_which_hash(mq, mq->unexpected_htab, table, cur);
+		if (tag_cmp_req(cur, srcsel, src, tagsel, tag)) {
+			match = cur;
+			cur = cur->next[NUM_HASH_CONFIGS];
+			break;
+		}
+	}
+	// continue populating, but done searching
+	for (; cur != NULL; cur = cur->next[NUM_HASH_CONFIGS]) {
+#ifdef PSM_DEBUG
+		cnt++;
+#endif
+		mq_qq_append_which_hash(mq, mq->unexpected_htab, table, cur);
+	}
+#ifdef PSM_DEBUG
+	if (table == NUM_HASH_CONFIGS-1) {	// first hash table learned
+		psmi_assert(cnt == mq->unexpected_list_len);
+		psmi_assert(! mq->unexpected_hash_len);
+	}
+#endif
+	mq->unexpected_hash_len = mq->unexpected_list_len;
+	UPDATE_UNEXP_SEARCH_LEN(mq, k);
+	return match;
+}
+#endif /* LEARN_HASH_SELECTOR */
+
 /*
  * ! @brief PSM exposed version to allow PTLs to match
  */
@@ -154,94 +318,131 @@ void psm3_mq_fastpath_try_reenable(psm2_mq_t mq)
  * @param[in] remove Non-zero to remove the req from the queue
  *
  * @returns NULL if no match or an mq request if there is a match
+ * When NULL is returned, *table will indicate expected hash table to add to
  */
 static
 psm2_mq_req_t
 mq_req_match_with_tagsel(psm2_mq_t mq, psm2_epaddr_t src,
-			 psm2_mq_tag_t *tag, psm2_mq_tag_t *tagsel, int remove)
+			 psm2_mq_tag_t *tag, psm2_mq_tag_t *tagsel, int remove, int *table)
 {
-	psm2_mq_req_t *curp;
 	psm2_mq_req_t cur;
 	unsigned hashval;
-	int i, j = 0;
-	struct mqq *qp;
+	int i, j;
+	struct mqq *q;
+	unsigned k = 0;
+	int srcsel = src != PSM2_MQ_ANY_ADDR;
+#ifndef LEARN_HASH_SELECTOR
+	j = 0;
+#endif
 
+	mq->stats.num_unexp_search++;
 	if_pt (mq->nohash_fastpath) {
-		i = j = PSM2_ANYTAG_ANYSRC;
-		qp = &mq->unexpected_q;
+		i = j = NUM_HASH_CONFIGS;
+		q = &mq->unexpected_q;
+#ifdef LEARN_HASH_SELECTOR
+	} else {
+		j = mq->min_table;	// 1st valid unexpected subqueue for remove
+		psmi_assert(j <= NUM_HASH_CONFIGS && j >= 0);
+		i = mq_qq_find_table(mq, src, tagsel);
+		psmi_assert(i <= NUM_HASH_CONFIGS && i >= mq->min_table);
+		if_pt (i < NUM_HASH_CONFIGS) {
+			if_pf (i < j)	{ // added a new table
+				/* Everything in the unexpected_q needs to be
+				   duplicated into the new unexpected hash table. */
+				psmi_assert(i == mq->min_table);
+				cur = populate_and_search(mq, i, srcsel, src, tagsel, tag);
+				if (cur) {
+					j = mq->min_table;	// new min for remove
+					goto found;
+				}
+				// hashval for add to expected table
+				mq->hashvals[0] = hash_src_tag_sel(&mq->table_sel[i], src, tag)
+							% NUM_HASH_BUCKETS;
+				goto notfound;
+			}
+			// search an existing hash table
+			hashval = hash_src_tag_sel(&mq->table_sel[i], src, tag)
+							% NUM_HASH_BUCKETS;
+			mq->hashvals[0] = hashval;	// for add to expected table
+			q = &mq->unexpected_htab[i][hashval];
+		} else {
+			// couldn't add a table, search the simple linear list
+			q = &mq->unexpected_q;
+		}
+	}
+#else /* LEARN_HASH_SELECTOR */
 	} else if ((tagsel->tag[0] == 0xFFFFFFFF) &&
 		   (tagsel->tag[1] == 0xFFFFFFFF)) {
 		i = PSM2_TAG_SRC;
 		hashval = hash_64(tag->tag64) % NUM_HASH_BUCKETS;
-		qp = &mq->unexpected_htab[i][hashval];
+		mq->hashvals[0] = hashval;	// for add to expected table
+		q = &mq->unexpected_htab[i][hashval];
 	} else if (tagsel->tag[0] == 0xFFFFFFFF) {
 		i = PSM2_TAG_ANYSRC;
 		hashval = hash_32(tag->tag[0]) % NUM_HASH_BUCKETS;
-		qp = &mq->unexpected_htab[i][hashval];
+		mq->hashvals[0] = hashval;	// for add to expected table
+		q = &mq->unexpected_htab[i][hashval];
 	} else if (tagsel->tag[1] == 0xFFFFFFFF) {
 		i = PSM2_ANYTAG_SRC;
 		hashval = hash_32(tag->tag[1]) % NUM_HASH_BUCKETS;
-		qp = &mq->unexpected_htab[i][hashval];
+		mq->hashvals[0] = hashval;	// for add to expected table
+		q = &mq->unexpected_htab[i][hashval];
 	} else {
 		/* unhashable tag */
 		i = PSM2_ANYTAG_ANYSRC;
-		qp = &mq->unexpected_q;
+		q = &mq->unexpected_q;
 	}
-
-	for (curp = &qp->first; (cur = *curp) != NULL; curp = &cur->next[i]) {
+#endif /* LEARN_HASH_SELECTOR */
+	for (cur = q->first; cur != NULL; cur = cur->next[i], k++) {
 		psmi_assert(cur->req_data.peer != PSM2_MQ_ANY_ADDR);
-		if ((src == PSM2_MQ_ANY_ADDR || src == cur->req_data.peer) &&
-		    !((tag->tag[0] ^ cur->req_data.tag.tag[0]) & tagsel->tag[0]) &&
-		    !((tag->tag[1] ^ cur->req_data.tag.tag[1]) & tagsel->tag[1]) &&
-		    !((tag->tag[2] ^ cur->req_data.tag.tag[2]) & tagsel->tag[2])) {
-			/* match! */
-			if (remove) {
-				if_pt (i == PSM2_ANYTAG_ANYSRC)
-					mq->unexpected_list_len--;
-				else
-					mq->unexpected_hash_len--;
-				for (; j < NUM_MQ_SUBLISTS; j++)
-					mq_qq_remove_which(cur, j);
-				psm3_mq_fastpath_try_reenable(mq);
-			}
-			return cur;
-		}
+		if (tag_cmp_req(cur, srcsel, src, tagsel, tag))
+			goto  found;
 	}
+#ifdef LEARN_HASH_SELECTOR
+notfound:
+#endif
+	psmi_assert(i <= NUM_HASH_CONFIGS && i >= mq->min_table);
+	*table = i;
+	UPDATE_UNEXP_SEARCH_LEN(mq, k);
 	return NULL;
+
+found:
+	if (remove) {
+		psmi_assert(j <= NUM_HASH_CONFIGS && j >= mq->min_table);
+		mq->unexpected_list_len--;
+		if_pf (j < NUM_HASH_CONFIGS) {
+			mq->unexpected_hash_len--;
+			psmi_assert(mq->unexpected_hash_len == mq->unexpected_list_len);
+		}
+		for (; j < NUM_HASH_CONFIGS+1; j++)
+			mq_qq_remove_which(cur, j);
+		psm3_mq_fastpath_try_reenable(mq);
+	}
+	UPDATE_UNEXP_SEARCH_LEN(mq, k);
+	return cur;
 }
 
-static void mq_add_to_expected_hashes(psm2_mq_t mq, psm2_mq_req_t req)
+// must be immediately preceeded by mq_req_match_with_tagsel so
+// mq->hashvals[0] is primed when using an expected hashtab
+static void mq_add_to_expected_hashes(psm2_mq_t mq, int table, psm2_mq_req_t req)
 {
-	unsigned hashval;
-	int i;
-
 	req->timestamp = mq->timestamp++;
 	if_pt (mq->nohash_fastpath) {
+		psmi_assert(table == NUM_HASH_CONFIGS);
 		mq_qq_append(&mq->expected_q, req);
-		req->q[PSM2_ANYTAG_ANYSRC] = &mq->expected_q;
 		mq->expected_list_len++;
-		if_pf (mq->expected_list_len >= HASH_THRESHOLD)
+		if_pf (mq->expected_list_len > mq->hash_thresh)
 			psm3_mq_fastpath_disable(mq);
-	} else if ((req->req_data.tagsel.tag[0] == 0xFFFFFFFF) &&
-		   (req->req_data.tagsel.tag[1] == 0xFFFFFFFF)) {
-		i = PSM2_TAG_SRC;
-		hashval = hash_64(req->req_data.tag.tag64) % NUM_HASH_BUCKETS;
-		mq_qq_append_which(mq->expected_htab, i, hashval, req);
-		mq->expected_hash_len++;
-	} else if (req->req_data.tagsel.tag[0] == 0xFFFFFFFF) {
-		i = PSM2_TAG_ANYSRC;
-		hashval = hash_32(req->req_data.tag.tag[0]) % NUM_HASH_BUCKETS;
-		mq_qq_append_which(mq->expected_htab, i, hashval, req);
-		mq->expected_hash_len++;
-	} else if (req->req_data.tagsel.tag[1] == 0xFFFFFFFF) {
-		i = PSM2_ANYTAG_SRC;
-		hashval = hash_32(req->req_data.tag.tag[1]) % NUM_HASH_BUCKETS;
-		mq_qq_append_which(mq->expected_htab, i, hashval, req);
+		UPDATE_EXP_LIST_COUNT(mq);
+	} else if (table != NUM_HASH_CONFIGS) {
+		mq_qq_append_which(mq->expected_htab, table, mq->hashvals[0], req);
 		mq->expected_hash_len++;
+		UPDATE_EXP_HASH_COUNT(mq);
 	} else {
+		// out of hashes, add all others to linear list
 		mq_qq_append(&mq->expected_q, req);
-		req->q[PSM2_ANYTAG_ANYSRC] = &mq->expected_q;
 		mq->expected_list_len++;
+		UPDATE_EXP_LIST_COUNT(mq);
 	}
 }
 
@@ -258,24 +459,28 @@ int mq_req_remove_single(psm2_mq_t mq, psm2_mq_req_t req)
 	int i;
 
 	/* item should only exist in one expected queue at a time */
-	psmi_assert((!!req->q[0] + !!req->q[1] + !!req->q[2] + !!req->q[3]) == 1);
-
-	for (i = 0; i < NUM_MQ_SUBLISTS; i++)
-		if (req->q[i]) /* found */
-			break;
-	switch (i) {
-	case PSM2_ANYTAG_ANYSRC:
-		mq->expected_list_len--;
-		break;
-	case PSM2_TAG_SRC:
-	case PSM2_TAG_ANYSRC:
-	case PSM2_ANYTAG_SRC:
-		mq->expected_hash_len--;
-		break;
-	default:
-		return 0;
+#ifdef PSM_DEBUG
+	int count=0;
+	for (i=mq->min_table; i < NUM_HASH_CONFIGS+1; i++) {
+		if (req->q[i])
+			count++;
 	}
+	psmi_assert(count == 1);
+#endif
 
+	if (req->q[NUM_HASH_CONFIGS]) {
+		mq->expected_list_len--;
+		i = NUM_HASH_CONFIGS;
+	} else {
+		for (i = mq->min_table; i < NUM_HASH_CONFIGS; i++) {
+			if (req->q[i]) {
+				mq->expected_hash_len--;
+				goto found;
+			}
+		}
+		return 0;	// should not happen, verified req is POSTED
+	}
+found:
 	mq_qq_remove_which(req, i);
 	psm3_mq_fastpath_try_reenable(mq);
 	return 1;
@@ -288,9 +493,10 @@ psm3_mq_iprobe_inner(psm2_mq_t mq, psm2_epaddr_t src,
 		     psm2_mq_tag_t *tagsel, int remove_req))
 {
 	psm2_mq_req_t req;
+	int table;
 
 	PSMI_LOCK(mq->progress_lock);
-	req = mq_req_match_with_tagsel(mq, src, tag, tagsel, remove_req);
+	req = mq_req_match_with_tagsel(mq, src, tag, tagsel, remove_req, &table);
 
 	if (req != NULL) {
 		PSMI_UNLOCK(mq->progress_lock);
@@ -299,7 +505,7 @@ psm3_mq_iprobe_inner(psm2_mq_t mq, psm2_epaddr_t src,
 
 	psm3_poll_internal(mq->ep, 1, 0);
 	/* try again */
-	req = mq_req_match_with_tagsel(mq, src, tag, tagsel, remove_req);
+	req = mq_req_match_with_tagsel(mq, src, tag, tagsel, remove_req, &table);
 
 	PSMI_UNLOCK(mq->progress_lock);
 	return req;
@@ -342,10 +548,10 @@ psm3_mq_iprobe(psm2_mq_t mq, uint64_t tag, uint64_t tagsel,
 
 	rtag.tag64 = tag;
 #ifdef PSM_DEBUG
-	rtag.tag[2] = 0;
+	rtag.rem32 = 0;
 #endif
 	rtagsel.tag64 = tagsel;
-	rtagsel.tag[2] = 0;
+	rtagsel.rem32 = 0;
 
 	req = psm3_mq_iprobe_inner(mq, PSM2_MQ_ANY_ADDR, &rtag, &rtagsel, 0);
 	psmi_assert_req_not_internal(req);
@@ -402,10 +608,10 @@ psm3_mq_improbe(psm2_mq_t mq, uint64_t tag, uint64_t tagsel,
 
 	rtag.tag64 = tag;
 #ifdef PSM_DEBUG
-	rtag.tag[2] = 0;
+	rtag.rem32 = 0;
 #endif
 	rtagsel.tag64 = tagsel;
-	rtagsel.tag[2] = 0;
+	rtagsel.rem32 = 0;
 
 	req = psm3_mq_iprobe_inner(mq, PSM2_MQ_ANY_ADDR, &rtag, &rtagsel, 1);
 	if (req != NULL) {
@@ -681,7 +887,7 @@ psm3_mq_isend(psm2_mq_t mq, psm2_epaddr_t dest, uint32_t flags, uint64_t stag,
 	PSM2_LOG_MSG("entering");
 
 	tag.tag64 = stag;
-	tag.tag[2] = 0;
+	tag.rem32 = 0;
 
 	PSMI_ASSERT_INITIALIZED();
 
@@ -726,7 +932,7 @@ psm3_mq_send(psm2_mq_t mq, psm2_epaddr_t dest, uint32_t flags, uint64_t stag,
 	PSM2_LOG_MSG("entering stag: 0x%" PRIx64, stag);
 
 	tag.tag64 = stag;
-	tag.tag[2] = 0;
+	tag.rem32 = 0;
 
 	PSMI_ASSERT_INITIALIZED();
 
@@ -853,6 +1059,7 @@ psm3_mq_fp_msg(psm2_ep_t ep, psm2_mq_t mq, psm2_epaddr_t addr, psm2_mq_tag_t *ta
 		(*req)->req_data.peer = addr;
 	} else if (fp_type == PSM2_MQ_IRECV_FP) {
 		psm2_mq_req_t recv_req;
+		int table;
 
 #if defined(PSM_CUDA) || defined(PSM_ONEAPI)
 		int gpu_mem = 0;
@@ -867,7 +1074,7 @@ psm3_mq_fp_msg(psm2_ep_t ep, psm2_mq_t mq, psm2_epaddr_t addr, psm2_mq_tag_t *ta
 #endif
 
 		/* First check unexpected Queue and remove req if found */
-		recv_req = mq_req_match_with_tagsel(mq, addr, tag, tagsel, REMOVE_ENTRY);
+		recv_req = mq_req_match_with_tagsel(mq, addr, tag, tagsel, REMOVE_ENTRY, &table);
 
 		if (recv_req == NULL) {
 			/* prepost before arrival, add to expected q */
@@ -892,7 +1099,7 @@ psm3_mq_fp_msg(psm2_ep_t ep, psm2_mq_t mq, psm2_epaddr_t addr, psm2_mq_tag_t *ta
 			recv_req->user_gpu_buffer = gpu_user_buffer;
 #endif
 
-			mq_add_to_expected_hashes(mq, recv_req);
+			mq_add_to_expected_hashes(mq, table, recv_req);
 			_HFI_VDBG("buf=%p,len=%d,tag=%08x.%08x.%08x "
 				  " tagsel=%08x.%08x.%08x req=%p\n",
 				  buf, len, tag->tag[0], tag->tag[1], tag->tag[2],
@@ -932,6 +1139,7 @@ psm3_mq_irecv2(psm2_mq_t mq, psm2_epaddr_t src,
 {
 	psm2_error_t err = PSM2_OK;
 	psm2_mq_req_t req;
+	int table;
 
 #if defined(PSM_CUDA) || defined(PSM_ONEAPI) 
 	int gpu_mem = 0;
@@ -949,7 +1157,7 @@ psm3_mq_irecv2(psm2_mq_t mq, psm2_epaddr_t src,
 	PSMI_LOCK(mq->progress_lock);
 
 	/* First check unexpected Queue and remove req if found */
-	req = mq_req_match_with_tagsel(mq, src, tag, tagsel, REMOVE_ENTRY);
+	req = mq_req_match_with_tagsel(mq, src, tag, tagsel, REMOVE_ENTRY, &table);
 
 	if (req == NULL) {
 		/* prepost before arrival, add to expected q */
@@ -977,7 +1185,7 @@ psm3_mq_irecv2(psm2_mq_t mq, psm2_epaddr_t src,
 			req->user_gpu_buffer = NULL;
 #endif
 
-		mq_add_to_expected_hashes(mq, req);
+		mq_add_to_expected_hashes(mq, table, req);
 		_HFI_VDBG("buf=%p,len=%d,tag=%08x.%08x.%08x "
 			  " tagsel=%08x.%08x.%08x req=%p\n",
 			  buf, len, tag->tag[0], tag->tag[1], tag->tag[2],
@@ -1023,10 +1231,10 @@ psm3_mq_irecv(psm2_mq_t mq, uint64_t tag, uint64_t tagsel, uint32_t flags,
 
 	rtag.tag64 = tag;
 #ifdef PSM_DEBUG
-	rtag.tag[2] = 0;
+	rtag.rem32 = 0;
 #endif
 	rtagsel.tag64 = tagsel;
-	rtagsel.tag[2] = 0;
+	rtagsel.rem32 = 0;
 	rv = psm3_mq_irecv2(mq, PSM2_MQ_ANY_ADDR, &rtag, &rtagsel,
 			       flags, buf, len, context, reqo);
 
@@ -1284,8 +1492,19 @@ psm2_error_t psm3_mq_setopt(psm2_mq_t mq, int key, const void *value)
 	STAT(tx_shm_num)		\
 	STAT(rx_shm_num)		\
 	STAT(rx_sysbuf_num)		\
-	STAT(rx_sysbuf_bytes)		\
+	STAT(rx_sysbuf_bytes)	\
 	STAT(comm_world_rank)
+//	STAT(max_subqueues)
+//	STAT(max_exp_list_len)
+//	STAT(max_exp_hash_len)
+//	STAT(num_exp_search)
+//	STAT(tot_exp_search_cmp)
+//	STAT(max_exp_search_cmp)
+//	STAT(max_unexp_list_len)
+//	STAT(max_unexp_hash_len)
+//	STAT(num_unexp_search)
+//	STAT(tot_unexp_search_cmp)
+//	STAT(max_unexp_search__cmp)
 
 static
 void
@@ -1297,7 +1516,7 @@ psm3_mq_print_stats(psm2_mq_t mq, FILE *perf_stats_fd)
 	psm3_mq_get_stats(sizeof(stats), mq, &stats);
 
 #define STAT(x) \
-	snprintf(msg_buffer, MSG_BUFFER_LEN, "%*lu",TAB_SIZE, stats.x); \
+	snprintf(msg_buffer, MSG_BUFFER_LEN, "%*lu ",TAB_SIZE-1, stats.x); \
 	fwrite(msg_buffer, sizeof(char), strlen(msg_buffer), perf_stats_fd);
 
 	STATS
@@ -1329,7 +1548,7 @@ void
 	}
 
 #define STAT(x) \
-	snprintf(msg_buffer, MSG_BUFFER_LEN, "%*s",TAB_SIZE, #x);\
+	snprintf(msg_buffer, MSG_BUFFER_LEN, "%*s ",TAB_SIZE-1, #x);\
 	fwrite(msg_buffer, sizeof(char), strlen(msg_buffer), perf_stats_fd);
 
 	STAT(delta_t)
@@ -1341,7 +1560,7 @@ void
 
 	/* Performance stats will be printed every $PSM3_MQ_PRINT_STATS seconds */
 	do {
-		snprintf(msg_buffer, MSG_BUFFER_LEN, "%*d",TAB_SIZE, delta_t);
+		snprintf(msg_buffer, MSG_BUFFER_LEN, "%*d ",TAB_SIZE-1, delta_t);
 		fwrite(msg_buffer, sizeof(char), strlen(msg_buffer), perf_stats_fd);
 		psm3_mq_print_stats(mq, perf_stats_fd);
 		fflush(perf_stats_fd);
@@ -1459,6 +1678,12 @@ void psm3_mq_get_stats(uint32_t len, psm2_mq_t mq, psm2_mq_stats_t *stats)
 		memset(stats, 0, len);
 	} else {
 		memcpy(stats, &mq->stats, len);
+#if ! defined(LEARN_HASH_SELECTOR) || ! defined(RETAIN_PAST_TABLE_SEL)
+		stats->max_subqueues = NUM_HASH_CONFIGS+1 - mq->min_min_table;
+#else
+		stats->max_subqueues = NUM_HASH_CONFIGS+1
+								- (mq->min_table + !mq->search_linear_expected);
+#endif
 	}
 	PSM2_LOG_MSG("leaving");
 }
@@ -1528,6 +1753,52 @@ static uint64_t expected_avg_msg_size_recv(void *context)
 		return 0;
 }
 
+static uint64_t num_subqueues(void *context)
+{
+	psm2_mq_t mq = (psm2_mq_t)context;
+	if (mq->nohash_fastpath)
+		return 1;
+	else
+#ifdef LEARN_HASH_SELECTOR
+		return NUM_HASH_CONFIGS+1
+					- (mq->min_table + !mq->search_linear_expected);
+#else
+		return NUM_HASH_CONFIGS+1 - mq->min_table;
+#endif
+}
+
+static uint64_t max_subqueues(void *context)
+{
+	psm2_mq_t mq = (psm2_mq_t)context;
+#if ! defined(LEARN_HASH_SELECTOR) || ! defined(RETAIN_PAST_TABLE_SEL)
+	return NUM_HASH_CONFIGS+1 - mq->min_min_table;
+#else
+	return NUM_HASH_CONFIGS+1 - (mq->min_table + !mq->search_linear_expected);
+#endif
+}
+
+static uint64_t expected_list_len(void *context)
+{
+	psm2_mq_t mq = (psm2_mq_t)context;
+	return mq->expected_list_len;
+}
+
+static uint64_t expected_hash_len(void *context)
+{
+	psm2_mq_t mq = (psm2_mq_t)context;
+	return mq->expected_hash_len;
+}
+
+static uint64_t avg_exp_search_cmp(void *context)
+{
+	psm2_mq_t mq = (psm2_mq_t)context;
+	if (mq->stats.num_exp_search)
+		return (mq->stats.tot_exp_search_cmp / mq->stats.num_exp_search);
+	else
+		return 0;
+}
+
+
 static uint64_t unexpected_avg_msg_size_recv(void *context)
 {
 	psm2_mq_t mq = (psm2_mq_t)context;
@@ -1537,6 +1808,27 @@ static uint64_t unexpected_avg_msg_size_recv(void *context)
 		return 0;
 }
 
+static uint64_t unexpected_list_len(void *context)
+{
+	psm2_mq_t mq = (psm2_mq_t)context;
+	return mq->unexpected_list_len;
+}
+
+static uint64_t unexpected_hash_len(void *context)
+{
+	psm2_mq_t mq = (psm2_mq_t)context;
+	return mq->unexpected_hash_len;
+}
+
+static uint64_t avg_unexp_search_cmp(void *context)
+{
+	psm2_mq_t mq = (psm2_mq_t)context;
+	if (mq->stats.num_unexp_search)
+		return (mq->stats.tot_unexp_search_cmp / mq->stats.num_unexp_search);
+	else
+		return 0;
+}
+
 static uint64_t sysbuf_avg_size_recv(void *context)
 {
 	psm2_mq_t mq = (psm2_mq_t)context;
@@ -1586,6 +1878,27 @@ static uint64_t shm_dsa_avg_copy_size_recv(void *context)
 }
 #endif /* PSM_DSA */
 
+#if defined(PSM_CUDA) || defined(PSM_ONEAPI)
+static uint64_t gpu_ipc_hit_rate(void *context)
+{
+	psm2_mq_t mq = (psm2_mq_t)context;
+	if (mq->stats.gpu_ipc_cache_miss)	// all entries start with a miss, then get hits
+		return((mq->stats.gpu_ipc_cache_hit*100)/(mq->stats.gpu_ipc_cache_miss+mq->stats.gpu_ipc_cache_hit));
+	else
+		return 0;
+}
+
+static uint64_t gpu_ipc_miss_rate(void *context)
+{
+	psm2_mq_t mq = (psm2_mq_t)context;
+	if (mq->stats.gpu_ipc_cache_miss)	// all entries start with a miss, then get hits
+		return((mq->stats.gpu_ipc_cache_miss*100)/(mq->stats.gpu_ipc_cache_miss+mq->stats.gpu_ipc_cache_hit));
+	else
+		return 0;
+}
+#endif /* defined(PSM_CUDA) || defined(PSM_ONEAPI) */
+
+
 static uint64_t self_avg_msg_size_sent(void *context)
 {
 	psm2_mq_t mq = (psm2_mq_t)context;
@@ -1698,6 +2011,14 @@ psm2_error_t psm3_mq_initstats(psm2_mq_t mq, psm2_epid_t epid)
 				"Overall average message size received",
 				overall_avg_msg_size_recv),
 
+		// ------------------------------------------------------------
+		PSMI_STATS_DECL_HELP("Overall expected queue tag matching statistics "
+			"across all PSM3 protocols (shm, nic, self). Newly arriving "
+			"messages first search the expected queue and if no match, post to "
+			"the unexpected queue. The expected queue has 1 or more subqueues. "
+			"The subqueues consist of a list or a list and a set of hash "
+			"tables. Each entry on the queue appears on the list or exactly 1 "
+			"hash table."),
 		PSMI_STATS_DECLU64("Expected_count_recv",
 				"Total messages received where receive posted by application prior to arrival",
 				&mq->stats.rx_user_num),
@@ -1705,19 +2026,90 @@ psm2_error_t psm3_mq_initstats(psm2_mq_t mq, psm2_epid_t epid)
 				"Total bytes received where receive posted by application prior to arrival",
 				&mq->stats.rx_user_bytes),
 		PSMI_STATS_DECL_FUNC("Expected_avg_msg_size_recv",
-				"average message size received where receive posted by application prior to arrival",
+				"Average message size received where receive posted by application prior to arrival",
 				expected_avg_msg_size_recv),
+		PSMI_STATS_DECL_FUNC("Expected_subqueues",
+				   "Current number of sub-queues used for expected message tag matching",
+				   num_subqueues),
+		PSMI_STATS_DECL_FUNC("Expected_max_subqueues",
+				   "Maximum number of sub-queues used for expected message tag matching",
+				   max_subqueues),
+		PSMI_STATS_DECL_FUNC("Expected_list_len",
+				   "Current number of items on list used for expected message tag matching",
+				   expected_list_len),
+		PSMI_STATS_DECLU64("Expected_max_list_len",
+				   "Maximum number of items on list used for expected message tag matching",
+				   &mq->stats.max_exp_list_len),
+		PSMI_STATS_DECL_FUNC("Expected_hash_len",
+				   "Current number of items on hash tables used for expected message tag matching",
+				   expected_hash_len),
+		PSMI_STATS_DECLU64("Expected_max_hash_len",
+				   "Maximum number of items on hash tables used for expected message tag matching",
+				   &mq->stats.max_exp_hash_len),
+		PSMI_STATS_DECLU64("Expected_num_search",
+				   "Total number of expected queue searches",
+				   &mq->stats.num_exp_search),
+		PSMI_STATS_DECLU64("Expected_total_search_cmp",
+				   "Total entries compared during expected queue searches",
+				   &mq->stats.tot_exp_search_cmp),
+		PSMI_STATS_DECL_FUNC("Expected_avg_search_cmp",
+				   "Average entries compared during expected queue searches",
+				   avg_exp_search_cmp),
+		PSMI_STATS_DECLU64("Expected_max_search_cmp",
+				   "Maximum entries compared during an expected queue search",
+				   &mq->stats.max_exp_search_cmp),
 
+		// ------------------------------------------------------------
+		PSMI_STATS_DECL_HELP("Overall unexpected queue tag matching statistics "
+			"across all PSM3 protocols (shm, nic, self).  Newly posted "
+			"receives first search the unexpected queue and if no match, post "
+			"to the expected queue. The unexpected queue has 1 or more "
+			"subqueues.  The subqueues consist of a list or a list and a set "
+			"of hash tables. Each entry on the queue appears on the list and "
+			"all hash tables."),
 		PSMI_STATS_DECLU64("Unexpected_count_recv",
 				"Total messages received where receive posted by application after arrival",
 				&mq->stats.rx_sys_num),
 		PSMI_STATS_DECLU64("Unexpected_bytes_recv",
 				"Total bytes received where receive posted by application after arrival",
 				&mq->stats.rx_sys_bytes),
-		PSMI_STATS_DECL_FUNC("Unxpected_avg_msg_size_recv",
+		PSMI_STATS_DECL_FUNC("Unexpected_avg_msg_size_recv",
 				"Average message size received where receive posted by application after arrival",
 				unexpected_avg_msg_size_recv),
+		PSMI_STATS_DECL_FUNC("Unexpected_subqueues",
+				   "Current number of sub-queues used for unexpected message tag matching",
+				   num_subqueues),
+		PSMI_STATS_DECL_FUNC("Unexpected_max_subqueues",
+				   "Maximum number of sub-queues used for unexpected message tag matching",
+				   max_subqueues),
+		PSMI_STATS_DECL_FUNC("Unexpected_list_len",
+				   "Current number of items on list used for unexpected message tag matching",
+				   unexpected_list_len),
+		PSMI_STATS_DECLU64("Unexpected_max_list_len",
+				   "Maximum number of items on list used for unexpected message tag matching",
+				   &mq->stats.max_unexp_list_len),
+		PSMI_STATS_DECL_FUNC("Unexpected_hash_len",
+				   "Current number of items on hash tables used for unexpected message tag matching",
+				   unexpected_hash_len),
+		PSMI_STATS_DECLU64("Unexpected_max_hash_len",
+				   "Maximum number of items on hash tables used for unexpected message tag matching",
+				   &mq->stats.max_unexp_hash_len),
+		PSMI_STATS_DECLU64("Unexpected_num_search",
+				   "Total number of unexpected queue searches",
+				   &mq->stats.num_unexp_search),
+		PSMI_STATS_DECLU64("Unexpected_total_search_cmp",
+				   "Total entries compared during unexpected queue searches",
+				   &mq->stats.tot_unexp_search_cmp),
+		PSMI_STATS_DECL_FUNC("Unexpected_avg_search_cmp",
+				   "Average entries compared during unexpected queue searches",
+				   avg_unexp_search_cmp),
+		PSMI_STATS_DECLU64("Unexpected_max_search_cmp",
+				   "Maximum entries compared during an unexpected queue search",
+				   &mq->stats.max_unexp_search_cmp),
 
+		// ------------------------------------------------------------
+		PSMI_STATS_DECL_HELP("Overall receive bounce buffer statistics across "
+			"all PSM3 protocols (shm, nic, self)"),
 		PSMI_STATS_DECLU64("sysbuf_count_recv",
 				"Total packets which used bounce buffers",
 				&mq->stats.rx_sysbuf_num),
@@ -1755,6 +2147,7 @@ psm2_error_t psm3_mq_initstats(psm2_mq_t mq, psm2_epid_t epid)
 				shm_avg_msg_size_recv),
 
 #ifdef PSM_DSA
+		// ------------------------------------------------------------
 		PSMI_STATS_DECL_HELP("For some Intel(r) Xeon(r) processor "
 			"models starting with the 4th Generation Intel(r) "
 			"Xeon(r) Scalable Processors, the Data Streaming "
@@ -1772,6 +2165,12 @@ psm2_error_t psm3_mq_initstats(psm2_mq_t mq, psm2_epid_t epid)
 		PSMI_STATS_DECL_FUNC("shm_dsa_avg_copy_size_sent",
 				"Average DSA send copy size",
 				shm_dsa_avg_copy_size_sent),
+		PSMI_STATS_DECLU64("shm_dsa_wait_swq_send_ns",
+				"Total nanoseconds spent during send waiting for DSA SWQ enqcmd",
+				&mq->stats.dsa_stats[0].dsa_swq_wait_ns),
+		PSMI_STATS_DECLU64("shm_dsa_no_wait_swq_send",
+				"Total DSA send copies with no wait for DSA SWQ enqcmd",
+				&mq->stats.dsa_stats[0].dsa_swq_no_wait),
 		PSMI_STATS_DECLU64("shm_dsa_wait_send_ns",
 				"Total nanoseconds spent during send waiting for DSA completions",
 				&mq->stats.dsa_stats[0].dsa_wait_ns),
@@ -1797,6 +2196,12 @@ psm2_error_t psm3_mq_initstats(psm2_mq_t mq, psm2_epid_t epid)
 		PSMI_STATS_DECL_FUNC("shm_dsa_avg_copy_size_recv",
 				"Average DSA receive copy size",
 				shm_dsa_avg_copy_size_recv),
+		PSMI_STATS_DECLU64("shm_dsa_wait_swq_recv_ns",
+				"Total nanoseconds spent during receive waiting for DSA SWQ enqcmd",
+				&mq->stats.dsa_stats[0].dsa_swq_wait_ns),
+		PSMI_STATS_DECLU64("shm_dsa_no_wait_swq_recv",
+				"Total DSA receive copies with no wait for DSA SWQ enqcmd",
+				&mq->stats.dsa_stats[0].dsa_swq_no_wait),
 		PSMI_STATS_DECLU64("shm_dsa_wait_recv_ns",
 				"Total nanoseconds spent during receive waiting for DSA completions",
 				&mq->stats.dsa_stats[1].dsa_wait_ns),
@@ -1813,6 +2218,52 @@ psm2_error_t psm3_mq_initstats(psm2_mq_t mq, psm2_epid_t epid)
 				"Total DSA receive copiess which failured for non-page fault error",
 				&mq->stats.dsa_stats[1].dsa_error),
 #endif /* PSM_DSA */
+#if defined(PSM_CUDA) || defined(PSM_ONEAPI)
+		// ------------------------------------------------------------
+		PSMI_STATS_DECL_HELP("Intra-node GPU messages may use GPU IPC Handles "
+			"to perform GPU to GPU rendezvous messages directly to and from "
+			"GPU application buffers. This may then take advantage of "
+			"optimized GPU to GPU HW mechanisms such as "
+#ifdef PSM_CUDA
+			"nvLink.\n"
+#else
+			"xeLink.\n"
+#endif
+			"To further optimize such transfers, a GPU IPC Handle cache is "
+			"maintained on the receiver."),
+		PSMI_STATS_DECLU64("gpu_ipc_cache_limit",
+				"Max IPC Handles allowed in GPU cache",
+				&mq->stats.gpu_ipc_cache_limit),
+		PSMI_STATS_DECLU64("gpu_ipc_cache_nelems",
+				"Current IPC Handles in GPU cache",
+				&mq->stats.gpu_ipc_cache_nelems),
+		PSMI_STATS_DECLU64("gpu_ipc_cache_max_nelems",
+				"Max Observed IPC Handles in GPU cache",
+				&mq->stats.gpu_ipc_cache_max_nelems),
+		PSMI_STATS_DECLU64("gpu_ipc_hit",
+				"Overall Number of hits",
+				&mq->stats.gpu_ipc_cache_hit),
+		PSMI_STATS_DECL("gpu_ipc_hit_%",
+				"Overall Cache hit rate",
+				MPSPAWN_STATS_REDUCTION_ALL,
+				gpu_ipc_hit_rate, NULL),
+		PSMI_STATS_DECLU64("gpu_ipc_miss",
+				"Overall Number of misses",
+				&mq->stats.gpu_ipc_cache_miss),
+		PSMI_STATS_DECL("gpu_ipc_miss_%",
+				"Overall Cache miss rate",
+				MPSPAWN_STATS_REDUCTION_ALL,
+				gpu_ipc_miss_rate, NULL),
+		PSMI_STATS_DECLU64("gpu_ipc_evict",
+				"Number of entries removed from cache due to no space",
+				&mq->stats.gpu_ipc_cache_evict),
+		PSMI_STATS_DECLU64("gpu_ipc_remove",
+				"Number of entries removed from cache due to being stale",
+				&mq->stats.gpu_ipc_cache_remove),
+		PSMI_STATS_DECLU64("gpu_ipc_clear",
+				"Number of times entire cache was cleared and reset due to error",
+				&mq->stats.gpu_ipc_cache_clear),
+#endif /* defined(PSM_CUDA) || defined(PSM_ONEAPI) */
 
 		// ------------------------------------------------------------
 		PSMI_STATS_DECL_HELP("The PSM3 self protocol is used in the "
@@ -1828,8 +2279,8 @@ psm2_error_t psm3_mq_initstats(psm2_mq_t mq, psm2_epid_t epid)
 				"Average message size sent using PSM3 self protocol",
 				self_avg_msg_size_sent),
 
-		// ------------------------------------------------------------
 #if defined(PSM_CUDA) || defined(PSM_ONEAPI)
+		// ------------------------------------------------------------
 		PSMI_STATS_DECL_HELP("Eager messages may be sent from GPU or "
 			"CPU application buffers.\n"
 			"When PSM3 receive bounce buffers (sysbuf) are used, the data "
@@ -1903,10 +2354,13 @@ psm2_error_t psm3_mq_initstats(psm2_mq_t mq, psm2_epid_t epid)
 		"receive request, the receiver issues one or more Clear to "
 		"Send (CTS). Upon receiving the CTS, the sender transfers the "
 		"corresponding message data.\n"
+		"If the application posts a receive before the matching message "
+		"arrives, it is considered 'expected' and the receive request "
+		"must be queued until a matching message arrives.\n"
 		"If a message arrives before the receiving application has "
 		"posted a receive, it is considered 'unexpected' and the "
 		"message must be queued until the application posts a receive. "
-		"If the received message was eager or an RTS contained "
+		"If the received message was eager or an RTS containing "
 		"payload, PSM3 receive bounce buffers (sysbuf) are used to hold the "
 		"data as it arrives. For multi-packet messages it's possible "
 		"to receive part of the message into bounce buffers and then "
@@ -1966,10 +2420,12 @@ fail:
 }
 
 // parse MQ env variables
+// called immediately after psm3_mq_malloc which has zeroed
+// fields already
 psm2_error_t psm3_mq_initialize_params(psm2_mq_t mq)
 {
 	union psmi_envvar_val env_hfitiny, env_rvwin, env_hfirv,
-		env_shmrv, env_stats;
+		env_shmrv, env_hash, env_stats;
 
 	// a limit of PSM_MQ_MAX_TINY btyes is hardcoded into the PSM protocol
 	psm3_getenv("PSM3_MQ_TINY_NIC_LIMIT",
@@ -1999,6 +2455,12 @@ psm2_error_t psm3_mq_initialize_params(psm2_mq_t mq)
 		    (union psmi_envvar_val)mq->shm_thresh_rv, &env_shmrv);
 	mq->shm_thresh_rv = env_shmrv.e_uint;
 
+	psm3_getenv("PSM3_MQ_HASH_THRESH",
+		    "linear list to hash tag matching switchover",
+		    PSMI_ENVVAR_LEVEL_HIDDEN, PSMI_ENVVAR_TYPE_UINT,
+		    (union psmi_envvar_val)DEFAULT_HASH_THRESH, &env_hash);
+	mq->hash_thresh = env_hash.e_uint;
+
 	psm3_getenv("PSM3_MQ_PRINT_STATS",
 		    "Prints MQ performance stats every n seconds to file "
 			"./psm3-perf-stat-ep-[epid]-pid-[pid] when set to -1 stats are "
@@ -2008,6 +2470,14 @@ psm2_error_t psm3_mq_initialize_params(psm2_mq_t mq)
 	mq->print_stats = env_stats.e_uint;
 
 	mq->nohash_fastpath = 1;
+	mq->min_table = NUM_HASH_CONFIGS;
+#ifdef LEARN_HASH_SELECTOR
+	mq->search_linear_expected = 0;
+#endif
+#if ! defined(LEARN_HASH_SELECTOR) || ! defined(RETAIN_PAST_TABLE_SEL)
+	//mq->min_min_table = mq->min_table + !mq->search_linear_expected;
+	mq->min_min_table = NUM_HASH_CONFIGS + 1;
+#endif
 	return PSM2_OK;
 }
 
