@@ -107,7 +107,7 @@ static ssize_t rxd_ep_cancel_recv(struct rxd_ep *ep, struct dlist_entry *list,
 	struct fi_cq_err_entry err_entry;
 	int ret = 0;
 
-	fastlock_acquire(&ep->util_ep.lock);
+	ofi_genlock_lock(&ep->util_ep.lock);
 
 	entry = dlist_remove_first_match(list, &rxd_match_ctx, context);
 	if (!entry)
@@ -127,14 +127,14 @@ static ssize_t rxd_ep_cancel_recv(struct rxd_ep *ep, struct dlist_entry *list,
 	rxd_rx_entry_free(ep, rx_entry);
 	ret = 1;
 out:
-	fastlock_release(&ep->util_ep.lock);
+	ofi_genlock_unlock(&ep->util_ep.lock);
 	return ret;
 }
 
 static ssize_t rxd_ep_cancel(fid_t fid, void *context)
 {
 	struct rxd_ep *ep;
-	int ret;
+	ssize_t ret;
 
 	ep = container_of(fid, struct rxd_ep, util_ep.ep_fid);
 
@@ -206,7 +206,7 @@ struct rxd_x_entry *rxd_rx_entry_init(struct rxd_ep *ep,
 	rx_entry->bytes_done = 0;
 	rx_entry->offset = 0;
 	rx_entry->next_seg_no = 0;
-	rx_entry->iov_count = iov_count;
+	rx_entry->iov_count = (uint8_t) iov_count;
 	rx_entry->op = op;
 	rx_entry->ignore = ignore;
 
@@ -223,7 +223,7 @@ struct rxd_x_entry *rxd_rx_entry_init(struct rxd_ep *ep,
 	return rx_entry;
 }
 
-int rxd_ep_post_buf(struct rxd_ep *ep)
+ssize_t rxd_ep_post_buf(struct rxd_ep *ep)
 {
 	struct rxd_pkt_entry *pkt_entry;
 	ssize_t ret;
@@ -250,7 +250,7 @@ int rxd_ep_post_buf(struct rxd_ep *ep)
 static int rxd_ep_enable(struct rxd_ep *ep)
 {
 	size_t i;
-	ssize_t ret;
+	int ret;
 
 	ret = fi_ep_bind(ep->dg_ep, &ep->dg_cq->fid, FI_TRANSMIT | FI_RECV);
 	if (ret)
@@ -263,26 +263,24 @@ static int rxd_ep_enable(struct rxd_ep *ep)
 	ep->tx_flags = rxd_tx_flags(ep->util_ep.tx_op_flags);
 	ep->rx_flags = rxd_rx_flags(ep->util_ep.rx_op_flags);
 
-	fastlock_acquire(&ep->util_ep.lock);
+	ofi_genlock_lock(&ep->util_ep.lock);
 	for (i = 0; i < ep->rx_size; i++) {
-		ret = rxd_ep_post_buf(ep);
-		if (ret)
+		if (rxd_ep_post_buf(ep))
 			break;
 	}
-
-	fastlock_release(&ep->util_ep.lock);
+	ofi_genlock_unlock(&ep->util_ep.lock);
 	return 0;
 }
 
 /*
  * Exponential back-off starting at 1ms, max 4s.
  */
-int rxd_get_timeout(uint8_t retry_cnt)
+int rxd_get_timeout(int retry_cnt)
 {
 	return MIN(1 << retry_cnt, 4000);
 }
 
-uint64_t rxd_get_retry_time(uint64_t start, uint8_t retry_cnt)
+uint64_t rxd_get_retry_time(uint64_t start, int retry_cnt)
 {
 	return start + rxd_get_timeout(retry_cnt);
 }
@@ -293,8 +291,8 @@ void rxd_init_data_pkt(struct rxd_ep *ep, struct rxd_x_entry *tx_entry,
 	struct rxd_data_pkt *data_pkt = (struct rxd_data_pkt *) (pkt_entry->pkt);
 	uint32_t seg_size;
 
-	seg_size = tx_entry->cq_entry.len - tx_entry->bytes_done;
-	seg_size = MIN(rxd_ep_domain(ep)->max_seg_sz, seg_size);
+	seg_size = (uint32_t) (tx_entry->cq_entry.len - tx_entry->bytes_done);
+	seg_size = (uint32_t) MIN(rxd_ep_domain(ep)->max_seg_sz, seg_size);
 
 	data_pkt->base_hdr.version = RXD_PROTOCOL_VERSION;
 	data_pkt->base_hdr.type = (tx_entry->cq_entry.flags &
@@ -305,7 +303,7 @@ void rxd_init_data_pkt(struct rxd_ep *ep, struct rxd_x_entry *tx_entry,
 	data_pkt->ext_hdr.rx_id = tx_entry->rx_id;
 	data_pkt->ext_hdr.tx_id = tx_entry->tx_id;
 	data_pkt->ext_hdr.seg_no = tx_entry->next_seg_no++;
-	data_pkt->base_hdr.peer = ep->peers[tx_entry->peer].peer_addr;
+	data_pkt->base_hdr.peer = (uint32_t) rxd_peer(ep, tx_entry->peer)->peer_addr;
 
 	pkt_entry->pkt_size = ofi_copy_from_iov(data_pkt->msg, seg_size,
 						tx_entry->iov,
@@ -343,7 +341,7 @@ struct rxd_x_entry *rxd_tx_entry_init_common(struct rxd_ep *ep, fi_addr_t addr,
 	tx_entry->bytes_done = 0;
 	tx_entry->offset = 0;
 	tx_entry->next_seg_no = 0;
-	tx_entry->iov_count = iov_count;
+	tx_entry->iov_count = (uint8_t) iov_count;
 	memcpy(&tx_entry->iov[0], iov, sizeof(*iov) * iov_count);
 
 	tx_entry->cq_entry.op_context = context;
@@ -360,7 +358,7 @@ struct rxd_x_entry *rxd_tx_entry_init_common(struct rxd_ep *ep, fi_addr_t addr,
 	rxd_init_base_hdr(ep, &(*ptr), tx_entry);
 
 	dlist_insert_tail(&tx_entry->entry,
-			  &ep->peers[tx_entry->peer].tx_list);
+			  &(rxd_peer(ep, tx_entry->peer)->tx_list));
 
 	return tx_entry;
 }
@@ -377,8 +375,8 @@ void rxd_insert_unacked(struct rxd_ep *ep, fi_addr_t peer,
 			struct rxd_pkt_entry *pkt_entry)
 {
 	dlist_insert_tail(&pkt_entry->d_entry,
-			  &ep->peers[peer].unacked);
-	ep->peers[peer].unacked_cnt++;
+			  &(rxd_peer(ep, peer)->unacked));
+	rxd_peer(ep, peer)->unacked_cnt++;
 }
 
 ssize_t rxd_ep_post_data_pkts(struct rxd_ep *ep, struct rxd_x_entry *tx_entry)
@@ -387,8 +385,8 @@ ssize_t rxd_ep_post_data_pkts(struct rxd_ep *ep, struct rxd_x_entry *tx_entry)
 	struct rxd_data_pkt *data;
 
 	while (tx_entry->bytes_done != tx_entry->cq_entry.len) {
-		if (ep->peers[tx_entry->peer].unacked_cnt >=
-		    ep->peers[tx_entry->peer].tx_window)
+		if (rxd_peer(ep, tx_entry->peer)->unacked_cnt >=
+		    rxd_peer(ep, tx_entry->peer)->tx_window)
 			return 0;
 
 		pkt_entry = rxd_get_tx_pkt(ep);
@@ -407,23 +405,24 @@ ssize_t rxd_ep_post_data_pkts(struct rxd_ep *ep, struct rxd_x_entry *tx_entry)
 		rxd_insert_unacked(ep, tx_entry->peer, pkt_entry);
 	}
 
-	return ep->peers[tx_entry->peer].unacked_cnt >=
-	       ep->peers[tx_entry->peer].tx_window;
+	return rxd_peer(ep, tx_entry->peer)->unacked_cnt >=
+	       rxd_peer(ep, tx_entry->peer)->tx_window;
 }
 
-int rxd_ep_send_pkt(struct rxd_ep *ep, struct rxd_pkt_entry *pkt_entry)
+ssize_t rxd_ep_send_pkt(struct rxd_ep *ep, struct rxd_pkt_entry *pkt_entry)
 {
-	int ret;
-
+	ssize_t ret;
+	fi_addr_t dg_addr;
 	pkt_entry->timestamp = ofi_gettime_ms();
 
+	dg_addr = (intptr_t) ofi_idx_lookup(&(rxd_ep_av(ep)->rxdaddr_dg_idx),
+					    (int)pkt_entry->peer);
 	ret = fi_send(ep->dg_ep, (const void *) rxd_pkt_start(pkt_entry),
-		      pkt_entry->pkt_size, pkt_entry->desc,
-		      rxd_ep_av(ep)->rxd_addr_table[pkt_entry->peer].dg_addr,
+		      pkt_entry->pkt_size, pkt_entry->desc, dg_addr,
 		      &pkt_entry->context);
 	if (ret) {
 		FI_WARN(&rxd_prov, FI_LOG_EP_CTRL, "error sending packet: %d (%s)\n",
-			ret, fi_strerror(-ret));
+			(int) ret, fi_strerror((int) -ret));
 		return ret;
 	}
 	pkt_entry->flags |= RXD_PKT_IN_USE;
@@ -461,15 +460,21 @@ static ssize_t rxd_ep_send_rts(struct rxd_ep *rxd_ep, fi_addr_t rxd_addr)
 
 	rxd_ep_send_pkt(rxd_ep, pkt_entry);
 	rxd_insert_unacked(rxd_ep, rxd_addr, pkt_entry);
-	dlist_insert_tail(&rxd_ep->peers[rxd_addr].entry, &rxd_ep->rts_sent_list);
+	dlist_insert_tail(&(rxd_peer(rxd_ep, rxd_addr)->entry),
+			  &rxd_ep->rts_sent_list);
 
 	return 0;
 }
 
 ssize_t rxd_send_rts_if_needed(struct rxd_ep *ep, fi_addr_t addr)
 {
-	if (ep->peers[addr].peer_addr == FI_ADDR_UNSPEC &&
-	    dlist_empty(&ep->peers[addr].unacked))
+	if (!rxd_peer(ep, addr)) {
+		if (rxd_create_peer(ep, addr) < 0)
+			return -FI_ENOMEM;
+	}
+
+	if (rxd_peer(ep, addr)->peer_addr == RXD_ADDR_INVALID &&
+	    dlist_empty(&(rxd_peer(ep, addr)->unacked)))
 		return rxd_ep_send_rts(ep, addr);
 	return 0;
 }
@@ -480,10 +485,10 @@ void rxd_init_base_hdr(struct rxd_ep *rxd_ep, void **ptr,
 	struct rxd_base_hdr *hdr = (struct rxd_base_hdr *) *ptr;
 
 	hdr->version = RXD_PROTOCOL_VERSION;
-	hdr->type = tx_entry->op;
+	hdr->type = (uint8_t) tx_entry->op;
 	hdr->seq_no = 0;
-	hdr->peer = rxd_ep->peers[tx_entry->peer].peer_addr;
-	hdr->flags = tx_entry->flags;
+	hdr->peer = (uint32_t) rxd_peer(rxd_ep, tx_entry->peer)->peer_addr;
+	hdr->flags = (uint16_t) tx_entry->flags;
 
 	*ptr = (char *) (*ptr) + sizeof(*hdr);
 }
@@ -496,7 +501,7 @@ void rxd_init_sar_hdr(void **ptr, struct rxd_x_entry *tx_entry,
 	hdr->size = tx_entry->cq_entry.len;
 	hdr->num_segs = tx_entry->num_segs;
 	hdr->tx_id = tx_entry->tx_id;
-	hdr->iov_count = iov_count;
+	hdr->iov_count = (uint8_t) iov_count;
 
 	*ptr = (char *) (*ptr) + sizeof(*hdr);
 }
@@ -569,10 +574,10 @@ void rxd_ep_send_ack(struct rxd_ep *rxd_ep, fi_addr_t peer)
 
 	ack->base_hdr.version = RXD_PROTOCOL_VERSION;
 	ack->base_hdr.type = RXD_ACK;
-	ack->base_hdr.peer = rxd_ep->peers[peer].peer_addr;
-	ack->base_hdr.seq_no = rxd_ep->peers[peer].rx_seq_no;
-	ack->ext_hdr.rx_id = rxd_ep->peers[peer].rx_window;
-	rxd_ep->peers[peer].last_tx_ack = ack->base_hdr.seq_no;
+	ack->base_hdr.peer = (uint32_t) rxd_peer(rxd_ep, peer)->peer_addr;
+	ack->base_hdr.seq_no = rxd_peer(rxd_ep, peer)->rx_seq_no;
+	ack->ext_hdr.rx_id = rxd_peer(rxd_ep, peer)->rx_window;
+	rxd_peer(rxd_ep, peer)->last_tx_ack = ack->base_hdr.seq_no;
 
 	dlist_insert_tail(&pkt_entry->d_entry, &rxd_ep->ctrl_pkts);
 	if (rxd_ep_send_pkt(rxd_ep, pkt_entry))
@@ -606,19 +611,19 @@ static void rxd_close_peer(struct rxd_ep *ep, struct rxd_peer *peer)
 		peer->unacked_cnt--;
 	}
 
-	while(!dlist_empty(&peer->tx_list)) {
+	while (!dlist_empty(&peer->tx_list)) {
 		dlist_pop_front(&peer->tx_list, struct rxd_x_entry,
 				x_entry, entry);
 		rxd_tx_entry_free(ep, x_entry);
 	}
 
-	while(!dlist_empty(&peer->rx_list)) {
+	while (!dlist_empty(&peer->rx_list)) {
 		dlist_pop_front(&peer->rx_list, struct rxd_x_entry,
 				x_entry, entry);
 		rxd_rx_entry_free(ep, x_entry);
 	}
 
-	while(!dlist_empty(&peer->rma_rx_list)) {
+	while (!dlist_empty(&peer->rma_rx_list)) {
 		dlist_pop_front(&peer->rma_rx_list, struct rxd_x_entry,
 				x_entry, entry);
 		rxd_tx_entry_free(ep, x_entry);
@@ -663,6 +668,9 @@ static int rxd_ep_close(struct fid *fid)
 
 	dlist_foreach_container(&ep->active_peers, struct rxd_peer, peer, entry)
 		rxd_close_peer(ep, peer);
+	dlist_foreach_container(&ep->rts_sent_list, struct rxd_peer, peer, entry)
+		rxd_close_peer(ep, peer);
+	ofi_idm_reset(&(ep->peers_idm), free);
 
 	ret = fi_close(&ep->dg_ep->fid);
 	if (ret)
@@ -910,7 +918,8 @@ static void rxd_progress_pkt_list(struct rxd_ep *ep, struct rxd_peer *peer)
 {
 	struct rxd_pkt_entry *pkt_entry;
 	uint64_t current;
-	int ret, retry = 0;
+	ssize_t ret;
+	int retry = 0;
 
 	current = ofi_gettime_ms();
 	if (peer->retry_cnt > RXD_MAX_PKT_RETRY) {
@@ -921,7 +930,8 @@ static void rxd_progress_pkt_list(struct rxd_ep *ep, struct rxd_peer *peer)
 	dlist_foreach_container(&peer->unacked, struct rxd_pkt_entry,
 				pkt_entry, d_entry) {
 		if (pkt_entry->flags & (RXD_PKT_IN_USE | RXD_PKT_ACKED) ||
-		    current < rxd_get_retry_time(pkt_entry->timestamp, peer->retry_cnt))
+		    current < rxd_get_retry_time(pkt_entry->timestamp,
+						 (uint8_t) peer->retry_cnt))
 			break;
 		retry = 1;
 		ret = rxd_ep_send_pkt(ep, pkt_entry);
@@ -947,7 +957,7 @@ void rxd_ep_progress(struct util_ep *util_ep)
 
 	ep = container_of(util_ep, struct rxd_ep, util_ep);
 
-	fastlock_acquire(&ep->util_ep.lock);
+	ofi_genlock_lock(&ep->util_ep.lock);
 	for(ret = 1, i = 0;
 	    ret > 0 && (!rxd_env.spin_count || i < rxd_env.spin_count);
 	    i++) {
@@ -982,7 +992,7 @@ void rxd_ep_progress(struct util_ep *util_ep)
 	}
 
 out:
-	fastlock_release(&ep->util_ep.lock);
+	ofi_genlock_unlock(&ep->util_ep.lock);
 }
 
 static int rxd_buf_region_alloc_fn(struct ofi_bufpool_region *region)
@@ -1027,9 +1037,9 @@ static void rxd_pkt_init_fn(struct ofi_bufpool_region *region, void *buf)
 	struct rxd_buf_pool *pool = (struct rxd_buf_pool *) region->pool->attr.context;
 
  	if (pool->type == RXD_BUF_POOL_TX)
-		entry->tx_id = ofi_buf_index(entry);
+		entry->tx_id = (uint16_t) ofi_buf_index(entry);
 	else
-		entry->rx_id = ofi_buf_index(entry);
+		entry->rx_id = (uint16_t) ofi_buf_index(entry);
 }
 
 static void rxd_buf_region_free_fn(struct ofi_bufpool_region *region)
@@ -1136,23 +1146,38 @@ err:
 	return ret;
 }
 
-static void rxd_init_peer(struct rxd_ep *ep, uint64_t rxd_addr)
+int rxd_create_peer(struct rxd_ep *ep, uint64_t rxd_addr)
 {
-	ep->peers[rxd_addr].peer_addr = FI_ADDR_UNSPEC;
-	ep->peers[rxd_addr].tx_seq_no = 0;
-	ep->peers[rxd_addr].rx_seq_no = 0;
-	ep->peers[rxd_addr].last_rx_ack = 0;
-	ep->peers[rxd_addr].last_tx_ack = 0;
-	ep->peers[rxd_addr].rx_window = rxd_env.max_unacked;
-	ep->peers[rxd_addr].tx_window = rxd_env.max_unacked;
-	ep->peers[rxd_addr].unacked_cnt = 0;
-	ep->peers[rxd_addr].retry_cnt = 0;
-	ep->peers[rxd_addr].active = 0;
-	dlist_init(&ep->peers[rxd_addr].unacked);
-	dlist_init(&ep->peers[rxd_addr].tx_list);
-	dlist_init(&ep->peers[rxd_addr].rx_list);
-	dlist_init(&ep->peers[rxd_addr].rma_rx_list);
-	dlist_init(&ep->peers[rxd_addr].buf_pkts);
+
+	struct rxd_peer *peer;
+
+	peer = calloc(1, sizeof(struct rxd_peer));
+	if (!peer)
+		return -FI_ENOMEM;
+
+	peer->peer_addr = RXD_ADDR_INVALID;
+	peer->tx_seq_no = 0;
+	peer->rx_seq_no = 0;
+	peer->last_rx_ack = 0;
+	peer->last_tx_ack = 0;
+	peer->rx_window = (uint16_t) rxd_env.max_unacked;
+	peer->tx_window = (uint16_t) rxd_env.max_unacked;
+	peer->unacked_cnt = 0;
+	peer->retry_cnt = 0;
+	peer->active = 0;
+	dlist_init(&(peer->unacked));
+	dlist_init(&(peer->tx_list));
+	dlist_init(&(peer->rx_list));
+	dlist_init(&(peer->rma_rx_list));
+	dlist_init(&(peer->buf_pkts));
+
+	if (ofi_idm_set(&(ep->peers_idm), (int) rxd_addr, peer) < 0)
+		goto err;
+
+	return 0;
+err:
+	free(peer);
+	return -FI_ENOMEM;
 }
 
 int rxd_endpoint(struct fid_domain *domain, struct fi_info *info,
@@ -1161,10 +1186,9 @@ int rxd_endpoint(struct fid_domain *domain, struct fi_info *info,
 	struct fi_info *dg_info;
 	struct rxd_domain *rxd_domain;
 	struct rxd_ep *rxd_ep;
-	int ret, i;
+	int ret;
 
-	rxd_ep = calloc(1, sizeof(*rxd_ep) + sizeof(struct rxd_peer) *
-			rxd_env.max_peers);
+	rxd_ep = calloc(1, sizeof(*rxd_ep));
 	if (!rxd_ep)
 		return -FI_ENOMEM;
 
@@ -1177,7 +1201,7 @@ int rxd_endpoint(struct fid_domain *domain, struct fi_info *info,
 		goto err1;
 
 	ret = ofi_get_core_info(rxd_domain->util_domain.fabric->fabric_fid.api_version,
-				NULL, NULL, 0, &rxd_util_prov, info,
+				NULL, NULL, 0, &rxd_util_prov, info, NULL,
 				rxd_info_to_core, &dg_info);
 	if (ret)
 		goto err2;
@@ -1206,8 +1230,7 @@ int rxd_endpoint(struct fid_domain *domain, struct fi_info *info,
 	if (ret)
 		goto err3;
 
-	for (i = 0; i < rxd_env.max_peers; rxd_init_peer(rxd_ep, i++))
-		;
+	memset(&(rxd_ep->peers_idm), 0, sizeof(rxd_ep->peers_idm));
 
 	rxd_ep->util_ep.ep_fid.fid.ops = &rxd_ep_fi_ops;
 	rxd_ep->util_ep.ep_fid.cm = &rxd_ep_cm;
