@@ -37,6 +37,7 @@
 
 #include "ofi_hmem.h"
 #include "ofi.h"
+#include "ofi_iov.h"
 
 #if HAVE_GDRCOPY
 
@@ -292,7 +293,45 @@ void cuda_gdrcopy_from_dev(uint64_t handle, void *hostptr,
 			  GDRCOPY_FROM_DEVICE);
 }
 
-int cuda_gdrcopy_dev_register(struct fi_mr_attr *mr_attr, uint64_t *handle)
+static ssize_t cuda_gdrcopy_iov_buf(uint64_t handle,
+                                    const struct iovec *iov, size_t iov_count,
+                                    uint64_t iov_offset, void *buf,
+                                    size_t size, enum gdrcopy_dir dir)
+{
+	uint64_t done = 0, len;
+	char *dev_buf;
+	size_t i;
+
+	for (i = 0; i < iov_count && size; i++) {
+		len = ofi_iov_bytes_to_copy(&iov[i], &size, &iov_offset, &dev_buf);
+
+		if (!len)
+			continue;
+
+		cuda_gdrcopy_impl(handle, dev_buf, (void *)((char *)buf + done), len, dir);
+
+		done += len;
+	}
+	return done;
+}
+
+ssize_t ofi_gdrcopy_to_cuda_iov(uint64_t handle, const struct iovec *iov,
+                                size_t iov_count, uint64_t iov_offset,
+                                const void *host, size_t len)
+{
+	return cuda_gdrcopy_iov_buf(handle, iov, iov_count, iov_offset,
+	                           (void *) host, len, GDRCOPY_TO_DEVICE);
+}
+
+ssize_t ofi_gdrcopy_from_cuda_iov(uint64_t handle, void *host,
+                                  const struct iovec *iov, size_t iov_count,
+                                  uint64_t iov_offset, size_t len)
+{
+	return cuda_gdrcopy_iov_buf(handle, iov, iov_count, iov_offset,
+	                            host, len, GDRCOPY_FROM_DEVICE);
+}
+
+int cuda_gdrcopy_dev_register(const void *buf, size_t len, uint64_t *handle)
 {
 	int err;
 	uintptr_t regbgn, regend;
@@ -303,8 +342,8 @@ int cuda_gdrcopy_dev_register(struct fi_mr_attr *mr_attr, uint64_t *handle)
 	assert(global_gdrcopy_ops.gdr_pin_buffer);
 	assert(global_gdrcopy_ops.gdr_map);
 
-	regbgn = (uintptr_t)ofi_get_page_start(mr_attr->mr_iov->iov_base, GPU_PAGE_SIZE);
-	regend = (uintptr_t)mr_attr->mr_iov->iov_base + mr_attr->mr_iov->iov_len;
+	regbgn = (uintptr_t)ofi_get_page_start(buf, GPU_PAGE_SIZE);
+	regend = (uintptr_t)buf + len;
 	reglen = ofi_get_aligned_size(regend - regbgn, GPU_PAGE_SIZE);
 
 	gdrcopy = malloc(sizeof(struct gdrcopy_handle));
@@ -318,7 +357,7 @@ int cuda_gdrcopy_dev_register(struct fi_mr_attr *mr_attr, uint64_t *handle)
 	if (err) {
 		FI_WARN(&core_prov, FI_LOG_CORE,
 			"gdr_pin_buffer failed! error: %s ptr: %p len: %ld\n",
-			strerror(err), mr_attr->mr_iov->iov_base, mr_attr->mr_iov->iov_len);
+			strerror(err), buf, len);
 		free(gdrcopy);
 		goto exit;
 	}
@@ -400,7 +439,7 @@ void cuda_gdrcopy_from_dev(uint64_t devhandle, void *hostptr,
 {
 }
 
-int cuda_gdrcopy_dev_register(struct fi_mr_attr *mr_attr, uint64_t *handle)
+int cuda_gdrcopy_dev_register(const void *buf, size_t len, uint64_t *handle)
 {
 	return FI_SUCCESS;
 }
@@ -410,4 +449,18 @@ int cuda_gdrcopy_dev_unregister(uint64_t handle)
 	return FI_SUCCESS;
 }
 
+ssize_t ofi_gdrcopy_to_cuda_iov(uint64_t handle, const struct iovec *iov,
+                                size_t iov_count, uint64_t iov_offset,
+                                const void *host, size_t len)
+{
+	return -FI_ENOSYS;
+}
+
+ssize_t ofi_gdrcopy_from_cuda_iov(uint64_t handle, void *host,
+                                  const struct iovec *iov, size_t iov_count,
+                                  uint64_t iov_offset, size_t len)
+{
+	return -FI_ENOSYS;
+}
+
 #endif /* HAVE_GDRCOPY */
