@@ -1,5 +1,5 @@
 /**
- * Copyright (C) Mellanox Technologies Ltd. 2018.       ALL RIGHTS RESERVED.
+ * Copyright (c) NVIDIA CORPORATION & AFFILIATES, 2018. ALL RIGHTS RESERVED.
  *
  * See file LICENSE for terms.
  */
@@ -11,25 +11,10 @@
 #include <stdint.h>
 
 #include <ucs/type/status.h>
-#include <ucs/sys/compiler_def.h>
 
 #define UCM_BISTRO_PROLOGUE
 #define UCM_BISTRO_EPILOGUE
 
-/* Patch by jumping to absolute address loaded from register */
-typedef struct ucm_bistro_jmp_r11_patch {
-    uint8_t mov_r11[2];  /* mov %r11, addr */
-    void    *ptr;
-    uint8_t jmp_r11[3];  /* jmp r11        */
-} UCS_S_PACKED ucm_bistro_jmp_r11_patch_t;
-
-
-/* Patch by jumping to relative address by immediate displacement */
-typedef struct ucm_bistro_jmp_near_patch {
-    uint8_t jmp_rel; /* opcode:  JMP rel32          */
-    int32_t disp;    /* operand: jump displacement */
-} UCS_S_PACKED ucm_bistro_jmp_near_patch_t;
-
 
 /**
  * Set library function call hook using Binary Instrumentation
