@@ -1,5 +1,5 @@
 /**
- * Copyright (C) Mellanox Technologies Ltd. 2001-2021.  ALL RIGHTS RESERVED.
+ * Copyright (c) NVIDIA CORPORATION & AFFILIATES, 2001-2021. ALL RIGHTS RESERVED.
  *
  * See file LICENSE for terms.
  */
@@ -15,12 +15,36 @@
 #include <ucs/sys/module.h>
 #include <ucs/sys/string.h>
 #include <ucs/vfs/base/vfs_obj.h>
+#include <uct/tcp/tcp.h>
+#include <uct/sm/self/self.h>
+#include <uct/sm/mm/base/mm_iface.h>
 #include <limits.h>
 #include <string.h>
 
 
 UCS_LIST_HEAD(uct_components_list);
 
+UCT_TL_DECL(self)
+UCT_TL_DECL(tcp)
+UCT_TL_DECL(posix)
+UCT_TL_DECL(sysv)
+
+void UCS_F_CTOR uct_init()
+{
+    uct_self_init();
+    uct_tcp_init();
+    uct_sysv_init();
+    uct_posix_init();
+}
+
+void UCS_F_DTOR uct_cleanup()
+{
+    uct_posix_cleanup();
+    uct_sysv_cleanup();
+    uct_tcp_cleanup();
+    uct_self_cleanup();
+}
+
 ucs_status_t uct_query_components(uct_component_h **components_p,
                                   unsigned *num_components_p)
 {
@@ -100,19 +124,19 @@ ucs_status_t uct_component_query(uct_component_h component,
 }
 
 ucs_status_t uct_config_read(uct_config_bundle_t **bundle,
-                             ucs_config_field_t *config_table,
-                             size_t config_size, const char *env_prefix,
-                             const char *cfg_prefix)
+                             ucs_config_global_list_entry_t *entry,
+                             const char *env_prefix)
 {
     char full_prefix[128] = UCS_DEFAULT_ENV_PREFIX;
     uct_config_bundle_t *config_bundle;
     ucs_status_t status;
 
-    if (config_table == NULL) {
+    if (entry->table == NULL) {
         return UCS_ERR_INVALID_PARAM;
     }
 
-    config_bundle = ucs_calloc(1, sizeof(*config_bundle) + config_size, "uct_config");
+    config_bundle = ucs_calloc(1, sizeof(*config_bundle) + entry->size,
+                               "uct_config");
     if (config_bundle == NULL) {
         status = UCS_ERR_NO_MEMORY;
         goto err;
@@ -124,14 +148,14 @@ ucs_status_t uct_config_read(uct_config_bundle_t **bundle,
                           env_prefix, UCS_DEFAULT_ENV_PREFIX);
     }
 
-    status = ucs_config_parser_fill_opts(config_bundle->data, config_table,
-                                         full_prefix, cfg_prefix, 0);
+    status = ucs_config_parser_fill_opts(config_bundle->data, entry,
+                                         full_prefix, 0);
     if (status != UCS_OK) {
         goto err_free_bundle;
     }
 
-    config_bundle->table = config_table;
-    config_bundle->table_prefix = ucs_strdup(cfg_prefix, "uct_config");
+    config_bundle->table        = entry->table;
+    config_bundle->table_prefix = ucs_strdup(entry->prefix, "uct_config");
     if (config_bundle->table_prefix == NULL) {
         status = UCS_ERR_NO_MEMORY;
         goto err_free_bundle;
@@ -145,3 +169,17 @@ err_free_bundle:
 err:
     return status;
 }
+
+void uct_component_register(uct_component_t *component)
+{
+    ucs_list_add_tail(&uct_components_list, &component->list);
+    ucs_list_add_tail(&ucs_config_global_list, &component->md_config.list);
+    ucs_list_add_tail(&ucs_config_global_list, &component->cm_config.list);
+}
+
+void uct_component_unregister(uct_component_t *component)
+{
+    /* TODO: add ucs_list_del(uct_components_list) */
+    ucs_list_del(&component->md_config.list);
+    ucs_list_del(&component->cm_config.list);
+}
