@@ -1,5 +1,5 @@
 /**
-* Copyright (C) Mellanox Technologies Ltd. 2001-2019.  ALL RIGHTS RESERVED.
+* Copyright (c) NVIDIA CORPORATION & AFFILIATES, 2001-2019. ALL RIGHTS RESERVED.
 *
 * See file LICENSE for terms.
 */
@@ -14,11 +14,11 @@
 #endif
 
 #include <uct/ib/mlx5/ib_mlx5_log.h>
-#include <uct/ib/mlx5/exp/ib_exp.h>
 #include <ucs/vfs/base/vfs_cb.h>
 #include <ucs/vfs/base/vfs_obj.h>
 #include <ucs/arch/cpu.h>
 #include <ucs/sys/compiler.h>
+#include <ucs/type/serialize.h>
 #include <arpa/inet.h> /* For htonl */
 
 #include "rc_mlx5.inl"
@@ -47,9 +47,9 @@ uct_rc_mlx5_ep_zcopy_post(uct_rc_mlx5_ep_t *ep, unsigned opcode,
                                    opcode, iov, iovcnt,
                                    am_id, am_hdr, am_hdr_len,
                                    rdma_raddr, uct_ib_md_direct_rkey(rdma_rkey),
-                                   tag, app_ctx, ib_imm_be,
-                                   NULL, NULL, 0,
+                                   tag, app_ctx, ib_imm_be, 0,
                                    (comp == NULL) ? force_sig : MLX5_WQE_CTRL_CQ_UPDATE,
+                                   0,
                                    UCT_IB_MAX_ZCOPY_LOG_SGE(&iface->super.super));
 
     uct_rc_txqp_add_send_comp(&iface->super, &ep->super.txqp, handler, comp, sn,
@@ -73,7 +73,8 @@ uct_rc_mlx5_ep_put_short_inline(uct_ep_h tl_ep, const void *buffer, unsigned len
                                  &ep->super.txqp, &ep->tx.wq,
                                  MLX5_OPCODE_RDMA_WRITE,
                                  buffer, length, 0, 0, 0, remote_addr, rkey,
-                                 NULL, NULL, 0, 0, INT_MAX);
+                                 0, 0, 0, INT_MAX);
+    uct_rc_ep_enable_flush_remote(&ep->super);
     UCT_TL_EP_STAT_OP(&ep->super.super, PUT, SHORT, length);
     return UCS_OK;
 }
@@ -92,11 +93,11 @@ uct_rc_mlx5_ep_am_short_inline(uct_ep_h tl_ep, uint8_t id, uint64_t hdr,
                                  payload, length,
                                  id, hdr, 0,
                                  0, 0,
-                                 NULL, NULL, 0,
+                                 0,
                                  MLX5_WQE_CTRL_SOLICITED,
-                                 INT_MAX);
+                                 0, INT_MAX);
     UCT_TL_EP_STAT_OP(&ep->super.super, AM, SHORT, sizeof(hdr) + length);
-    UCT_RC_UPDATE_FC(&iface->super, &ep->super, id);
+    UCT_RC_UPDATE_FC(&ep->super, id);
     return UCS_OK;
 }
 
@@ -109,9 +110,9 @@ static ucs_status_t UCS_F_ALWAYS_INLINE uct_rc_mlx5_ep_am_short_iov_inline(
     UCT_RC_CHECK_RES_AND_FC(&iface->super, &ep->super, id);
     uct_rc_mlx5_txqp_inline_iov_post(iface, IBV_QPT_RC, &ep->super.txqp,
                                      &ep->tx.wq, iov, iovcnt, iov_length, id,
-                                     NULL, NULL, 0);
+                                     0, 0);
     UCT_TL_EP_STAT_OP(&ep->super.super, AM, SHORT, iov_length);
-    UCT_RC_UPDATE_FC(&iface->super, &ep->super, id);
+    UCT_RC_UPDATE_FC(&ep->super, id);
 
     return UCS_OK;
 }
@@ -137,13 +138,14 @@ uct_rc_mlx5_ep_put_short(uct_ep_h tl_ep, const void *buffer, unsigned length,
                              ep->super.atomic_mr_offset);
     status = uct_rc_mlx5_common_ep_short_dm(iface, IBV_QPT_RC, NULL, 0, buffer,
                                             length, MLX5_OPCODE_RDMA_WRITE,
-                                            MLX5_WQE_CTRL_CQ_UPDATE,
+                                            MLX5_WQE_CTRL_CQ_UPDATE, 0,
                                             remote_addr, rkey, &ep->super.txqp,
-                                            &ep->tx.wq, NULL, NULL, 0);
+                                            &ep->tx.wq, 0);
     if (UCS_STATUS_IS_ERR(status)) {
         return status;
     }
 
+    uct_rc_ep_enable_flush_remote(&ep->super);
     UCT_TL_EP_STAT_OP(&ep->super.super, PUT, SHORT, length);
     return UCS_OK;
 #endif
@@ -163,9 +165,10 @@ ssize_t uct_rc_mlx5_ep_put_bcopy(uct_ep_h tl_ep, uct_pack_callback_t pack_cb,
                              ep->super.atomic_mr_offset);
     uct_rc_mlx5_common_txqp_bcopy_post(iface, IBV_QPT_RC, &ep->super.txqp,
                                        &ep->tx.wq, MLX5_OPCODE_RDMA_WRITE,
-                                       length, remote_addr, rkey, NULL, NULL, 0,
-                                       MLX5_WQE_CTRL_CQ_UPDATE, 0, desc,
+                                       length, remote_addr, rkey, 0,
+                                       MLX5_WQE_CTRL_CQ_UPDATE, 0, 0, desc,
                                        desc + 1, NULL);
+    uct_rc_ep_enable_flush_remote(&ep->super);
     UCT_TL_EP_STAT_OP(&ep->super.super, PUT, BCOPY, length);
 
     return length;
@@ -194,6 +197,7 @@ ucs_status_t uct_rc_mlx5_ep_put_zcopy(uct_ep_h tl_ep, const uct_iov_t *iov, size
                                        comp);
     UCT_TL_EP_STAT_OP_IF_SUCCESS(status, &ep->super.super, PUT, ZCOPY,
                                  uct_iov_total_length(iov, iovcnt));
+    uct_rc_ep_enable_flush_remote(&ep->super);
     return status;
 }
 
@@ -215,8 +219,8 @@ ucs_status_t uct_rc_mlx5_ep_get_bcopy(uct_ep_h tl_ep,
     uct_rc_mlx5_ep_fence_get(iface, &ep->tx.wq, &rkey, &fm_ce_se);
     uct_rc_mlx5_common_txqp_bcopy_post(iface, IBV_QPT_RC, &ep->super.txqp,
                                        &ep->tx.wq, MLX5_OPCODE_RDMA_READ,
-                                       length, remote_addr, rkey, NULL, NULL, 0,
-                                       fm_ce_se, 0, desc, desc + 1, NULL);
+                                       length, remote_addr, rkey, 0, fm_ce_se,
+                                       0, 0, desc, desc + 1, NULL);
     UCT_TL_EP_STAT_OP(&ep->super.super, GET, BCOPY, length);
     UCT_RC_RDMA_READ_POSTED(&iface->super, length);
 
@@ -258,7 +262,6 @@ uct_rc_mlx5_ep_am_short(uct_ep_h tl_ep, uint8_t id, uint64_t hdr,
 {
 #if HAVE_IBV_DM
     UCT_RC_MLX5_EP_DECL(tl_ep, iface, ep);
-    uct_rc_iface_t *rc_iface = &iface->super;
     ucs_status_t status;
     uct_rc_mlx5_dm_copy_data_t cache;
 
@@ -280,13 +283,13 @@ uct_rc_mlx5_ep_am_short(uct_ep_h tl_ep, uint8_t id, uint64_t hdr,
     status = uct_rc_mlx5_common_ep_short_dm(
             iface, IBV_QPT_RC, &cache, sizeof(cache.am_hdr), payload, length,
             MLX5_OPCODE_SEND, MLX5_WQE_CTRL_SOLICITED | MLX5_WQE_CTRL_CQ_UPDATE,
-            0, 0, &ep->super.txqp, &ep->tx.wq, NULL, NULL, 0);
+            0, 0, 0, &ep->super.txqp, &ep->tx.wq, 0);
     if (UCS_STATUS_IS_ERR(status)) {
         return status;
     }
 
     UCT_TL_EP_STAT_OP(&ep->super.super, AM, SHORT, sizeof(cache.am_hdr) + length);
-    UCT_RC_UPDATE_FC(rc_iface, &ep->super, id);
+    UCT_RC_UPDATE_FC(&ep->super, id);
     return UCS_OK;
 #endif
 }
@@ -314,12 +317,12 @@ ucs_status_t uct_rc_mlx5_ep_am_short_iov(uct_ep_h tl_ep, uint8_t id,
     status = uct_rc_mlx5_common_ep_am_short_iov_dm(&ep->super.super, id, iface,
                                                    iov, iovcnt, iov_length,
                                                    IBV_QPT_RC, &ep->super.txqp,
-                                                   &ep->tx.wq, NULL, NULL, 0);
+                                                   &ep->tx.wq, 0, 0);
     if (ucs_unlikely(UCS_STATUS_IS_ERR(status))) {
         return status;
     }
 
-    UCT_RC_UPDATE_FC(&iface->super, &ep->super, id);
+    UCT_RC_UPDATE_FC(&ep->super, id);
 
     return UCS_OK;
 #endif
@@ -342,10 +345,10 @@ ssize_t uct_rc_mlx5_ep_am_bcopy(uct_ep_h tl_ep, uint8_t id,
     uct_rc_mlx5_common_txqp_bcopy_post(iface, IBV_QPT_RC, &ep->super.txqp,
                                        &ep->tx.wq, MLX5_OPCODE_SEND,
                                        sizeof(uct_rc_mlx5_hdr_t) + length, 0, 0,
-                                       NULL, NULL, 0, MLX5_WQE_CTRL_SOLICITED,
-                                       0, desc, desc + 1, NULL);
+                                       0, MLX5_WQE_CTRL_SOLICITED, 0, 0, desc,
+                                       desc + 1, NULL);
     UCT_TL_EP_STAT_OP(&ep->super.super, AM, BCOPY, length);
-    UCT_RC_UPDATE_FC(&iface->super, &ep->super, id);
+    UCT_RC_UPDATE_FC(&ep->super, id);
 
     return length;
 }
@@ -372,7 +375,7 @@ ucs_status_t uct_rc_mlx5_ep_am_zcopy(uct_ep_h tl_ep, uint8_t id, const void *hea
     if (ucs_likely(status >= 0)) {
         UCT_TL_EP_STAT_OP(&ep->super.super, AM, ZCOPY,
                           header_length + uct_iov_total_length(iov, iovcnt));
-        UCT_RC_UPDATE_FC(&iface->super, &ep->super, id);
+        UCT_RC_UPDATE_FC(&ep->super, id);
     }
     return status;
 }
@@ -394,9 +397,9 @@ uct_rc_mlx5_ep_atomic_post(uct_ep_h tl_ep, unsigned opcode,
                                opcode, desc + 1, length, &desc->lkey,
                                remote_addr, ib_rkey,
                                compare_mask, compare, swap_mask, swap_add,
-                               NULL, NULL, 0, MLX5_WQE_CTRL_CQ_UPDATE,
-                               0, INT_MAX, NULL);
+                               0, MLX5_WQE_CTRL_CQ_UPDATE, 0, 0, INT_MAX, NULL);
 
+    uct_rc_ep_enable_flush_remote(&ep->super);
     UCT_TL_EP_STAT_ATOMIC(&ep->super.super);
     uct_rc_txqp_add_send_op(&ep->super.txqp, &desc->super);
 }
@@ -452,6 +455,7 @@ uct_rc_mlx5_ep_atomic_op_post(uct_ep_h tl_ep, unsigned opcode, unsigned size,
 
     uct_rc_mlx5_ep_atomic_post(tl_ep, op, desc, size, remote_addr, rkey,
                                compare_mask, compare, swap_mask, swap);
+    uct_rc_ep_enable_flush_remote(&ep->super);
     return UCS_OK;
 }
 
@@ -546,8 +550,8 @@ void uct_rc_mlx5_ep_post_check(uct_ep_h tl_ep)
                                  MLX5_OPCODE_RDMA_WRITE, &dummy, 0,
                                  0, 0, 0,
                                  0, 0,
-                                 NULL, NULL, 0, 0,
-                                 INT_MAX);
+                                 0, 0,
+                                 0, INT_MAX);
 }
 
 void uct_rc_mlx5_ep_vfs_populate(uct_rc_ep_t *rc_ep)
@@ -561,15 +565,48 @@ void uct_rc_mlx5_ep_vfs_populate(uct_rc_ep_t *rc_ep)
     uct_rc_txqp_vfs_populate(&ep->super.txqp, ep);
 }
 
+static ucs_status_t
+uct_rc_mlx5_ep_flush_remote(uct_ep_h tl_ep, uct_completion_t *comp)
+{
+    UCT_RC_MLX5_EP_DECL(tl_ep, iface, ep);
+    uct_rc_iface_send_desc_t *desc;
+
+    UCT_RC_CHECK_RES(&iface->super, &ep->super);
+
+    UCT_RC_IFACE_GET_TX_DESC(iface, &iface->super.tx.mp, desc);
+    desc->super.handler   = uct_rc_ep_flush_remote_handler;
+    desc->super.user_comp = comp;
+
+    uct_rc_mlx5_common_txqp_bcopy_post(iface, IBV_QPT_RC, &ep->super.txqp,
+                                       &ep->tx.wq, MLX5_OPCODE_RDMA_READ,
+                                       UCT_IB_MD_FLUSH_REMOTE_LENGTH, 0,
+                                       ep->super.flush_rkey, 0,
+                                       MLX5_WQE_CTRL_CQ_UPDATE, 0, 0, desc,
+                                       desc + 1, NULL);
+    ep->super.flags &= ~UCT_RC_EP_FLAG_FLUSH_REMOTE;
+
+    return UCS_INPROGRESS;
+}
+
 ucs_status_t uct_rc_mlx5_ep_flush(uct_ep_h tl_ep, unsigned flags,
                                   uct_completion_t *comp)
 {
     UCT_RC_MLX5_EP_DECL(tl_ep, iface, ep);
-    uct_ib_mlx5_md_t *md = ucs_derived_of(iface->super.super.super.md,
-                                          uct_ib_mlx5_md_t);
     int already_canceled = ep->super.flags & UCT_RC_EP_FLAG_FLUSH_CANCEL;
     ucs_status_t status;
-    uint16_t sn;
+
+    UCT_CHECK_PARAM(!ucs_test_all_flags(flags, UCT_FLUSH_FLAG_CANCEL |
+                                               UCT_FLUSH_FLAG_REMOTE),
+                    "flush flags CANCEL and REMOTE are mutually exclusive");
+
+    if (flags & UCT_FLUSH_FLAG_REMOTE) {
+        UCT_RC_IFACE_CHECK_FLUSH_REMOTE(
+                uct_ib_md_is_flush_rkey_valid(ep->super.flush_rkey), ep,
+                &iface->super, rcx);
+        if (ep->super.flags & UCT_RC_EP_FLAG_FLUSH_REMOTE) {
+            return uct_rc_mlx5_ep_flush_remote(tl_ep, comp);
+        }
+    }
 
     status = uct_rc_ep_flush(&ep->super, ep->tx.wq.bb_max, flags);
     if (status != UCS_INPROGRESS) {
@@ -577,28 +614,37 @@ ucs_status_t uct_rc_mlx5_ep_flush(uct_ep_h tl_ep, unsigned flags,
     }
 
     if (uct_rc_txqp_unsignaled(&ep->super.txqp) != 0) {
-        sn = ep->tx.wq.sw_pi;
         UCT_RC_CHECK_RES(&iface->super, &ep->super);
         uct_rc_mlx5_txqp_inline_post(iface, IBV_QPT_RC,
                                      &ep->super.txqp, &ep->tx.wq,
                                      MLX5_OPCODE_NOP, NULL, 0,
                                      0, 0, 0,
                                      0, 0,
-                                     NULL, NULL, 0, 0,
-                                     INT_MAX);
-    } else {
-        sn = ep->tx.wq.sig_pi;
+                                     0, 0,
+                                     0, INT_MAX);
     }
 
     if (ucs_unlikely((flags & UCT_FLUSH_FLAG_CANCEL) && !already_canceled)) {
-        status = uct_ib_mlx5_modify_qp_state(md, &ep->tx.wq.super, IBV_QPS_ERR);
+        status = uct_ib_mlx5_modify_qp_state(&iface->super.super,
+                                             &ep->tx.wq.super, IBV_QPS_ERR);
         if (status != UCS_OK) {
             return status;
         }
+
+        uct_ib_mlx5_txwq_update_flags(&ep->tx.wq, UCT_IB_MLX5_TXWQ_FLAG_FAILED,
+                                      0);
     }
 
     return uct_rc_txqp_add_flush_comp(&iface->super, &ep->super.super,
-                                      &ep->super.txqp, comp, sn);
+                                      &ep->super.txqp, comp, ep->tx.wq.sig_pi);
+}
+
+ucs_status_t uct_rc_mlx5_ep_invalidate(uct_ep_h tl_ep, unsigned flags)
+{
+    UCT_RC_MLX5_EP_DECL(tl_ep, iface, ep);
+
+    return uct_ib_mlx5_modify_qp_state(&iface->super.super, &ep->tx.wq.super,
+                                       IBV_QPS_ERR);
 }
 
 ucs_status_t uct_rc_mlx5_ep_fc_ctrl(uct_ep_t *tl_ep, unsigned op,
@@ -617,17 +663,19 @@ ucs_status_t uct_rc_mlx5_ep_fc_ctrl(uct_ep_t *tl_ep, unsigned op,
                                  NULL, 0,
                                  UCT_RC_EP_FC_PURE_GRANT, 0, 0,
                                  0, 0,
-                                 NULL, NULL, 0, 0,
-                                 INT_MAX);
+                                 0, 0,
+                                 0, INT_MAX);
     return UCS_OK;
 }
 
 ucs_status_t uct_rc_mlx5_ep_get_address(uct_ep_h tl_ep, uct_ep_addr_t *addr)
 {
     UCT_RC_MLX5_EP_DECL(tl_ep, iface, ep);
-    uct_rc_mlx5_ep_address_t *rc_addr = (uct_rc_mlx5_ep_address_t*)addr;
-    uct_ib_md_t *md                   = uct_ib_iface_md(ucs_derived_of(
-                                        tl_ep->iface, uct_ib_iface_t));
+    uct_rc_mlx5_ep_address_t *rc_addr       = (uct_rc_mlx5_ep_address_t*)addr;
+    uct_ib_md_t *md                         = uct_ib_iface_md(ucs_derived_of(
+                                              tl_ep->iface, uct_ib_iface_t));
+    uct_rc_mlx5_ep_ext_address_t *ext_addr;
+    void *ptr;
 
     uct_ib_pack_uint24(rc_addr->qp_num, ep->tx.wq.super.qp_num);
     uct_ib_mlx5_md_get_atomic_mr_id(md, &rc_addr->atomic_mr_id);
@@ -636,6 +684,14 @@ ucs_status_t uct_rc_mlx5_ep_get_address(uct_ep_h tl_ep, uct_ep_addr_t *addr)
         uct_ib_pack_uint24(rc_addr->tm_qp_num, ep->tm_qp.qp_num);
     }
 
+    if (uct_rc_mlx5_iface_flush_rkey_enabled(iface)) {
+        ext_addr                            = ucs_derived_of(rc_addr,
+                                                             uct_rc_mlx5_ep_ext_address_t);
+        ext_addr->flags                     = UCT_RC_MLX5_EP_ADDR_FLAG_FLUSH_RKEY;
+        ptr                                 = ext_addr + 1;
+        *ucs_serialize_next(&ptr, uint16_t) = md->flush_rkey >> 16;
+    }
+
     return UCS_OK;
 }
 
@@ -688,25 +744,31 @@ uct_rc_mlx5_ep_connect_qp(uct_rc_mlx5_iface_common_t *iface,
 
     ucs_assert(path_mtu != UCT_IB_ADDRESS_INVALID_PATH_MTU);
     if (md->flags & UCT_IB_MLX5_MD_FLAG_DEVX) {
-        return uct_rc_mlx5_iface_common_devx_connect_qp(iface, qp, qp_num,
-                                                        ah_attr, path_mtu,
-                                                        path_index);
+        return uct_rc_mlx5_iface_common_devx_connect_qp(
+                iface, qp, qp_num, ah_attr, path_mtu, path_index,
+                iface->super.config.max_rd_atomic);
     } else {
         return uct_rc_iface_qp_connect(&iface->super, qp->verbs.qp, qp_num,
                                        ah_attr, path_mtu);
     }
 }
 
-ucs_status_t uct_rc_mlx5_ep_connect_to_ep(uct_ep_h tl_ep,
-                                          const uct_device_addr_t *dev_addr,
-                                          const uct_ep_addr_t *ep_addr)
+ucs_status_t
+uct_rc_mlx5_ep_connect_to_ep_v2(uct_ep_h tl_ep,
+                                const uct_device_addr_t *device_addr,
+                                const uct_ep_addr_t *ep_addr,
+                                const uct_ep_connect_to_ep_params_t *params)
 {
     UCT_RC_MLX5_EP_DECL(tl_ep, iface, ep);
-    const uct_ib_address_t *ib_addr = (const uct_ib_address_t *)dev_addr;
+    const uct_ib_address_t *ib_addr         = (const uct_ib_address_t *)device_addr;
     const uct_rc_mlx5_ep_address_t *rc_addr = (const uct_rc_mlx5_ep_address_t*)ep_addr;
+    const uct_rc_mlx5_ep_ext_address_t *ext_addr;
+    size_t addr_length;
     uint32_t qp_num;
     struct ibv_ah_attr ah_attr;
     enum ibv_mtu path_mtu;
+    const void *ptr;
+    uint32_t flush_rkey_hi;
     ucs_status_t status;
 
     uct_ib_iface_fill_ah_attr_from_addr(&iface->super.super, ib_addr,
@@ -742,6 +804,25 @@ ucs_status_t uct_rc_mlx5_ep_connect_to_ep(uct_ep_h tl_ep,
 
     ep->super.atomic_mr_offset = uct_ib_md_atomic_offset(rc_addr->atomic_mr_id);
     ep->super.flags           |= UCT_RC_EP_FLAG_CONNECTED;
+    addr_length                = UCS_PARAM_VALUE(UCT_EP_CONNECT_TO_EP_PARAM_FIELD,
+                                                 params, ep_addr_length,
+                                                 EP_ADDR_LENGTH,
+                                                 sizeof(uct_rc_mlx5_ep_address_t));
+
+    if (addr_length <= sizeof(uct_rc_mlx5_ep_address_t)) {
+        ep->super.flush_rkey = UCT_IB_MD_INVALID_FLUSH_RKEY;
+        return UCS_OK;
+    }
+
+    ext_addr = ucs_derived_of(rc_addr, uct_rc_mlx5_ep_ext_address_t);
+    if (ext_addr->flags & UCT_RC_MLX5_EP_ADDR_FLAG_FLUSH_RKEY) {
+        ptr                  = ext_addr + 1;
+        flush_rkey_hi        = *ucs_serialize_next(&ptr, uint16_t);
+        ep->super.flush_rkey = (flush_rkey_hi << 16) |
+                               ((uint32_t)rc_addr->atomic_mr_id << 8);
+    } else {
+        ep->super.flush_rkey = UCT_IB_MD_INVALID_FLUSH_RKEY;
+    }
 
     return UCS_OK;
 }
@@ -769,8 +850,8 @@ uct_rc_mlx5_ep_tag_eager_short_inline(uct_ep_h tl_ep, uct_tag_t tag,
 
     uct_rc_mlx5_txqp_tag_inline_post(iface, IBV_QPT_RC, &ep->super.txqp,
                                      &ep->tx.wq, MLX5_OPCODE_SEND, data, length,
-                                     NULL, tag, 0, IBV_TMH_EAGER, 0, NULL,
-                                     NULL, 0, NULL, 0, MLX5_WQE_CTRL_SOLICITED);
+                                     NULL, tag, 0, IBV_TMH_EAGER, 0, 0, NULL, 0,
+                                     MLX5_WQE_CTRL_SOLICITED, 0);
 
     UCT_TL_EP_STAT_OP(&ep->super.super, TAG, SHORT, length);
 
@@ -801,7 +882,7 @@ ucs_status_t uct_rc_mlx5_ep_tag_eager_short(uct_ep_h tl_ep, uct_tag_t tag,
     status = uct_rc_mlx5_common_ep_short_dm(
             iface, IBV_QPT_RC, &cache, sizeof(cache.tm_hdr), data, length,
             MLX5_OPCODE_SEND, MLX5_WQE_CTRL_SOLICITED | MLX5_WQE_CTRL_CQ_UPDATE,
-            0, 0, &ep->super.txqp, &ep->tx.wq, NULL, NULL, 0);
+            0, 0, 0, &ep->super.txqp, &ep->tx.wq, 0);
     if (!UCS_STATUS_IS_ERR(status)) {
         UCT_TL_EP_STAT_OP(&ep->super.super, TAG, SHORT, length);
     }
@@ -832,8 +913,8 @@ ssize_t uct_rc_mlx5_ep_tag_eager_bcopy(uct_ep_h tl_ep, uct_tag_t tag,
     uct_rc_mlx5_common_txqp_bcopy_post(iface, IBV_QPT_RC, &ep->super.txqp,
                                        &ep->tx.wq, opcode,
                                        sizeof(struct ibv_tmh) + length, 0, 0,
-                                       NULL, NULL, 0, MLX5_WQE_CTRL_SOLICITED,
-                                       ib_imm, desc, desc + 1, NULL);
+                                       0, MLX5_WQE_CTRL_SOLICITED, 0, ib_imm,
+                                       desc, desc + 1, NULL);
 
     UCT_TL_EP_STAT_OP(&ep->super.super, TAG, BCOPY, length);
 
@@ -893,8 +974,8 @@ ucs_status_ptr_t uct_rc_mlx5_ep_tag_rndv_zcopy(uct_ep_h tl_ep, uct_tag_t tag,
     uct_rc_mlx5_txqp_tag_inline_post(iface, IBV_QPT_RC, &ep->super.txqp,
                                      &ep->tx.wq, MLX5_OPCODE_SEND, header,
                                      header_length, iov, tag, op_index,
-                                     IBV_TMH_RNDV, 0, NULL, NULL, 0,
-                                     NULL, 0, MLX5_WQE_CTRL_SOLICITED);
+                                     IBV_TMH_RNDV, 0, 0, NULL, 0,
+                                     MLX5_WQE_CTRL_SOLICITED, 0);
 
     return (ucs_status_ptr_t)((uint64_t)op_index);
 }
@@ -912,8 +993,8 @@ ucs_status_t uct_rc_mlx5_ep_tag_rndv_request(uct_ep_h tl_ep, uct_tag_t tag,
     uct_rc_mlx5_txqp_tag_inline_post(iface, IBV_QPT_RC, &ep->super.txqp,
                                      &ep->tx.wq, MLX5_OPCODE_SEND_IMM, header,
                                      header_length, NULL, tag, 0,
-                                     IBV_TMH_EAGER, 0, NULL, NULL, 0,
-                                     NULL, 0, MLX5_WQE_CTRL_SOLICITED);
+                                     IBV_TMH_EAGER, 0, 0, NULL, 0,
+                                     MLX5_WQE_CTRL_SOLICITED, 0);
     return UCS_OK;
 }
 #endif /* IBV_HW_TM */
@@ -924,13 +1005,12 @@ UCS_CLASS_INIT_FUNC(uct_rc_mlx5_ep_t, const uct_ep_params_t *params)
                                                        uct_rc_mlx5_iface_common_t);
     uct_ib_mlx5_md_t *md              = ucs_derived_of(iface->super.super.super.md,
                                                        uct_ib_mlx5_md_t);
-    uct_ib_mlx5_qp_attr_t attr = {};
+    uct_ib_mlx5_qp_attr_t attr        = {};
     ucs_status_t status;
 
     /* Need to create QP before super constructor to get QP number */
     uct_rc_mlx5_iface_fill_attr(iface, &attr, iface->super.config.tx_qp_len,
                                 &iface->rx.srq);
-    uct_ib_exp_qp_fill_attr(&iface->super.super, &attr.super);
     status = uct_rc_mlx5_iface_create_qp(iface, &self->tx.wq.super, &self->tx.wq, &attr);
     if (status != UCS_OK) {
         return status;
@@ -942,7 +1022,7 @@ UCS_CLASS_INIT_FUNC(uct_rc_mlx5_ep_t, const uct_ep_params_t *params)
     if (self->tx.wq.super.type == UCT_IB_MLX5_OBJ_TYPE_VERBS) {
         status = uct_rc_iface_qp_init(&iface->super, self->tx.wq.super.verbs.qp);
         if (status != UCS_OK) {
-            goto err;
+            goto err_destroy_txwq_qp;
         }
     }
 
@@ -950,37 +1030,47 @@ UCS_CLASS_INIT_FUNC(uct_rc_mlx5_ep_t, const uct_ep_params_t *params)
                                                 IBV_EVENT_QP_LAST_WQE_REACHED,
                                                 self->tx.wq.super.qp_num);
     if (status != UCS_OK) {
-        goto err;
+        goto err_destroy_txwq_qp;
     }
 
-    uct_rc_iface_add_qp(&iface->super, &self->super, self->tx.wq.super.qp_num);
+    status = uct_rc_iface_add_qp(&iface->super, &self->super,
+                                 self->tx.wq.super.qp_num);
+    if (status != UCS_OK) {
+        goto err_event_unreg;
+    }
 
     if (UCT_RC_MLX5_TM_ENABLED(iface)) {
         /* Send queue of this QP will be used by FW for HW RNDV. Driver requires
          * such a QP to be initialized with zero send queue length. */
         memset(&attr, 0, sizeof(attr));
         uct_rc_mlx5_iface_fill_attr(iface, &attr, 0, &iface->rx.srq);
-        uct_ib_exp_qp_fill_attr(&iface->super.super, &attr.super);
         status = uct_rc_mlx5_iface_create_qp(iface, &self->tm_qp, NULL, &attr);
         if (status != UCS_OK) {
-            goto err_unreg;
+            goto err_remove_txwq_qp;
         }
 
-        uct_rc_iface_add_qp(&iface->super, &self->super, self->tm_qp.qp_num);
+        status = uct_rc_iface_add_qp(&iface->super, &self->super,
+                                     self->tm_qp.qp_num);
+        if (status != UCS_OK) {
+            goto err_destroy_tm_qp;
+        }
     }
 
     self->tx.wq.bb_max = ucs_min(self->tx.wq.bb_max, iface->tx.bb_max);
     self->mp.free      = 1;
     uct_rc_txqp_available_set(&self->super.txqp, self->tx.wq.bb_max);
+    uct_rc_mlx5_iface_common_prepost_recvs(iface);
     return UCS_OK;
 
-err_unreg:
+err_destroy_tm_qp:
+    uct_ib_mlx5_destroy_qp(md, &self->tm_qp);
+err_remove_txwq_qp:
+    uct_rc_iface_remove_qp(&iface->super, self->tx.wq.super.qp_num);
+err_event_unreg:
     uct_ib_device_async_event_unregister(&md->super.dev,
                                          IBV_EVENT_QP_LAST_WQE_REACHED,
                                          self->tx.wq.super.qp_num);
-    uct_rc_iface_remove_qp(&iface->super, self->tx.wq.super.qp_num);
-
-err:
+err_destroy_txwq_qp:
     uct_ib_mlx5_destroy_qp(md, &self->tx.wq.super);
     return status;
 }
@@ -989,8 +1079,6 @@ UCS_CLASS_CLEANUP_FUNC(uct_rc_mlx5_ep_t)
 {
     uct_rc_mlx5_iface_common_t *iface = ucs_derived_of(self->super.super.super.iface,
                                                        uct_rc_mlx5_iface_common_t);
-    uct_ib_mlx5_md_t *md              = ucs_derived_of(iface->super.super.super.md,
-                                                       uct_ib_mlx5_md_t);
     uct_rc_mlx5_iface_qp_cleanup_ctx_t *cleanup_ctx;
     uint16_t outstanding, wqe_count;
 
@@ -1009,7 +1097,8 @@ UCS_CLASS_CLEANUP_FUNC(uct_rc_mlx5_ep_t)
 #endif
 
     ucs_assert(self->mp.free == 1);
-    (void)uct_ib_mlx5_modify_qp_state(md, &self->tx.wq.super, IBV_QPS_ERR);
+    (void)uct_ib_mlx5_modify_qp_state(&iface->super.super, &self->tx.wq.super,
+                                      IBV_QPS_ERR);
 
     /* Keep only one unreleased CQ credit per WQE, so we will not have CQ
        overflow. These CQ credits will be released by error CQE handler. */
