@@ -1,7 +1,6 @@
-/* -*- Mode: C; c-basic-offset:4 ; indent-tabs-mode:nil ; -*- */
 /*
- *  (C) 2014 by Argonne National Laboratory.
- *      See COPYRIGHT in top-level directory.
+ * Copyright (C) by Argonne National Laboratory
+ *     See COPYRIGHT in top-level directory
  */
 
 /* style:allow:snprintf:9 sig:0 */
@@ -338,7 +337,7 @@ const char *mpit_scopeToStr(int scope)
             p = "SCOPE_ALL_EQ";
             break;
         default:
-            p = "Unrecoginized scope";
+            p = "Unrecognized scope";
             break;
     }
     return p;
@@ -381,6 +380,9 @@ const char *mpit_bindingToStr(int binding)
         case MPI_T_BIND_MPI_INFO:
             p = "MPI_INFO";
             break;
+        case MPI_T_BIND_MPI_SESSION:
+            p = "MPI_SESSION";
+            break;
         default:
             p = "Unknown object binding";
     }
@@ -511,10 +513,10 @@ const char *mpit_errclasscheck(int err)
     const char *p = 0;
     switch (err) {
         case MPI_T_ERR_CVAR_SET_NOT_NOW:
-            p = "MPI_T cvar not set";
+            p = "MPI_T cvar cannot be set at this moment";
             break;
         case MPI_T_ERR_CVAR_SET_NEVER:
-            p = "MPI_T cvar was never set";
+            p = "MPI_T cvar cannot be set until end of execution";
             break;
         case MPI_T_ERR_PVAR_NO_STARTSTOP:
             p = "MPI_T pvar does not support start and stop";
@@ -537,9 +539,6 @@ const char *mpit_errclasscheck(int err)
         case MPI_T_ERR_INVALID_INDEX:
             p = "MPI_T index invalid";
             break;
-        case MPI_T_ERR_INVALID_ITEM:
-            p = "MPI_T item index out of range";
-            break;
         case MPI_T_ERR_INVALID_HANDLE:
             p = "MPI_T handle invalid";
             break;
