@@ -8,40 +8,6 @@
 
 #include "mpir_process.h"
 
-/*
-=== BEGIN_MPI_T_CVAR_INFO_BLOCK ===
-
-categories :
-    - name : REQUEST
-      description : A category for requests management variables
-
-cvars:
-    - name        : MPIR_CVAR_REQUEST_POLL_FREQ
-      category    : REQUEST
-      type        : int
-      default     : 8
-      class       : none
-      verbosity   : MPI_T_VERBOSITY_USER_BASIC
-      scope       : MPI_T_SCOPE_LOCAL
-      description : >-
-        How frequent to poll during completion calls (wait/test) in terms
-        of number of processed requests before polling.
-
-    - name        : MPIR_CVAR_REQUEST_BATCH_SIZE
-      category    : REQUEST
-      type        : int
-      default     : 64
-      class       : none
-      verbosity   : MPI_T_VERBOSITY_USER_BASIC
-      scope       : MPI_T_SCOPE_LOCAL
-      description : >-
-        The number of requests to make completion as a batch
-        in MPI_Waitall and MPI_Testall implementation. A large number
-        is likely to cause more cache misses.
-
-=== END_MPI_T_CVAR_INFO_BLOCK ===
-*/
-
 /* NOTE-R1: MPIR_REQUEST_KIND__MPROBE signifies that this is a request created by
  * MPI_Mprobe or MPI_Improbe.  Since we use MPI_Request objects as our
  * MPI_Message objects, we use this separate kind in order to provide stronger
@@ -65,6 +31,7 @@ typedef enum MPIR_Request_kind_t {
     MPIR_REQUEST_KIND__PART_SEND,       /* Partitioned send req returned to user */
     MPIR_REQUEST_KIND__PART_RECV,       /* Partitioned recv req returned to user */
     MPIR_REQUEST_KIND__PART,    /* Partitioned pt2pt internal reqs */
+    MPIR_REQUEST_KIND__ENQUEUE, /* enqueued (to gpu stream) request */
     MPIR_REQUEST_KIND__GREQUEST,
     MPIR_REQUEST_KIND__COLL,
     MPIR_REQUEST_KIND__MPROBE,  /* see NOTE-R1 */
@@ -86,8 +53,6 @@ typedef enum MPIR_Request_kind_t {
 
 #define MPIR_REQUEST_NULL_RECV     (MPI_Request)0x6c000010
 
-#define MPIR_REQUEST_BUILTIN_COUNT      0x11
-
 /* This currently defines a single structure type for all requests.
    Eventually, we may want a union type, as used in MPICH-1 */
 /* Typedefs for Fortran generalized requests */
@@ -131,6 +96,7 @@ typedef struct MPIR_Grequest_class {
 extern MPIR_Grequest_class MPIR_Grequest_class_direct[];
 extern MPIR_Object_alloc_t MPIR_Grequest_class_mem;
 
+#ifndef ENABLE_THREADCOMM
 #define MPIR_Request_extract_status(request_ptr_, status_)              \
     {                                                                   \
         if ((status_) != MPI_STATUS_IGNORE)                             \
@@ -147,6 +113,22 @@ extern MPIR_Object_alloc_t MPIR_Grequest_class_mem;
             (status_)->MPI_ERROR = error__;                             \
         }                                                               \
     }
+#else
+/* Same as above but with additional threadcomm tag reset */
+#define MPIR_Request_extract_status(request_ptr_, status_)              \
+    do {                                                                \
+        if ((status_) != MPI_STATUS_IGNORE) {                           \
+            int error__;                                                \
+            error__ = (status_)->MPI_ERROR;                             \
+            *(status_) = (request_ptr_)->status;                        \
+            (status_)->MPI_ERROR = error__;                             \
+                                                                        \
+            if ((request_ptr_)->comm && (request_ptr_)->comm->threadcomm) { \
+                MPIR_Threadcomm_adjust_status((request_ptr_)->comm->threadcomm, status_); \
+            } \
+        }                                                               \
+    } while (0)
+#endif
 
 #define MPIR_Request_is_complete(req_) (MPIR_cc_is_complete((req_)->cc_ptr))
 
@@ -172,6 +154,8 @@ enum MPIR_sched_type {
   (e.g., 'MPIR_Request_send_t') that extends the 'MPIR_Request'.
 
   S*/
+
+#define MPIR_REQUEST_UNION_SIZE  40
 struct MPIR_Request {
     MPIR_OBJECT_HEADER;         /* adds handle and ref_count fields */
 
@@ -185,12 +169,6 @@ struct MPIR_Request {
      * 32 bytes and 32-bit integers */
     MPIR_cc_t cc;
 
-    /* completion notification counter: this must be decremented by
-     * the request completion routine, when the completion count hits
-     * zero.  this counter allows us to keep track of the completion
-     * of multiple requests in a single place. */
-    MPIR_cc_t *completion_notification;
-
     /* A comm is needed to find the proper error handler */
     MPIR_Comm *comm;
     /* Status is needed for wait/test/recv */
@@ -204,19 +182,11 @@ struct MPIR_Request {
             MPIR_Errflag_t errflag;
             MPII_Coll_req_t coll;
         } nbc;                  /* kind : MPIR_REQUEST_KIND__COLL */
-#if defined HAVE_DEBUGGER_SUPPORT
         struct {
-            struct MPIR_Sendq *dbg_next;
-        } send;                 /* kind : MPID_REQUEST_SEND */
-#endif                          /* HAVE_DEBUGGER_SUPPORT */
-        struct {
-#if defined HAVE_DEBUGGER_SUPPORT
-            struct MPIR_Sendq *dbg_next;
-#endif                          /* HAVE_DEBUGGER_SUPPORT */
             /* Persistent requests have their own "real" requests */
             struct MPIR_Request *real_request;
             MPIR_TSP_sched_t sched;
-        } persist;              /* kind : MPID_PREQUEST_SEND or MPID_PREQUEST_RECV */
+        } persist;              /* kind : MPIR_REQUEST_KIND__PREQUEST_SEND or MPIR_REQUEST_KIND__PREQUEST_RECV */
         struct {
             struct MPIR_Request *real_request;
             enum MPIR_sched_type sched_type;
@@ -228,12 +198,32 @@ struct MPIR_Request {
             MPL_atomic_int_t active_flag;       /* flag indicating whether in a start-complete active period.
                                                  * Value is 0 or 1. */
         } part;                 /* kind : MPIR_REQUEST_KIND__PART_SEND or MPIR_REQUEST_KIND__PART_RECV */
+        struct {
+            MPIR_Stream *stream_ptr;
+            struct MPIR_Request *real_request;
+            bool is_send;
+            void *data;
+        } enqueue;
         struct {
             MPIR_Win *win;
         } rma;                  /* kind : MPIR_REQUEST_KIND__RMA */
+        /* Reserve space for local usages. For example, threadcomm, the actual struct
+         * is defined locally and is used via casting */
+        char dummy[MPIR_REQUEST_UNION_SIZE];
     } u;
 
+#if defined HAVE_DEBUGGER_SUPPORT
+    struct MPIR_Debugq *send;
+    struct MPIR_Debugq *recv;
+    struct MPIR_Debugq *unexp;
+#endif                          /* HAVE_DEBUGGER_SUPPORT */
+
     struct MPIR_Request *next, *prev;
+    UT_hash_handle hh;
+
+#ifdef MPICH_DEBUG_PROGRESS
+    char info[100];
+#endif
 
     /* Other, device-specific information */
 #ifdef MPID_DEV_REQUEST_DECL
@@ -253,40 +243,48 @@ void MPIR_Persist_coll_free_cb(MPIR_Request * request);
  * MPIR_Request_create_from_pool is used to create request objects from a specific pool.
  * MPIR_Request_create is a wrapper to create request from pool 0.
  */
-/* Handle Bits - 2+4+6+8+12 - Type, Kind, Pool_idx, Block_idx, Object_idx */
-#define REQUEST_POOL_MASK    0x03f00000
-#define REQUEST_POOL_SHIFT   20
+/* Handle Bits - 2+4+1+6+7+12 - Type, Kind, NonZeroVCI, Pool_idx, Block_idx, Object_idx */
+/* UPDATE: If the 6th bit (NonZeroVCI) is 0, it allocated from Pool 0, Block_idx extends to 13 bits (taking over Pool_idx).
+ *         If the 6th bit is 1, BLOCK_idx shrinks to 7 bits.
+ */
+#define REQUEST_POOL_FLAG    0x02000000
+#define REQUEST_POOL_MASK    0x01f80000
+#define REQUEST_POOL_SHIFT   19
 #define REQUEST_POOL_MAX     64
-#define REQUEST_BLOCK_MASK   0x000ff000
+#define REQUEST_BLOCK_MASK0  0x01fff000
+#define REQUEST_BLOCK_MASK   0x0007f000
 #define REQUEST_BLOCK_SHIFT  12
-#define REQUEST_BLOCK_MAX    256
+#define REQUEST_BLOCK_MAX    128
 #define REQUEST_OBJECT_MASK  0x00000fff
 #define REQUEST_OBJECT_SHIFT 0
 #define REQUEST_OBJECT_MAX   4096
 
-#define REQUEST_NUM_BLOCKS   256
-#define REQUEST_NUM_INDICES  1024
+#define REQUEST_NUM_BLOCKS0  8192
+#define REQUEST_NUM_BLOCKS   128
+#define REQUEST_NUM_INDICES  4096
 
+#define MPIR_MAX_VCIS REQUEST_POOL_MAX
 #define MPIR_REQUEST_NUM_POOLS REQUEST_POOL_MAX
-#define MPIR_REQUEST_PREALLOC 8
 
-#define MPIR_REQUEST_POOL(req_) (((req_)->handle & REQUEST_POOL_MASK) >> REQUEST_POOL_SHIFT)
+#define MPIR_REQUEST_POOL_FROM_HANDLE(handle) (((handle) & REQUEST_POOL_FLAG) ? \
+    (((handle) & REQUEST_POOL_MASK) >> REQUEST_POOL_SHIFT) : 0)
+#define MPIR_REQUEST_POOL(req_) MPIR_REQUEST_POOL_FROM_HANDLE((req_)->handle)
 
-extern MPIR_Request MPIR_Request_builtins[MPIR_REQUEST_BUILTIN_COUNT];
+extern MPIR_Request MPIR_Request_builtin[MPIR_REQUEST_N_BUILTIN];
 extern MPIR_Object_alloc_t MPIR_Request_mem[MPIR_REQUEST_NUM_POOLS];
 extern MPIR_Request MPIR_Request_direct[MPIR_REQUEST_PREALLOC];
 
 #define MPIR_Request_get_ptr(a, ptr) \
     do { \
         int pool, blk, idx; \
-        pool = ((a) & REQUEST_POOL_MASK) >> REQUEST_POOL_SHIFT; \
+        pool = MPIR_REQUEST_POOL_FROM_HANDLE(a); \
         switch (HANDLE_GET_KIND(a)) { \
         case HANDLE_KIND_BUILTIN: \
             if (a == MPI_MESSAGE_NO_PROC) { \
                 ptr = NULL; \
             } else { \
-                MPIR_Assert(HANDLE_INDEX(a) < MPIR_REQUEST_BUILTIN_COUNT); \
-                ptr = MPIR_Request_builtins + HANDLE_INDEX(a); \
+                MPIR_Assert(HANDLE_INDEX(a) < MPIR_REQUEST_N_BUILTIN); \
+                ptr = MPIR_Request_builtin + HANDLE_INDEX(a); \
             } \
             break; \
         case HANDLE_KIND_DIRECT: \
@@ -294,7 +292,11 @@ extern MPIR_Request MPIR_Request_direct[MPIR_REQUEST_PREALLOC];
             ptr = MPIR_Request_direct + HANDLE_INDEX(a); \
             break; \
         case HANDLE_KIND_INDIRECT: \
-            blk = ((a) & REQUEST_BLOCK_MASK) >> REQUEST_BLOCK_SHIFT; \
+            if (pool == 0) { \
+                blk = ((a) & REQUEST_BLOCK_MASK0) >> REQUEST_BLOCK_SHIFT; \
+            } else { \
+                blk = ((a) & REQUEST_BLOCK_MASK) >> REQUEST_BLOCK_SHIFT; \
+            } \
             idx = ((a) & REQUEST_OBJECT_MASK) >> REQUEST_OBJECT_SHIFT; \
             ptr = ((MPIR_Request *) MPIR_Request_mem[pool].indirect[blk]) + idx; \
             break; \
@@ -304,6 +306,54 @@ extern MPIR_Request MPIR_Request_direct[MPIR_REQUEST_PREALLOC];
         } \
     } while (0)
 
+#ifdef MPICH_DEBUG_PROGRESS
+#define MPIR_REQUEST_SET_INFO(req, ...) \
+    do { \
+        MPL_snprintf_nowarn((req)->info, 100, __VA_ARGS__); \
+    } while (0)
+
+#define MPIR_REQUEST_DEBUG(req) \
+    do { \
+        if (MPIR_cc_get((req)->cc) > 0) { \
+            printf("    %x: %s\n", (req)->handle, (req)->info); \
+        } \
+    } while (0)
+
+#define DEBUG_PROGRESS_START \
+    int iter = 0; \
+    bool progress_timed_out = false; \
+    MPL_time_t time_start; \
+    if (MPIR_CVAR_DEBUG_PROGRESS_TIMEOUT > 0) { \
+        MPL_wtime(&time_start); \
+    }
+
+#define DEBUG_PROGRESS_CHECK \
+    if (MPIR_CVAR_DEBUG_PROGRESS_TIMEOUT > 0) { \
+        iter++; \
+        if (iter == 0xffff) {\
+            double time_diff = 0.0; \
+            MPL_time_t time_cur; \
+            MPL_wtime(&time_cur); \
+            MPL_wtime_diff(&time_start, &time_cur, &time_diff); \
+            if (time_diff > MPIR_CVAR_DEBUG_PROGRESS_TIMEOUT && !progress_timed_out) { \
+                MPIR_Request_debug(); \
+                MPL_backtrace_show(stdout); \
+                progress_timed_out = true; \
+            } else if (time_diff > MPIR_CVAR_DEBUG_PROGRESS_TIMEOUT * 2) { \
+                MPIR_ERR_SETANDJUMP(mpi_errno, MPI_ERR_OTHER, "**timeout"); \
+            } \
+            iter = 0; \
+        } \
+    }
+
+#else
+
+#define MPIR_REQUEST_SET_INFO(req, ...) do { } while (0)
+#define MPIR_REQUEST_DEBUG(req) do { } while (0)
+#define DEBUG_PROGRESS_START do {} while (0)
+#define DEBUG_PROGRESS_CHECK do {} while (0)
+#endif
+
 void MPII_init_request(void);
 
 /* To get the benefit of multiple request pool, device layer need register their per-vci lock
@@ -319,12 +369,8 @@ static inline int MPIR_Request_is_persistent(MPIR_Request * req_ptr)
 {
     return (req_ptr->kind == MPIR_REQUEST_KIND__PREQUEST_SEND ||
             req_ptr->kind == MPIR_REQUEST_KIND__PREQUEST_RECV ||
-            req_ptr->kind == MPIR_REQUEST_KIND__PREQUEST_COLL);
-}
-
-static inline int MPIR_Request_is_partitioned(MPIR_Request * req_ptr)
-{
-    return (req_ptr->kind == MPIR_REQUEST_KIND__PART_SEND ||
+            req_ptr->kind == MPIR_REQUEST_KIND__PREQUEST_COLL ||
+            req_ptr->kind == MPIR_REQUEST_KIND__PART_SEND ||
             req_ptr->kind == MPIR_REQUEST_KIND__PART_RECV);
 }
 
@@ -356,6 +402,8 @@ static inline int MPIR_Request_is_active(MPIR_Request * req_ptr)
             case MPIR_REQUEST_KIND__PREQUEST_SEND:
             case MPIR_REQUEST_KIND__PREQUEST_RECV:
                 return (req_ptr)->u.persist.real_request != NULL;
+            case MPIR_REQUEST_KIND__PREQUEST_COLL:
+                return (req_ptr)->u.persist_coll.real_request != NULL;
             case MPIR_REQUEST_KIND__PART_SEND:
             case MPIR_REQUEST_KIND__PART_RECV:
                 return MPIR_Part_request_is_active(req_ptr);
@@ -383,13 +431,15 @@ static inline MPIR_Request *MPIR_Request_create_from_pool(MPIR_Request_kind_t ki
 #ifdef MPICH_DEBUG_MUTEX
     MPID_THREAD_ASSERT_IN_CS(VCI, (*(MPID_Thread_mutex_t *) MPIR_Request_mem[pool].lock));
 #endif
-    req = MPIR_Handle_obj_alloc_unsafe(&MPIR_Request_mem[pool],
-                                       REQUEST_NUM_BLOCKS, REQUEST_NUM_INDICES);
+    int max_blocks = (pool == 0) ? REQUEST_NUM_BLOCKS0 : REQUEST_NUM_BLOCKS;
+    req = MPIR_Handle_obj_alloc_unsafe(&MPIR_Request_mem[pool], max_blocks, REQUEST_NUM_INDICES);
     if (req == NULL)
         goto fn_fail;
 
     /* Patch the handle for pool index. */
-    req->handle |= (pool << REQUEST_POOL_SHIFT);
+    if (pool > 0) {
+        req->handle |= (REQUEST_POOL_FLAG | (pool << REQUEST_POOL_SHIFT));
+    }
 
     MPL_DBG_MSG_P(MPIR_DBG_REQUEST, VERBOSE, "allocated request, handle=0x%08x", req->handle);
 #ifdef MPICH_DBG_OUTPUT
@@ -412,8 +462,9 @@ static inline MPIR_Request *MPIR_Request_create_from_pool(MPIR_Request_kind_t ki
     req->kind = kind;
     MPIR_cc_set(&req->cc, 1);
     req->cc_ptr = &req->cc;
-
-    req->completion_notification = NULL;
+#ifdef MPICH_DEBUG_PROGRESS
+    req->info[0] = '\0';
+#endif
 
     req->status.MPI_ERROR = MPI_SUCCESS;
     MPIR_STATUS_SET_CANCEL_BIT(req->status, FALSE);
@@ -421,9 +472,6 @@ static inline MPIR_Request *MPIR_Request_create_from_pool(MPIR_Request_kind_t ki
     req->comm = NULL;
 
     switch (kind) {
-        case MPIR_REQUEST_KIND__SEND:
-            MPII_REQUEST_CLEAR_DBG(req);
-            break;
         case MPIR_REQUEST_KIND__COLL:
             req->u.nbc.errflag = MPIR_ERR_NONE;
             req->u.nbc.coll.host_sendbuf = NULL;
@@ -437,6 +485,7 @@ static inline MPIR_Request *MPIR_Request_create_from_pool(MPIR_Request_kind_t ki
         default:
             break;
     }
+    MPII_REQUEST_CLEAR_DBG(req);
 
     MPID_Request_create_hook(req);
 
@@ -465,14 +514,14 @@ static inline MPIR_Request *MPIR_Request_create_from_pool_safe(MPIR_Request_kind
 static inline MPIR_Request *MPIR_Request_create(MPIR_Request_kind_t kind)
 {
     MPIR_Request *req;
-    MPID_THREAD_CS_ENTER(POBJ, MPIR_THREAD_POBJ_HANDLE_MUTEX);
     MPID_THREAD_CS_ENTER(VCI, (*(MPID_Thread_mutex_t *) MPIR_Request_mem[0].lock));
     req = MPIR_Request_create_from_pool(kind, 0, 1);
-    MPID_THREAD_CS_EXIT(POBJ, MPIR_THREAD_POBJ_HANDLE_MUTEX);
     MPID_THREAD_CS_EXIT(VCI, (*(MPID_Thread_mutex_t *) MPIR_Request_mem[0].lock));
     return req;
 }
 
+int MPIR_allocate_enqueue_request(MPIR_Comm * comm_ptr, MPIR_Request ** req);
+
 #define MPIR_Request_add_ref(req_p_) \
     do { MPIR_Object_add_ref(req_p_); } while (0)
 
@@ -481,7 +530,7 @@ static inline MPIR_Request *MPIR_Request_create(MPIR_Request_kind_t kind)
 
 MPL_STATIC_INLINE_PREFIX MPIR_Request *get_builtin_req(int idx, MPIR_Request_kind_t kind)
 {
-    return MPIR_Request_builtins + (idx);
+    return MPIR_Request_builtin + (idx);
 }
 
 MPL_STATIC_INLINE_PREFIX MPIR_Request *MPIR_Request_create_complete(MPIR_Request_kind_t kind)
@@ -495,9 +544,10 @@ MPL_STATIC_INLINE_PREFIX MPIR_Request *MPIR_Request_create_null_recv(void)
     return get_builtin_req(HANDLE_INDEX(MPIR_REQUEST_NULL_RECV), MPIR_REQUEST_KIND__RECV);
 }
 
-static inline void MPIR_Request_free_with_safety(MPIR_Request * req, int need_safety)
+int MPIR_Grequest_free(MPIR_Request * request_ptr);
+static inline void MPIR_Request_free_with_safety(MPIR_Request * req,
+                                                 int need_safety, int *errno_out)
 {
-    int inuse;
     int pool = MPIR_REQUEST_POOL(req);
 
     if (HANDLE_IS_BUILTIN(req->handle)) {
@@ -505,8 +555,6 @@ static inline void MPIR_Request_free_with_safety(MPIR_Request * req, int need_sa
         return;
     }
 
-    MPIR_Request_release_ref(req, &inuse);
-
     if (need_safety) {
         MPID_THREAD_CS_ENTER(VCI, (*(MPID_Thread_mutex_t *) MPIR_Request_mem[pool].lock));
     }
@@ -517,6 +565,8 @@ static inline void MPIR_Request_free_with_safety(MPIR_Request * req, int need_sa
      * this request */
     MPID_Request_free_hook(req);
 
+    int inuse;
+    MPIR_Request_release_ref(req, &inuse);
     if (inuse == 0) {
         MPL_DBG_MSG_P(MPIR_DBG_REQUEST, VERBOSE, "freeing request, handle=0x%08x", req->handle);
 
@@ -538,28 +588,54 @@ static inline void MPIR_Request_free_with_safety(MPIR_Request * req, int need_sa
         }
 #endif
 
+        int mpi_errno;          /* for Grequest free_fn return */
+        switch (req->kind) {
+            case MPIR_REQUEST_KIND__SEND:
+                MPII_SENDQ_FORGET(req);
+                break;
+            case MPIR_REQUEST_KIND__RECV:
+                MPII_RECVQ_FORGET(req);
+                break;
+            case MPIR_REQUEST_KIND__PREQUEST_SEND:
+            case MPIR_REQUEST_KIND__PREQUEST_RECV:
+                MPID_Prequest_free_hook(req);
+                break;
+            case MPIR_REQUEST_KIND__PREQUEST_COLL:
+                MPIR_Persist_coll_free_cb(req);
+                break;
+            case MPIR_REQUEST_KIND__PART_SEND:
+            case MPIR_REQUEST_KIND__PART_RECV:
+                MPID_Part_request_free_hook(req);
+                break;
+            case MPIR_REQUEST_KIND__GREQUEST:
+                mpi_errno = MPIR_Grequest_free(req);
+                MPL_free(req->u.ureq.greq_fns);
+                if (errno_out) {
+                    *errno_out = mpi_errno;
+                }
+                break;
+            default:
+                break;
+        }
+
         /* FIXME: We need a better way to handle these so that we do
          * not always need to initialize these fields and check them
          * when we destroy a request */
         /* FIXME: We need a way to call these routines ONLY when the
          * related ref count has become zero. */
         if (req->comm != NULL) {
+            if (MPIR_Request_is_persistent(req)) {
+                MPIR_Comm_delete_inactive_request(req->comm, req);
+            }
             MPIR_Comm_release(req->comm);
         }
 
-        if (req->kind == MPIR_REQUEST_KIND__GREQUEST) {
-            MPL_free(req->u.ureq.greq_fns);
-        }
-
         MPID_Request_destroy_hook(req);
+#ifdef MPICH_DEBUG_PROGRESS
+        MPIR_cc_set(&req->cc, 0);
+#endif
 
-        if (need_safety) {
-            MPID_THREAD_CS_ENTER(POBJ, MPIR_THREAD_POBJ_HANDLE_MUTEX);
-            MPIR_Handle_obj_free_unsafe(&MPIR_Request_mem[pool], req, /* not info */ FALSE);
-            MPID_THREAD_CS_EXIT(POBJ, MPIR_THREAD_POBJ_HANDLE_MUTEX);
-        } else {
-            MPIR_Handle_obj_free_unsafe(&MPIR_Request_mem[pool], req, /* not info */ FALSE);
-        }
+        MPIR_Handle_obj_free_unsafe(&MPIR_Request_mem[pool], req, /* not info */ FALSE);
     }
     if (need_safety) {
         MPID_THREAD_CS_EXIT(VCI, (*(MPID_Thread_mutex_t *) MPIR_Request_mem[pool].lock));
@@ -568,18 +644,26 @@ static inline void MPIR_Request_free_with_safety(MPIR_Request * req, int need_sa
 
 MPL_STATIC_INLINE_PREFIX void MPIR_Request_free_safe(MPIR_Request * req)
 {
-    MPIR_Request_free_with_safety(req, 1);
+    MPIR_Request_free_with_safety(req, 1, NULL);
 }
 
 MPL_STATIC_INLINE_PREFIX void MPIR_Request_free_unsafe(MPIR_Request * req)
 {
-    MPIR_Request_free_with_safety(req, 0);
+    MPIR_Request_free_with_safety(req, 0, NULL);
 }
 
 MPL_STATIC_INLINE_PREFIX void MPIR_Request_free(MPIR_Request * req)
 {
     /* The default is to assume we need safety unless it's global thread granularity */
-    MPIR_Request_free_with_safety(req, 1);
+    MPIR_Request_free_with_safety(req, 1, NULL);
+}
+
+MPL_STATIC_INLINE_PREFIX int MPIR_Request_free_return(MPIR_Request * req)
+{
+    /* Same as MPIR_Request_free, bur return mpi_errno. It is relevant for Grequest free_fn */
+    int mpi_errno = MPI_SUCCESS;
+    MPIR_Request_free_with_safety(req, 1, &mpi_errno);
+    return mpi_errno;
 }
 
 /* Requests that are not created inside device (general requests, nonblocking collective
@@ -608,11 +692,6 @@ MPL_STATIC_INLINE_PREFIX int MPIR_Request_completion_processing_fastpath(MPI_Req
     MPIR_Assert(request_ptr->kind == MPIR_REQUEST_KIND__SEND ||
                 request_ptr->kind == MPIR_REQUEST_KIND__RECV);
 
-    if (request_ptr->kind == MPIR_REQUEST_KIND__SEND) {
-        /* FIXME: are Ibsend requests added to the send queue? */
-        MPII_SENDQ_FORGET(request_ptr);
-    }
-
     /* the completion path for SEND and RECV is the same at this time, modulo
      * the SENDQ hook above */
     mpi_errno = request_ptr->status.MPI_ERROR;
@@ -723,19 +802,20 @@ int MPIR_Waitany_impl(int count, MPIR_Request * request_ptrs[], int *indx, MPI_S
 int MPIR_Waitsome_impl(int incount, MPIR_Request * request_ptrs[],
                        int *outcount, int array_of_indices[], MPI_Status array_of_statuses[]);
 
-int MPIR_Test(MPI_Request * request, int *flag, MPI_Status * status);
-int MPIR_Testall(int count, MPI_Request array_of_requests[], int *flag,
+int MPIR_Test(MPIR_Request * request_ptr, int *flag, MPI_Status * status);
+int MPIR_Testall(int count, MPIR_Request * request_ptrs[], int *flag,
                  MPI_Status array_of_statuses[]);
-int MPIR_Testany(int count, MPI_Request array_of_requests[], MPIR_Request * request_ptrs[],
+int MPIR_Testany(int count, MPIR_Request * request_ptrs[],
                  int *indx, int *flag, MPI_Status * status);
-int MPIR_Testsome(int incount, MPI_Request array_of_requests[], MPIR_Request * request_ptrs[],
+int MPIR_Testsome(int incount, MPIR_Request * request_ptrs[],
                   int *outcount, int array_of_indices[], MPI_Status array_of_statuses[]);
-int MPIR_Wait(MPI_Request * request, MPI_Status * status);
-int MPIR_Waitall(int count, MPI_Request array_of_requests[], MPI_Status array_of_statuses[]);
-int MPIR_Waitany(int count, MPI_Request array_of_requests[], MPIR_Request * request_ptrs[],
-                 int *indx, MPI_Status * status);
-int MPIR_Waitsome(int incount, MPI_Request array_of_requests[], MPIR_Request * request_ptrs[],
+int MPIR_Wait(MPIR_Request * request_ptr, MPI_Status * status);
+int MPIR_Waitall(int count, MPIR_Request * request_ptrs[], MPI_Status array_of_statuses[]);
+int MPIR_Waitany(int count, MPIR_Request * request_ptrs[], int *indx, MPI_Status * status);
+int MPIR_Waitsome(int incount, MPIR_Request * request_ptrs[],
                   int *outcount, int array_of_indices[], MPI_Status array_of_statuses[]);
 int MPIR_Parrived(MPIR_Request * request_ptr, int partition, int *flag);
 
+void MPIR_Request_debug(void);
+
 #endif /* MPIR_REQUEST_H_INCLUDED */
