@@ -5,62 +5,20 @@
 
 #include "mpiimpl.h"
 
-/* -- Begin Profiling Symbol Block for routine MPI_Type_create_darray */
-#if defined(HAVE_PRAGMA_WEAK)
-#pragma weak MPI_Type_create_darray = PMPI_Type_create_darray
-#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
-#pragma _HP_SECONDARY_DEF PMPI_Type_create_darray  MPI_Type_create_darray
-#elif defined(HAVE_PRAGMA_CRI_DUP)
-#pragma _CRI duplicate MPI_Type_create_darray as PMPI_Type_create_darray
-#elif defined(HAVE_WEAK_ATTRIBUTE)
-int MPI_Type_create_darray(int size, int rank, int ndims, const int array_of_gsizes[],
-                           const int array_of_distribs[], const int array_of_dargs[],
-                           const int array_of_psizes[], int order, MPI_Datatype oldtype,
-                           MPI_Datatype * newtype)
-    __attribute__ ((weak, alias("PMPI_Type_create_darray")));
-#endif
-/* -- End Profiling Symbol Block */
-
-PMPI_LOCAL int MPIR_Type_block(const int *array_of_gsizes,
-                               int dim,
-                               int ndims,
-                               int nprocs,
-                               int rank,
-                               int darg,
-                               int order,
-                               MPI_Aint orig_extent,
-                               MPI_Datatype type_old,
-                               MPI_Datatype * type_new, MPI_Aint * st_offset);
-PMPI_LOCAL int MPIR_Type_cyclic(const int *array_of_gsizes,
-                                int dim,
-                                int ndims,
-                                int nprocs,
-                                int rank,
-                                int darg,
-                                int order,
-                                MPI_Aint orig_extent,
-                                MPI_Datatype type_old,
-                                MPI_Datatype * type_new, MPI_Aint * st_offset);
-
-/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build
-   the MPI routines */
-#ifndef MPICH_MPI_FROM_PMPI
-#undef MPI_Type_create_darray
-#define MPI_Type_create_darray PMPI_Type_create_darray
-
-PMPI_LOCAL int MPIR_Type_block(const int *array_of_gsizes,
-                               int dim,
-                               int ndims,
-                               int nprocs,
-                               int rank,
-                               int darg,
-                               int order,
-                               MPI_Aint orig_extent,
-                               MPI_Datatype type_old, MPI_Datatype * type_new, MPI_Aint * st_offset)
+static int MPIR_Type_block(const MPI_Aint * array_of_gsizes,
+                           int dim,
+                           int ndims,
+                           int nprocs,
+                           int rank,
+                           int darg,
+                           int order,
+                           MPI_Aint orig_extent,
+                           MPI_Datatype type_old, MPI_Datatype * type_new, MPI_Aint * st_offset)
 {
 /* nprocs = no. of processes in dimension dim of grid
    rank = coordinate of this process in dimension dim */
-    int mpi_errno, blksize, global_size, mysize, i, j;
+    int mpi_errno;
+    MPI_Aint global_size, blksize, mysize;
     MPI_Aint stride;
 
     global_size = array_of_gsizes[dim];
@@ -93,8 +51,7 @@ PMPI_LOCAL int MPIR_Type_block(const int *array_of_gsizes,
 #endif
     }
 
-    j = global_size - blksize * rank;
-    mysize = MPL_MIN(blksize, j);
+    mysize = MPL_MIN(blksize, global_size - blksize * rank);
     if (mysize < 0)
         mysize = 0;
 
@@ -111,8 +68,8 @@ PMPI_LOCAL int MPIR_Type_block(const int *array_of_gsizes,
             }
             /* --END ERROR HANDLING-- */
         } else {
-            for (i = 0; i < dim; i++)
-                stride *= (MPI_Aint) (array_of_gsizes[i]);
+            for (int i = 0; i < dim; i++)
+                stride *= array_of_gsizes[i];
             mpi_errno = MPIR_Type_vector(mysize, 1, stride, 1,  /* stride in bytes */
                                          type_old, type_new);
             /* --BEGIN ERROR HANDLING-- */
@@ -136,8 +93,8 @@ PMPI_LOCAL int MPIR_Type_block(const int *array_of_gsizes,
             }
             /* --END ERROR HANDLING-- */
         } else {
-            for (i = ndims - 1; i > dim; i--)
-                stride *= (MPI_Aint) (array_of_gsizes[i]);
+            for (int i = ndims - 1; i > dim; i--)
+                stride *= array_of_gsizes[i];
             mpi_errno = MPIR_Type_vector(mysize, 1, stride, 1,  /* stride in bytes */
                                          type_old, type_new);
             /* --BEGIN ERROR HANDLING-- */
@@ -151,30 +108,37 @@ PMPI_LOCAL int MPIR_Type_block(const int *array_of_gsizes,
         }
     }
 
-    *st_offset = (MPI_Aint) blksize *(MPI_Aint) rank;
+    *st_offset = blksize * rank;
     /* in terms of no. of elements of type oldtype in this dimension */
     if (mysize == 0)
         *st_offset = 0;
 
+    MPI_Datatype type_tmp;
+    MPI_Aint ex;
+    MPIR_Datatype_get_extent_macro(type_old, ex);
+    mpi_errno = MPIR_Type_create_resized(*type_new, 0, array_of_gsizes[dim] * ex, &type_tmp);
+    MPIR_Type_free_impl(type_new);
+    *type_new = type_tmp;
+
     return MPI_SUCCESS;
 }
 
-PMPI_LOCAL int MPIR_Type_cyclic(const int *array_of_gsizes,
-                                int dim,
-                                int ndims,
-                                int nprocs,
-                                int rank,
-                                int darg,
-                                int order,
-                                MPI_Aint orig_extent,
-                                MPI_Datatype type_old,
-                                MPI_Datatype * type_new, MPI_Aint * st_offset)
+static int MPIR_Type_cyclic(const MPI_Aint * array_of_gsizes,
+                            int dim,
+                            int ndims,
+                            int nprocs,
+                            int rank,
+                            int darg,
+                            int order,
+                            MPI_Aint orig_extent,
+                            MPI_Datatype type_old, MPI_Datatype * type_new, MPI_Aint * st_offset)
 {
 /* nprocs = no. of processes in dimension dim of grid
    rank = coordinate of this process in dimension dim */
-    int mpi_errno, blksize, i, blklens[3], st_index, end_index, local_size, rem, count;
-    MPI_Aint stride, disps[3];
-    MPI_Datatype type_tmp, type_indexed, types[3];
+    int mpi_errno;
+    MPI_Aint blksize, st_index, end_index, local_size, rem, count;
+    MPI_Aint stride;
+    MPI_Datatype type_tmp, type_indexed;
 
     if (darg == MPI_DISTRIBUTE_DFLT_DARG)
         blksize = 1;
@@ -209,11 +173,11 @@ PMPI_LOCAL int MPIR_Type_cyclic(const int *array_of_gsizes,
 
     stride = (MPI_Aint) nprocs *(MPI_Aint) blksize *orig_extent;
     if (order == MPI_ORDER_FORTRAN)
-        for (i = 0; i < dim; i++)
-            stride *= (MPI_Aint) (array_of_gsizes[i]);
+        for (int i = 0; i < dim; i++)
+            stride *= array_of_gsizes[i];
     else
-        for (i = ndims - 1; i > dim; i--)
-            stride *= (MPI_Aint) (array_of_gsizes[i]);
+        for (int i = ndims - 1; i > dim; i--)
+            stride *= array_of_gsizes[i];
 
     mpi_errno = MPIR_Type_vector(count, blksize, stride, 1,     /* stride in bytes */
                                  type_old, type_new);
@@ -229,6 +193,9 @@ PMPI_LOCAL int MPIR_Type_cyclic(const int *array_of_gsizes,
     if (rem) {
         /* if the last block is of size less than blksize, include
          * it separately using MPI_Type_struct */
+        MPI_Aint blklens[3];
+        MPI_Aint disps[3];
+        MPI_Datatype types[3];
 
         types[0] = *type_new;
         types[1] = type_old;
@@ -255,9 +222,10 @@ PMPI_LOCAL int MPIR_Type_cyclic(const int *array_of_gsizes,
      * dimension correctly. */
     if (((order == MPI_ORDER_FORTRAN) && (dim == 0)) ||
         ((order == MPI_ORDER_C) && (dim == ndims - 1))) {
+        MPI_Aint disps[3];
         disps[0] = 0;
         disps[1] = (MPI_Aint) rank *(MPI_Aint) blksize *orig_extent;
-        disps[2] = orig_extent * (MPI_Aint) (array_of_gsizes[dim]);
+        disps[2] = orig_extent * array_of_gsizes[dim];
 
         mpi_errno = MPIR_Type_blockindexed(1, 1, &disps[1], 1,  /* 1 means disp is in bytes */
                                            *type_new, &type_indexed);
@@ -297,193 +265,38 @@ PMPI_LOCAL int MPIR_Type_cyclic(const int *array_of_gsizes,
     if (local_size == 0)
         *st_offset = 0;
 
+    MPI_Aint ex;
+    MPIR_Datatype_get_extent_macro(type_old, ex);
+    mpi_errno = MPIR_Type_create_resized(*type_new, 0, array_of_gsizes[dim] * ex, &type_tmp);
+    MPIR_Type_free_impl(type_new);
+    *type_new = type_tmp;
+
     return MPI_SUCCESS;
 }
-#endif
 
-/*@
-   MPI_Type_create_darray - Create a datatype representing a distributed array
-
-Input Parameters:
-+ size - size of process group (positive integer)
-. rank - rank in process group (nonnegative integer)
-. ndims - number of array dimensions as well as process grid dimensions (positive integer)
-. array_of_gsizes - number of elements of type oldtype in each dimension of global array (array of positive integers)
-. array_of_distribs - distribution of array in each dimension (array of state)
-. array_of_dargs - distribution argument in each dimension (array of positive integers)
-. array_of_psizes - size of process grid in each dimension (array of positive integers)
-. order - array storage order flag (state)
-- oldtype - old datatype (handle)
-
-Output Parameters:
-. newtype - new datatype (handle)
-
-.N ThreadSafe
-
-.N Fortran
-
-.N Errors
-.N MPI_SUCCESS
-.N MPI_ERR_TYPE
-.N MPI_ERR_ARG
-@*/
-int MPI_Type_create_darray(int size,
-                           int rank,
-                           int ndims,
-                           const int array_of_gsizes[],
-                           const int array_of_distribs[],
-                           const int array_of_dargs[],
-                           const int array_of_psizes[],
-                           int order, MPI_Datatype oldtype, MPI_Datatype * newtype)
+static int MPIR_Type_create_darray(int size, int rank, int ndims,
+                                   const MPI_Aint array_of_gsizes[], const int array_of_distribs[],
+                                   const int array_of_dargs[], const int array_of_psizes[],
+                                   int order, MPI_Datatype oldtype, MPI_Datatype * newtype)
 {
-    int mpi_errno = MPI_SUCCESS, i;
+    int mpi_errno = MPI_SUCCESS;
     MPI_Datatype new_handle;
 
-    int procs, tmp_rank, tmp_size, *coords;
+    int procs, tmp_rank, *coords;
+    MPI_Aint tmp_size;
     MPI_Aint *st_offsets, orig_extent, disps[3];
     MPI_Datatype type_old, type_new = MPI_DATATYPE_NULL, tmp_type;
 
-#ifdef HAVE_ERROR_CHECKING
-    MPI_Aint size_with_aint;
-    MPI_Offset size_with_offset;
-#endif
+    MPIR_CHKLMEM_DECL(2);
 
-    int *ints;
-    MPIR_Datatype *datatype_ptr = NULL;
-    MPIR_CHKLMEM_DECL(3);
-    MPIR_FUNC_TERSE_STATE_DECL(MPID_STATE_MPI_TYPE_CREATE_DARRAY);
-
-    MPIR_ERRTEST_INITIALIZED_ORDIE();
-
-    MPID_THREAD_CS_ENTER(GLOBAL, MPIR_THREAD_GLOBAL_ALLFUNC_MUTEX);
-    MPIR_FUNC_TERSE_ENTER(MPID_STATE_MPI_TYPE_CREATE_DARRAY);
-
-    /* Validate parameters, especially handles needing to be converted */
-#ifdef HAVE_ERROR_CHECKING
-    {
-        MPID_BEGIN_ERROR_CHECKS;
-        {
-            MPIR_ERRTEST_DATATYPE(oldtype, "datatype", mpi_errno);
-        }
-        MPID_END_ERROR_CHECKS;
-    }
-#endif
-
-    /* Convert MPI object handles to object pointers */
-    MPIR_Datatype_get_ptr(oldtype, datatype_ptr);
-    MPIR_Datatype_get_extent_macro(oldtype, orig_extent);
-
-    /* Validate parameters and objects (post conversion) */
-#ifdef HAVE_ERROR_CHECKING
-    {
-        MPID_BEGIN_ERROR_CHECKS;
-        {
-            /* Check parameters */
-            MPIR_ERRTEST_ARGNONPOS(size, "size", mpi_errno, MPI_ERR_ARG);
-            /* use MPI_ERR_RANK class for PE-MPI compatibility */
-            MPIR_ERR_CHKANDJUMP3((rank < 0 || rank >= size), mpi_errno, MPI_ERR_RANK,
-                                 "**argrange", "**argrange %s %d %d", "rank", rank, (size - 1));
-            MPIR_ERRTEST_ARGNONPOS(ndims, "ndims", mpi_errno, MPI_ERR_DIMS);
-
-            MPIR_ERRTEST_ARGNULL(array_of_gsizes, "array_of_gsizes", mpi_errno);
-            MPIR_ERRTEST_ARGNULL(array_of_distribs, "array_of_distribs", mpi_errno);
-            MPIR_ERRTEST_ARGNULL(array_of_dargs, "array_of_dargs", mpi_errno);
-            MPIR_ERRTEST_ARGNULL(array_of_psizes, "array_of_psizes", mpi_errno);
-            if (order != MPI_ORDER_C && order != MPI_ORDER_FORTRAN) {
-                mpi_errno = MPIR_Err_create_code(MPI_SUCCESS,
-                                                 MPIR_ERR_RECOVERABLE,
-                                                 __func__,
-                                                 __LINE__,
-                                                 MPI_ERR_ARG, "**arg", "**arg %s", "order");
-                goto fn_fail;
-            }
-
-            tmp_size = 1;
-            for (i = 0; mpi_errno == MPI_SUCCESS && i < ndims; i++) {
-                MPIR_ERRTEST_ARGNONPOS(array_of_gsizes[i], "gsize", mpi_errno, MPI_ERR_ARG);
-                MPIR_ERRTEST_ARGNONPOS(array_of_psizes[i], "psize", mpi_errno, MPI_ERR_ARG);
-
-                if ((array_of_distribs[i] != MPI_DISTRIBUTE_NONE) &&
-                    (array_of_distribs[i] != MPI_DISTRIBUTE_BLOCK) &&
-                    (array_of_distribs[i] != MPI_DISTRIBUTE_CYCLIC)) {
-                    mpi_errno = MPIR_Err_create_code(MPI_SUCCESS,
-                                                     MPIR_ERR_RECOVERABLE,
-                                                     __func__,
-                                                     __LINE__, MPI_ERR_ARG, "**darrayunknown", 0);
-                    goto fn_fail;
-                }
-
-                if ((array_of_dargs[i] != MPI_DISTRIBUTE_DFLT_DARG) && (array_of_dargs[i] <= 0)) {
-                    mpi_errno = MPIR_Err_create_code(MPI_SUCCESS,
-                                                     MPIR_ERR_RECOVERABLE,
-                                                     __func__,
-                                                     __LINE__,
-                                                     MPI_ERR_ARG,
-                                                     "**arg", "**arg %s", "array_of_dargs");
-                    goto fn_fail;
-                }
-
-                if ((array_of_distribs[i] == MPI_DISTRIBUTE_NONE) && (array_of_psizes[i] != 1)) {
-                    mpi_errno = MPIR_Err_create_code(MPI_SUCCESS,
-                                                     MPIR_ERR_RECOVERABLE,
-                                                     __func__,
-                                                     __LINE__,
-                                                     MPI_ERR_ARG,
-                                                     "**darraydist",
-                                                     "**darraydist %d %d", i, array_of_psizes[i]);
-                    goto fn_fail;
-                }
-
-                tmp_size *= array_of_psizes[i];
-            }
-
-            MPIR_ERR_CHKANDJUMP1((tmp_size != size), mpi_errno, MPI_ERR_ARG,
-                                 "**arg", "**arg %s", "array_of_psizes");
-
-            /* TODO: GET THIS CHECK IN ALSO */
-
-            /* check if MPI_Aint is large enough for size of global array.
-             * if not, complain. */
-
-            size_with_aint = orig_extent;
-            for (i = 0; i < ndims; i++)
-                size_with_aint *= array_of_gsizes[i];
-            size_with_offset = orig_extent;
-            for (i = 0; i < ndims; i++)
-                size_with_offset *= array_of_gsizes[i];
-            if (size_with_aint != size_with_offset) {
-                mpi_errno = MPIR_Err_create_code(MPI_SUCCESS,
-                                                 MPIR_ERR_FATAL,
-                                                 __func__,
-                                                 __LINE__,
-                                                 MPI_ERR_ARG,
-                                                 "**darrayoverflow",
-                                                 "**darrayoverflow %L", size_with_offset);
-                goto fn_fail;
-            }
-
-            /* Validate datatype_ptr */
-            MPIR_Datatype_valid_ptr(datatype_ptr, mpi_errno);
-            /* If datatype_ptr is not valid, it will be reset to null */
-            /* --BEGIN ERROR HANDLING-- */
-            if (mpi_errno)
-                goto fn_fail;
-            /* --END ERROR HANDLING-- */
-        }
-        MPID_END_ERROR_CHECKS;
-    }
-#endif /* HAVE_ERROR_CHECKING */
-
-    /* ... body of routine ... */
-
-/* calculate position in Cartesian grid as MPI would (row-major
-   ordering) */
+    /* calculate position in Cartesian grid as MPI would (row-major ordering) */
     MPIR_CHKLMEM_MALLOC_ORJUMP(coords, int *, ndims * sizeof(int), mpi_errno,
                                "position is Cartesian grid", MPL_MEM_COMM);
 
+    MPIR_Datatype_get_extent_macro(oldtype, orig_extent);
     procs = size;
     tmp_rank = rank;
-    for (i = 0; i < ndims; i++) {
+    for (int i = 0; i < ndims; i++) {
         procs = procs / array_of_psizes[i];
         coords[i] = tmp_rank / procs;
         tmp_rank = tmp_rank % procs;
@@ -496,7 +309,7 @@ int MPI_Type_create_darray(int size,
 
     if (order == MPI_ORDER_FORTRAN) {
         /* dimension 0 changes fastest */
-        for (i = 0; i < ndims; i++) {
+        for (int i = 0; i < ndims; i++) {
             switch (array_of_distribs[i]) {
                 case MPI_DISTRIBUTE_BLOCK:
                     mpi_errno = MPIR_Type_block(array_of_gsizes,
@@ -535,16 +348,13 @@ int MPI_Type_create_darray(int size,
             }
             type_old = type_new;
 
-            /* --BEGIN ERROR HANDLING-- */
-            if (mpi_errno != MPI_SUCCESS)
-                goto fn_fail;
-            /* --END ERROR HANDLING-- */
+            MPIR_ERR_CHECK(mpi_errno);
         }
 
         /* add displacement and UB */
         disps[1] = st_offsets[0];
         tmp_size = 1;
-        for (i = 1; i < ndims; i++) {
+        for (int i = 1; i < ndims; i++) {
             tmp_size *= array_of_gsizes[i - 1];
             disps[1] += (MPI_Aint) tmp_size *st_offsets[i];
         }
@@ -554,7 +364,7 @@ int MPI_Type_create_darray(int size,
     else {      /* order == MPI_ORDER_C */
 
         /* dimension ndims-1 changes fastest */
-        for (i = ndims - 1; i >= 0; i--) {
+        for (int i = ndims - 1; i >= 0; i--) {
             switch (array_of_distribs[i]) {
                 case MPI_DISTRIBUTE_BLOCK:
                     mpi_errno = MPIR_Type_block(array_of_gsizes,
@@ -593,16 +403,13 @@ int MPI_Type_create_darray(int size,
             }
             type_old = type_new;
 
-            /* --BEGIN ERROR HANDLING-- */
-            if (mpi_errno != MPI_SUCCESS)
-                goto fn_fail;
-            /* --END ERROR HANDLING-- */
+            MPIR_ERR_CHECK(mpi_errno);
         }
 
         /* add displacement and UB */
         disps[1] = st_offsets[ndims - 1];
         tmp_size = 1;
-        for (i = ndims - 2; i >= 0; i--) {
+        for (int i = ndims - 2; i >= 0; i--) {
             tmp_size *= array_of_gsizes[i + 1];
             disps[1] += (MPI_Aint) tmp_size *st_offsets[i];
         }
@@ -611,29 +418,51 @@ int MPI_Type_create_darray(int size,
     disps[1] *= orig_extent;
 
     disps[2] = orig_extent;
-    for (i = 0; i < ndims; i++)
+    for (int i = 0; i < ndims; i++)
         disps[2] *= (MPI_Aint) (array_of_gsizes[i]);
 
     disps[0] = 0;
 
     mpi_errno = MPIR_Type_blockindexed(1, 1, &disps[1], 1,      /* 1 means disp is in bytes */
                                        type_new, &tmp_type);
-
-    /* --BEGIN ERROR HANDLING-- */
-    if (mpi_errno != MPI_SUCCESS)
-        goto fn_fail;
-    /* --END ERROR HANDLING-- */
+    MPIR_ERR_CHECK(mpi_errno);
 
     mpi_errno = MPIR_Type_create_resized(tmp_type, 0, disps[2], &new_handle);
-
-    /* --BEGIN ERROR HANDLING-- */
-    if (mpi_errno != MPI_SUCCESS)
-        goto fn_fail;
-    /* --END ERROR HANDLING-- */
+    MPIR_ERR_CHECK(mpi_errno);
 
     MPIR_Type_free_impl(&tmp_type);
     MPIR_Type_free_impl(&type_new);
 
+    *newtype = new_handle;
+
+  fn_exit:
+    MPIR_CHKLMEM_FREEALL();
+    return mpi_errno;
+  fn_fail:
+    goto fn_exit;
+}
+
+int MPIR_Type_create_darray_impl(int size, int rank, int ndims,
+                                 const int array_of_gsizes[], const int array_of_distribs[],
+                                 const int array_of_dargs[], const int array_of_psizes[],
+                                 int order, MPI_Datatype oldtype, MPI_Datatype * newtype)
+{
+    int mpi_errno = MPI_SUCCESS;
+
+    MPI_Aint *real_array_of_gsizes;
+    int *ints;
+    MPIR_CHKLMEM_DECL(2);
+
+    MPIR_CHKLMEM_MALLOC_ORJUMP(real_array_of_gsizes, MPI_Aint *, ndims * sizeof(MPI_Aint),
+                               mpi_errno, "real_array_of_gsizes", MPL_MEM_COMM);
+    for (int i = 0; i < ndims; i++) {
+        real_array_of_gsizes[i] = array_of_gsizes[i];
+    }
+
+    mpi_errno = MPIR_Type_create_darray(size, rank, ndims, real_array_of_gsizes, array_of_distribs,
+                                        array_of_dargs, array_of_psizes, order, oldtype, newtype);
+    MPIR_ERR_CHECK(mpi_errno);
+
     /* at this point we have the new type, and we've cleaned up any
      * intermediate types created in the process.  we just need to save
      * all our contents/envelope information.
@@ -647,55 +476,85 @@ int MPI_Type_create_darray(int size,
     ints[1] = rank;
     ints[2] = ndims;
 
-    for (i = 0; i < ndims; i++) {
+    for (int i = 0; i < ndims; i++) {
         ints[i + 3] = array_of_gsizes[i];
     }
-    for (i = 0; i < ndims; i++) {
+    for (int i = 0; i < ndims; i++) {
         ints[i + ndims + 3] = array_of_distribs[i];
     }
-    for (i = 0; i < ndims; i++) {
+    for (int i = 0; i < ndims; i++) {
         ints[i + 2 * ndims + 3] = array_of_dargs[i];
     }
-    for (i = 0; i < ndims; i++) {
+    for (int i = 0; i < ndims; i++) {
         ints[i + 3 * ndims + 3] = array_of_psizes[i];
     }
     ints[4 * ndims + 3] = order;
-    MPIR_Datatype_get_ptr(new_handle, datatype_ptr);
+
+    MPIR_Datatype *datatype_ptr;
+    MPIR_Datatype_get_ptr(*newtype, datatype_ptr);
     mpi_errno = MPIR_Datatype_set_contents(datatype_ptr,
                                            MPI_COMBINER_DARRAY,
-                                           4 * ndims + 4, 0, 1, ints, NULL, &oldtype);
-    /* --BEGIN ERROR HANDLING-- */
-    if (mpi_errno != MPI_SUCCESS)
-        goto fn_fail;
-    /* --END ERROR HANDLING-- */
-
-    mpi_errno = MPIR_Typerep_create_darray(size, rank, ndims, array_of_gsizes, array_of_distribs,
-                                           array_of_dargs, array_of_psizes, order, oldtype,
-                                           datatype_ptr);
+                                           4 * ndims + 4, 0, 0, 1, ints, NULL, NULL, &oldtype);
+    MPIR_ERR_CHECK(mpi_errno);
+
+  fn_exit:
+    MPIR_CHKLMEM_FREEALL();
+    return mpi_errno;
+  fn_fail:
+    goto fn_exit;
+}
+
+int MPIR_Type_create_darray_large_impl(int size, int rank, int ndims,
+                                       const MPI_Aint array_of_gsizes[],
+                                       const int array_of_distribs[], const int array_of_dargs[],
+                                       const int array_of_psizes[], int order, MPI_Datatype oldtype,
+                                       MPI_Datatype * newtype)
+{
+    int mpi_errno = MPI_SUCCESS;
+
+    int *ints;
+    MPI_Aint *counts;
+    MPIR_CHKLMEM_DECL(2);
+
+    mpi_errno = MPIR_Type_create_darray(size, rank, ndims, array_of_gsizes, array_of_distribs,
+                                        array_of_dargs, array_of_psizes, order, oldtype, newtype);
+    MPIR_ERR_CHECK(mpi_errno);
+
+    /* Save contents */
+    MPIR_CHKLMEM_MALLOC_ORJUMP(ints, int *, (3 * ndims + 4) * sizeof(int), mpi_errno,
+                               "content ints array", MPL_MEM_BUFFER);
+    MPIR_CHKLMEM_MALLOC_ORJUMP(counts, MPI_Aint *, ndims * sizeof(MPI_Aint), mpi_errno,
+                               "content counts array", MPL_MEM_BUFFER);
+
+    ints[0] = size;
+    ints[1] = rank;
+    ints[2] = ndims;
+
+    for (int i = 0; i < ndims; i++) {
+        counts[i] = array_of_gsizes[i];
+    }
+    for (int i = 0; i < ndims; i++) {
+        ints[i + 3] = array_of_distribs[i];
+    }
+    for (int i = 0; i < ndims; i++) {
+        ints[i + ndims + 3] = array_of_dargs[i];
+    }
+    for (int i = 0; i < ndims; i++) {
+        ints[i + 2 * ndims + 3] = array_of_psizes[i];
+    }
+    ints[3 * ndims + 3] = order;
+
+    MPIR_Datatype *datatype_ptr;
+    MPIR_Datatype_get_ptr(*newtype, datatype_ptr);
+    mpi_errno = MPIR_Datatype_set_contents(datatype_ptr,
+                                           MPI_COMBINER_DARRAY,
+                                           3 * ndims + 4, 0, ndims, 1, ints, NULL, counts,
+                                           &oldtype);
     MPIR_ERR_CHECK(mpi_errno);
 
-    MPIR_OBJ_PUBLISH_HANDLE(*newtype, new_handle);
-    /* ... end of body of routine ... */
-
   fn_exit:
     MPIR_CHKLMEM_FREEALL();
-    MPIR_FUNC_TERSE_EXIT(MPID_STATE_MPI_TYPE_CREATE_DARRAY);
-    MPID_THREAD_CS_EXIT(GLOBAL, MPIR_THREAD_GLOBAL_ALLFUNC_MUTEX);
     return mpi_errno;
-
   fn_fail:
-    /* --BEGIN ERROR HANDLING-- */
-#ifdef HAVE_ERROR_CHECKING
-    {
-        mpi_errno =
-            MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, __func__, __LINE__, MPI_ERR_OTHER,
-                                 "**mpi_type_create_darray",
-                                 "**mpi_type_create_darray %d %d %d %p %p %p %p %d %D %p", size,
-                                 rank, ndims, array_of_gsizes, array_of_distribs, array_of_dargs,
-                                 array_of_psizes, order, oldtype, newtype);
-    }
-#endif
-    mpi_errno = MPIR_Err_return_comm(NULL, __func__, mpi_errno);
     goto fn_exit;
-    /* --END ERROR HANDLING-- */
 }
