@@ -1,9 +1,6 @@
-/* -*- Mode: C; c-basic-offset:4 ; indent-tabs-mode:nil ; -*- */
 /*
- *   ad_panfs_open.c
- *
- *   Copyright (C) 2001 University of Chicago.
- *   See COPYRIGHT notice in top-level directory.
+ * Copyright (C) by Argonne National Laboratory
+ *     See COPYRIGHT in top-level directory
  */
 
 #include "ad_panfs.h"
@@ -133,7 +130,7 @@ void ADIOI_PANFS_Open6(ADIO_File fd, int *error_code)
                         myname);
                 layout_max_faults = 2;
             }
-            /* as of 6.0.x release, we only support RS enconding */
+            /* as of 6.0.x release, we only support RS encoding */
             if (layout_encoding != PAN_FS_CLIENT_LAYOUT_RAIDN_ENCODING_RS) {
                 FPRINTF(stderr,
                         "%s: panfs_layout_encoding is not a valid value: %u. Setting to default of %u\n",
@@ -352,62 +349,59 @@ void ADIOI_PANFS_Open6(ADIO_File fd, int *error_code)
             /* Error - set layout type to unknown */
             ADIOI_Info_set(fd->info, "panfs_layout_type", "PAN_FS_CLIENT_LAYOUT_TYPE__INVALID");
         } else {
-            MPL_snprintf(temp_buffer, TEMP_BUFFER_SIZE, "%u", file_query_args.layout.agg_type);
+            snprintf(temp_buffer, TEMP_BUFFER_SIZE, "%u", file_query_args.layout.agg_type);
             ADIOI_Info_set(fd->info, "panfs_layout_type", temp_buffer);
             if (file_query_args.layout.layout_is_valid == 1) {
                 switch (file_query_args.layout.agg_type) {
                     case PAN_FS_CLIENT_LAYOUT_TYPE__RAIDN_PARITY_STRIPE:
-                        MPL_snprintf(temp_buffer, TEMP_BUFFER_SIZE, "%u",
-                                     file_query_args.layout.u.raidn_parity_stripe.stripe_unit);
+                        snprintf(temp_buffer, TEMP_BUFFER_SIZE, "%u",
+                                 file_query_args.layout.u.raidn_parity_stripe.stripe_unit);
                         ADIOI_Info_set(fd->info, "panfs_layout_stripe_unit", temp_buffer);
-                        MPL_snprintf(temp_buffer, TEMP_BUFFER_SIZE, "%u",
-                                     file_query_args.layout.u.raidn_parity_stripe.
-                                     parity_stripe_width);
+                        snprintf(temp_buffer, TEMP_BUFFER_SIZE, "%u",
+                                 file_query_args.layout.u.raidn_parity_stripe.parity_stripe_width);
                         ADIOI_Info_set(fd->info, "panfs_layout_parity_stripe_width", temp_buffer);
-                        MPL_snprintf(temp_buffer, TEMP_BUFFER_SIZE, "%u",
-                                     file_query_args.layout.u.raidn_parity_stripe.
-                                     parity_stripe_depth);
+                        snprintf(temp_buffer, TEMP_BUFFER_SIZE, "%u",
+                                 file_query_args.layout.u.raidn_parity_stripe.parity_stripe_depth);
                         ADIOI_Info_set(fd->info, "panfs_layout_parity_stripe_depth", temp_buffer);
-                        MPL_snprintf(temp_buffer, TEMP_BUFFER_SIZE, "%u",
-                                     file_query_args.layout.u.raidn_parity_stripe.total_num_comps);
+                        snprintf(temp_buffer, TEMP_BUFFER_SIZE, "%u",
+                                 file_query_args.layout.u.raidn_parity_stripe.total_num_comps);
                         ADIOI_Info_set(fd->info, "panfs_layout_total_num_comps", temp_buffer);
-                        MPL_snprintf(temp_buffer, TEMP_BUFFER_SIZE, "%u",
-                                     file_query_args.layout.u.raidn_parity_stripe.max_faults);
+                        snprintf(temp_buffer, TEMP_BUFFER_SIZE, "%u",
+                                 file_query_args.layout.u.raidn_parity_stripe.max_faults);
                         ADIOI_Info_set(fd->info, "panfs_layout_max_faults", temp_buffer);
-                        MPL_snprintf(temp_buffer, TEMP_BUFFER_SIZE, "%u",
-                                     file_query_args.layout.u.raidn_parity_stripe.encoding);
+                        snprintf(temp_buffer, TEMP_BUFFER_SIZE, "%u",
+                                 file_query_args.layout.u.raidn_parity_stripe.encoding);
                         ADIOI_Info_set(fd->info, "panfs_layout_encoding", temp_buffer);
                         break;
                     case PAN_FS_CLIENT_LAYOUT_TYPE__RAID1_5_PARITY_STRIPE:
-                        MPL_snprintf(temp_buffer, TEMP_BUFFER_SIZE, "%u",
-                                     file_query_args.layout.u.raid1_5_parity_stripe.stripe_unit);
+                        snprintf(temp_buffer, TEMP_BUFFER_SIZE, "%u",
+                                 file_query_args.layout.u.raid1_5_parity_stripe.stripe_unit);
                         ADIOI_Info_set(fd->info, "panfs_layout_stripe_unit", temp_buffer);
-                        MPL_snprintf(temp_buffer, TEMP_BUFFER_SIZE, "%u",
-                                     file_query_args.layout.u.raid1_5_parity_stripe.
-                                     parity_stripe_width);
+                        snprintf(temp_buffer, TEMP_BUFFER_SIZE, "%u",
+                                 file_query_args.layout.u.raid1_5_parity_stripe.
+                                 parity_stripe_width);
                         ADIOI_Info_set(fd->info, "panfs_layout_parity_stripe_width", temp_buffer);
-                        MPL_snprintf(temp_buffer, TEMP_BUFFER_SIZE, "%u",
-                                     file_query_args.layout.u.raid1_5_parity_stripe.
-                                     parity_stripe_depth);
+                        snprintf(temp_buffer, TEMP_BUFFER_SIZE, "%u",
+                                 file_query_args.layout.u.raid1_5_parity_stripe.
+                                 parity_stripe_depth);
                         ADIOI_Info_set(fd->info, "panfs_layout_parity_stripe_depth", temp_buffer);
-                        MPL_snprintf(temp_buffer, TEMP_BUFFER_SIZE, "%u",
-                                     file_query_args.layout.u.
-                                     raid1_5_parity_stripe.total_num_comps);
+                        snprintf(temp_buffer, TEMP_BUFFER_SIZE, "%u",
+                                 file_query_args.layout.u.raid1_5_parity_stripe.total_num_comps);
                         ADIOI_Info_set(fd->info, "panfs_layout_total_num_comps", temp_buffer);
-                        MPL_snprintf(temp_buffer, TEMP_BUFFER_SIZE, "%u",
-                                     file_query_args.layout.u.raid1_5_parity_stripe.
-                                     layout_visit_policy);
+                        snprintf(temp_buffer, TEMP_BUFFER_SIZE, "%u",
+                                 file_query_args.layout.u.raid1_5_parity_stripe.
+                                 layout_visit_policy);
                         ADIOI_Info_set(fd->info, "panfs_layout_visit_policy", temp_buffer);
                         break;
                     case PAN_FS_CLIENT_LAYOUT_TYPE__RAID10:
-                        MPL_snprintf(temp_buffer, TEMP_BUFFER_SIZE, "%u",
-                                     file_query_args.layout.u.raid10.stripe_unit);
+                        snprintf(temp_buffer, TEMP_BUFFER_SIZE, "%u",
+                                 file_query_args.layout.u.raid10.stripe_unit);
                         ADIOI_Info_set(fd->info, "panfs_layout_stripe_unit", temp_buffer);
-                        MPL_snprintf(temp_buffer, TEMP_BUFFER_SIZE, "%u",
-                                     file_query_args.layout.u.raid10.total_num_comps);
+                        snprintf(temp_buffer, TEMP_BUFFER_SIZE, "%u",
+                                 file_query_args.layout.u.raid10.total_num_comps);
                         ADIOI_Info_set(fd->info, "panfs_layout_total_num_comps", temp_buffer);
-                        MPL_snprintf(temp_buffer, TEMP_BUFFER_SIZE, "%u",
-                                     file_query_args.layout.u.raid10.layout_visit_policy);
+                        snprintf(temp_buffer, TEMP_BUFFER_SIZE, "%u",
+                                 file_query_args.layout.u.raid10.layout_visit_policy);
                         ADIOI_Info_set(fd->info, "panfs_layout_visit_policy", temp_buffer);
                         break;
                     case PAN_FS_CLIENT_LAYOUT_TYPE__INVALID:
