@@ -11,7 +11,7 @@
 #include <assert.h>
 /* #define DEBUG */
 
-void ADIOI_Calc_file_realms_user_size(ADIO_File fd, int fr_size,
+void ADIOI_Calc_file_realms_user_size(ADIO_File fd, MPI_Count fr_size,
                                       int nprocs_for_coll,
                                       ADIO_Offset * file_realm_st_offs,
                                       MPI_Datatype * file_realm_types);
@@ -25,9 +25,9 @@ void ADIOI_Calc_file_realms_fsize(ADIO_File fd,
                                   ADIO_Offset max_end_offset,
                                   ADIO_Offset * file_realm_st_offs,
                                   MPI_Datatype * file_realm_types);
-void ADIOI_Create_fr_simpletype(int size, int nprocs_for_coll, MPI_Datatype * simpletype);
-static void align_fr(int fr_size, ADIO_Offset fr_off, int alignment,
-                     int *aligned_fr_size, ADIO_Offset * aligned_fr_off);
+void ADIOI_Create_fr_simpletype(MPI_Count size, int nprocs_for_coll, MPI_Datatype * simpletype);
+static void align_fr(MPI_Count fr_size, ADIO_Offset fr_off, int alignment,
+                     MPI_Count * aligned_fr_size, ADIO_Offset * aligned_fr_off);
 void ADIOI_Verify_fr(int nprocs_for_coll, ADIO_Offset * file_realm_st_offs,
                      MPI_Datatype * file_realm_types);
 
@@ -68,9 +68,9 @@ void ADIOI_Calc_file_realms(ADIO_File fd, ADIO_Offset min_st_offset, ADIO_Offset
             file_realm_types = fd->file_realm_types;
         }
         *file_realm_st_offs = min_st_offset;
-        MPI_Type_contiguous((max_end_offset - min_st_offset + 1), MPI_BYTE, file_realm_types);
+        MPI_Type_contiguous_c((max_end_offset - min_st_offset + 1), MPI_BYTE, file_realm_types);
         MPI_Type_commit(file_realm_types);
-        ADIOI_Flatten_datatype(*file_realm_types);
+        ADIOI_Flatten_and_find(*file_realm_types);
     } else if (fd->file_realm_st_offs == NULL) {
         file_realm_st_offs = (ADIO_Offset *)
             ADIOI_Malloc(nprocs_for_coll * sizeof(ADIO_Offset));
@@ -84,13 +84,13 @@ void ADIOI_Calc_file_realms(ADIO_File fd, ADIO_Offset min_st_offset, ADIO_Offset
                                        file_realm_st_offs, file_realm_types);
             /* flatten file realm datatype for future use - only one
              * because all are the same*/
-            ADIOI_Flatten_datatype(file_realm_types[0]);
+            ADIOI_Flatten_and_find(file_realm_types[0]);
         } else if (file_realm_calc_type == ADIOI_FR_FSZ) {
             ADIOI_Calc_file_realms_fsize(fd, nprocs_for_coll, max_end_offset,
                                          file_realm_st_offs, file_realm_types);
             /* flatten file realm datatype for future use - only one
              * because all are the same*/
-            ADIOI_Flatten_datatype(file_realm_types[0]);
+            ADIOI_Flatten_and_find(file_realm_types[0]);
         } else if (file_realm_calc_type == ADIOI_FR_USR_REALMS) {
             /* copy user provided realm datatypes and realm offsets in
              * hints to file descriptor. may also want to verify that
@@ -101,7 +101,7 @@ void ADIOI_Calc_file_realms(ADIO_File fd, ADIO_Offset min_st_offset, ADIO_Offset
                                              nprocs_for_coll, file_realm_st_offs, file_realm_types);
             /* flatten file realm datatype for future use - only one
              * because all are the same */
-            ADIOI_Flatten_datatype(file_realm_types[0]);
+            ADIOI_Flatten_and_find(file_realm_types[0]);
         }
     }
     fd->file_realm_st_offs = file_realm_st_offs;
@@ -111,13 +111,13 @@ void ADIOI_Calc_file_realms(ADIO_File fd, ADIO_Offset min_st_offset, ADIO_Offset
 #endif
 }
 
-void ADIOI_Calc_file_realms_user_size(ADIO_File fd, int fr_size,
+void ADIOI_Calc_file_realms_user_size(ADIO_File fd, MPI_Count fr_size,
                                       int nprocs_for_coll,
                                       ADIO_Offset * file_realm_st_offs,
                                       MPI_Datatype * file_realm_types)
 {
     int i;
-    int aligned_fr_size;
+    MPI_Count aligned_fr_size;
     ADIO_Offset aligned_fr_off;
     MPI_Datatype simpletype;
 
@@ -151,7 +151,8 @@ void ADIOI_Calc_file_realms_aar(ADIO_File fd, int nprocs_for_coll, int cb_pfr,
                                 ADIO_Offset max_end_offset,
                                 ADIO_Offset * file_realm_st_offs, MPI_Datatype * file_realm_types)
 {
-    int fr_size, aligned_fr_size, i;
+    MPI_Count fr_size, aligned_fr_size;
+    int i;
     MPI_Datatype simpletype;
     ADIO_Offset aligned_start_off;
     char value[9];
@@ -178,7 +179,7 @@ void ADIOI_Calc_file_realms_aar(ADIO_File fd, int nprocs_for_coll, int cb_pfr,
 #endif
     }
     if (fd->hints->cb_pfr == ADIOI_HINT_ENABLE) {
-        MPL_snprintf(value, sizeof(value), "%d", fr_size);
+        snprintf(value, sizeof(value), "%lld", (long long) fr_size);
         ADIOI_Info_set(fd->info, "romio_cb_fr_type", value);
     }
 }
@@ -187,8 +188,8 @@ void ADIOI_Calc_file_realms_fsize(ADIO_File fd, int nprocs_for_coll,
                                   ADIO_Offset max_end_offset,
                                   ADIO_Offset * file_realm_st_offs, MPI_Datatype * file_realm_types)
 {
-    int fr_size, aligned_fr_size, error_code, i;
-    int fsize;
+    int error_code, i;
+    MPI_Count fr_size, aligned_fr_size, fsize;
     ADIO_Offset aligned_fr_off;
     ADIO_Fcntl_t fcntl_struct;
     MPI_Datatype simpletype;
@@ -208,7 +209,7 @@ void ADIOI_Calc_file_realms_fsize(ADIO_File fd, int nprocs_for_coll,
 }
 
 /* creates a datatype with an empty trailing edge */
-void ADIOI_Create_fr_simpletype(int size, int nprocs_for_coll, MPI_Datatype * simpletype)
+void ADIOI_Create_fr_simpletype(MPI_Count size, int nprocs_for_coll, MPI_Datatype * simpletype)
 {
     MPI_Aint lb, ub;
     MPI_Datatype type;
@@ -216,7 +217,7 @@ void ADIOI_Create_fr_simpletype(int size, int nprocs_for_coll, MPI_Datatype * si
     lb = 0;
     ub = size * nprocs_for_coll;
 
-    MPI_Type_contiguous(size, MPI_BYTE, &type);
+    MPI_Type_contiguous_c(size, MPI_BYTE, &type);
     MPI_Type_create_resized(type, lb, ub, simpletype);
 
     MPI_Type_free(&type);
@@ -243,8 +244,8 @@ int ADIOI_Agg_idx(int rank, ADIO_File fd)
     return -1;
 }
 
-static void align_fr(int fr_size, ADIO_Offset fr_off, int alignment,
-                     int *aligned_fr_size, ADIO_Offset * aligned_fr_off)
+static void align_fr(MPI_Count fr_size, ADIO_Offset fr_off, int alignment,
+                     MPI_Count * aligned_fr_size, ADIO_Offset * aligned_fr_off)
 {
     *aligned_fr_off = fr_off - (fr_off % alignment);
     *aligned_fr_size = ((fr_off + fr_size) / alignment) * alignment - *aligned_fr_off;
