@@ -1,8 +1,6 @@
-/* -*- Mode: C; c-basic-offset:4 ; indent-tabs-mode:nil ; -*- */
 /*
- *
- *   Copyright (C) 2001 University of Chicago.
- *   See COPYRIGHT notice in top-level directory.
+ * Copyright (C) by Argonne National Laboratory
+ *     See COPYRIGHT in top-level directory
  */
 
 /* cb_config_list.c
@@ -18,7 +16,6 @@
  */
 
 #include "adio.h"
-#include "mpi.h"
 #include "adio_cb_config_list.h"
 #include <stdio.h>
 #include <string.h>
@@ -36,11 +33,9 @@
 
 /* a couple of globals keep things simple */
 int ADIOI_cb_config_list_keyval = MPI_KEYVAL_INVALID;
-static char *yylval;
-static char *token_ptr;
 
 /* internal stuff */
-static int get_max_procs(int cb_nodes);
+static int get_max_procs(int cb_nodes, char *yylval, char **token_ptr);
 static int match_procs(char *name, int max_per_proc, char *procnames[],
                        char used_procnames[],
                        int nr_procnames, int ranks[], int nr_ranks, int *nr_ranks_allocated);
@@ -49,7 +44,7 @@ static int match_this_proc(char *name, int cur_proc, int max_matches,
                            int nr_procnames, int ranks[], int nr_ranks, int nr_ranks_allocated);
 static int find_name(char *name, char *procnames[], char used_procnames[],
                      int nr_procnames, int start_ind);
-static int cb_config_list_lex(void);
+static int cb_config_list_lex(char *yylval, char **token_ptr);
 
 
 /* ADIOI_cb_bcast_rank_map() - broadcast the rank array
@@ -86,7 +81,7 @@ int ADIOI_cb_bcast_rank_map(ADIO_File fd)
     /* TEMPORARY -- REMOVE WHEN NO LONGER UPDATING INFO FOR
      * FS-INDEP. */
     value = (char *) ADIOI_Malloc((MPI_MAX_INFO_VAL + 1) * sizeof(char));
-    MPL_snprintf(value, MPI_MAX_INFO_VAL + 1, "%d", fd->hints->cb_nodes);
+    snprintf(value, MPI_MAX_INFO_VAL + 1, "%d", fd->hints->cb_nodes);
     ADIOI_Info_set(fd->info, "cb_nodes", value);
     p = value;
     /* the (by MPI rank) list of aggregators can be larger than
@@ -94,8 +89,8 @@ int ADIOI_cb_bcast_rank_map(ADIO_File fd)
      * wasn't clever enough to figure out how to rewind and put '...' at the
      * end in the truncate case */
     for (i = 0; i < fd->hints->cb_nodes; i++) {
-        int incr, remain = (MPI_MAX_INFO_VAL) - (p - value);
-        incr = MPL_snprintf(p, remain, "%d ", fd->hints->ranklist[i]);
+        long incr, remain = (MPI_MAX_INFO_VAL) - (p - value);
+        incr = snprintf(p, remain, "%d ", fd->hints->ranklist[i]);
         if (incr >= remain)
             break;
         p += incr;
@@ -128,11 +123,11 @@ int ADIOI_cb_gather_name_array(MPI_Comm comm, MPI_Comm dupcomm, ADIO_cb_name_arr
 
     if (ADIOI_cb_config_list_keyval == MPI_KEYVAL_INVALID) {
         /* cleaned up by ADIOI_End_call */
-        MPI_Keyval_create((MPI_Copy_function *) ADIOI_cb_copy_name_array,
-                          (MPI_Delete_function *) ADIOI_cb_delete_name_array,
-                          &ADIOI_cb_config_list_keyval, NULL);
+        MPI_Comm_create_keyval((MPI_Copy_function *) ADIOI_cb_copy_name_array,
+                               (MPI_Delete_function *) ADIOI_cb_delete_name_array,
+                               &ADIOI_cb_config_list_keyval, NULL);
     } else {
-        MPI_Attr_get(comm, ADIOI_cb_config_list_keyval, (void *) &array, &found);
+        MPI_Comm_get_attr(comm, ADIOI_cb_config_list_keyval, (void *) &array, &found);
         if (found) {
             ADIOI_Assert(array != NULL);
             *arrayp = array;
@@ -244,8 +239,8 @@ int ADIOI_cb_gather_name_array(MPI_Comm comm, MPI_Comm dupcomm, ADIO_cb_name_arr
      * it next time an open is performed on this same comm, and on the
      * dupcomm, so we can use it in I/O operations.
      */
-    MPI_Attr_put(comm, ADIOI_cb_config_list_keyval, array);
-    MPI_Attr_put(dupcomm, ADIOI_cb_config_list_keyval, array);
+    MPI_Comm_set_attr(comm, ADIOI_cb_config_list_keyval, array);
+    MPI_Comm_set_attr(dupcomm, ADIOI_cb_config_list_keyval, array);
     *arrayp = array;
     return 0;
 }
@@ -265,6 +260,8 @@ int ADIOI_cb_config_list_parse(char *config_list,
     int token, max_procs, cur_rank = 0, nr_procnames;
     char *cur_procname, *cur_procname_p, **procnames;
     char *used_procnames;
+    char *yylval;
+    char *token_ptr;
 
     nr_procnames = array->namect;
     procnames = array->names;
@@ -318,7 +315,7 @@ int ADIOI_cb_config_list_parse(char *config_list,
     }
 
     while (cur_rank < cb_nodes) {
-        token = cb_config_list_lex();
+        token = cb_config_list_lex(yylval, &token_ptr);
 
         if (token == AGG_EOS) {
             ADIOI_Free(cur_procname);
@@ -348,7 +345,7 @@ int ADIOI_cb_config_list_parse(char *config_list,
         }
 
         /* after we have saved the current procname, we can grab max_procs */
-        max_procs = get_max_procs(cb_nodes);
+        max_procs = get_max_procs(cb_nodes, yylval, &token_ptr);
 
 #ifdef CB_CONFIG_LIST_DEBUG
         if (token == AGG_WILDCARD) {
@@ -632,19 +629,19 @@ static int find_name(char *name,
  *
  * Returns # of processes, or -1 on error.
  */
-static int get_max_procs(int cb_nodes)
+static int get_max_procs(int cb_nodes, char *yylval, char **token_ptr)
 {
     int token, max_procs = -1;
     char *errptr;
 
-    token = cb_config_list_lex();
+    token = cb_config_list_lex(yylval, token_ptr);
 
     switch (token) {
         case AGG_EOS:
         case AGG_COMMA:
             return 1;
         case AGG_COLON:
-            token = cb_config_list_lex();
+            token = cb_config_list_lex(yylval, token_ptr);
             if (token != AGG_WILDCARD && token != AGG_STRING)
                 return -1;
             if (token == AGG_WILDCARD)
@@ -657,7 +654,7 @@ static int get_max_procs(int cb_nodes)
                 }
             }
             /* strip off next comma (if there is one) */
-            token = cb_config_list_lex();
+            token = cb_config_list_lex(yylval, token_ptr);
             if (token != AGG_COMMA && token != AGG_EOS)
                 return -1;
 
@@ -688,28 +685,29 @@ static int get_max_procs(int cb_nodes)
 #define DELIMS ":,"
 #endif
 
-static int cb_config_list_lex(void)
+static int cb_config_list_lex(char *yylval, char **token_ptr)
 {
     int slen;
+    char *token = *token_ptr;
 
-    if (*token_ptr == '\0')
+    if (*token == '\0')
         return AGG_EOS;
 
-    slen = (int) strcspn(token_ptr, DELIMS);
+    slen = (int) strcspn(token, DELIMS);
 
-    if (*token_ptr == COLON) {
-        token_ptr++;
+    if (*token == COLON) {
+        *token_ptr = token + 1;
         return AGG_COLON;
     }
-    if (*token_ptr == COMMA) {
-        token_ptr++;
+    if (*token == COMMA) {
+        *token_ptr = token + 1;
         return AGG_COMMA;
     }
 
-    if (*token_ptr == '*') {
+    if (*token == '*') {
         /* make sure that we don't have characters after the '*' */
         if (slen == 1) {
-            token_ptr++;
+            *token_ptr = token + 1;
             return AGG_WILDCARD;
         } else
             return AGG_ERROR;
@@ -722,8 +720,8 @@ static int cb_config_list_lex(void)
      * should ensure that no one tries to use wildcards with strings
      * (e.g. "ccn*").
      */
-    ADIOI_Strncpy(yylval, token_ptr, slen);
+    ADIOI_Strncpy(yylval, token, slen);
     yylval[slen] = '\0';
-    token_ptr += slen;
+    *token_ptr = token + slen;
     return AGG_STRING;
 }
