@@ -1,7 +1,6 @@
-/* -*- Mode: C; c-basic-offset:4 ; indent-tabs-mode:nil ; -*- */
 /*
- *  (C) 2006 by Argonne National Laboratory.
- *      See COPYRIGHT in top-level directory.
+ * Copyright (C) by Argonne National Laboratory
+ *     See COPYRIGHT in top-level directory
  */
 
 #ifndef MPID_NEM_IMPL_H_INCLUDED
@@ -21,7 +20,7 @@
 
 int MPID_nem_mpich_init(void);
 int MPID_nem_coll_init (void);
-int MPID_nem_send_iov(MPIDI_VC_t *vc, MPIR_Request **sreq_ptr, MPL_IOV *iov, int n_iov);
+int MPID_nem_send_iov(MPIDI_VC_t *vc, MPIR_Request **sreq_ptr, struct iovec *iov, int n_iov);
 int MPID_nem_lmt_pkthandler_init(MPIDI_CH3_PktHandler_Fcn *pktArray[], int arraySize);
 int MPID_nem_register_initcomp_cb(int (* callback)(void));
 int MPID_nem_choose_netmod(void);
@@ -35,16 +34,13 @@ int MPID_nem_lmt_RndvRecv(struct MPIDI_VC *vc, MPIR_Request *rreq);
 
 #define MPID_nem_mpich_release_fbox(cell)                               \
     do {                                                                \
-        OPA_store_release_int(&MPID_nem_mem_region.mailboxes.in[(cell)->pkt.header.source]->mpich.flag.value, 0); \
+        MPL_atomic_release_store_int(&MPID_nem_mem_region.mailboxes.in[(cell)->header.source]->flag, 0); \
     } while (0)
 
-/* initialize shared-memory MPI_Barrier variables */
-int MPID_nem_barrier_vars_init (MPID_nem_barrier_vars_t *barrier_region);
-
-/* assumes value!=0 means the fbox is full.  Contains acquire barrier to
+/* assumes flag!=0 means the fbox is full.  Contains acquire barrier to
  * ensure that later operations that are dependent on this check don't
  * escape earlier than this check. */
-#define MPID_nem_fbox_is_full(pbox_) (OPA_load_acquire_int(&(pbox_)->flag.value))
+#define MPID_nem_fbox_is_full(pbox_) (MPL_atomic_acquire_load_int(&(pbox_)->flag))
 
 typedef struct MPID_nem_pkt_lmt_rts
 {
@@ -135,15 +131,15 @@ typedef union MPIDI_CH3_nem_pkt
 
 #define MPID_nem_lmt_send_RTS(vc, rts_pkt, s_cookie_buf, s_cookie_len) do {                             \
         MPIR_Request *_rts_req;                                                                         \
-        MPL_IOV _iov[2];                                                                               \
+        struct iovec _iov[2];                                                                               \
                                                                                                         \
         MPL_DBG_MSG(MPIDI_CH3_DBG_OTHER,VERBOSE,"sending rndv RTS packet");                                      \
         (rts_pkt)->cookie_len = (s_cookie_len);                                                         \
                                                                                                         \
-        _iov[0].MPL_IOV_BUF = (MPL_IOV_BUF_CAST)(rts_pkt);                                            \
-        _iov[0].MPL_IOV_LEN = sizeof(*(rts_pkt));                                                      \
-        _iov[1].MPL_IOV_BUF = (MPL_IOV_BUF_CAST)(s_cookie_buf);                                       \
-        _iov[1].MPL_IOV_LEN = (s_cookie_len);                                                          \
+        _iov[0].iov_base = (void *)(rts_pkt);                                            \
+        _iov[0].iov_len = sizeof(*(rts_pkt));                                                      \
+        _iov[1].iov_base = (void *)(s_cookie_buf);                                       \
+        _iov[1].iov_len = (s_cookie_len);                                                          \
                                                                                                         \
         MPL_DBG_MSGPKT((vc), (rts_pkt)->match.parts.tag, (rts_pkt)->match.parts.context_id, (rts_pkt)->match.parts.rank, \
                         (rts_pkt)->data_sz, "Rndv");                                                    \
@@ -179,7 +175,7 @@ typedef union MPIDI_CH3_nem_pkt
 #define MPID_nem_lmt_send_CTS(vc, rreq, r_cookie_buf, r_cookie_len) do {                                \
         MPID_PKT_DECL_CAST(_upkt, MPID_nem_pkt_lmt_cts_t, _cts_pkt);                                    \
         MPIR_Request *_cts_req;                                                                         \
-        MPL_IOV _iov[2];                                                                               \
+        struct iovec _iov[2];                                                                               \
                                                                                                         \
         MPL_DBG_MSG(MPIDI_CH3_DBG_OTHER,VERBOSE,"sending rndv CTS packet");                                      \
         MPIDI_Pkt_init(_cts_pkt, MPIDI_NEM_PKT_LMT_CTS);                                                \
@@ -188,10 +184,10 @@ typedef union MPIDI_CH3_nem_pkt
         _cts_pkt->cookie_len = (r_cookie_len);                                                          \
         _cts_pkt->data_sz = (rreq)->ch.lmt_data_sz;                                                     \
                                                                                                         \
-        _iov[0].MPL_IOV_BUF = (MPL_IOV_BUF_CAST)_cts_pkt;                                             \
-        _iov[0].MPL_IOV_LEN = sizeof(*_cts_pkt);                                                       \
-        _iov[1].MPL_IOV_BUF = (MPL_IOV_BUF_CAST)(r_cookie_buf);                                       \
-        _iov[1].MPL_IOV_LEN = (r_cookie_len);                                                          \
+        _iov[0].iov_base = (void *)_cts_pkt;                                             \
+        _iov[0].iov_len = sizeof(*_cts_pkt);                                                       \
+        _iov[1].iov_base = (void *)(r_cookie_buf);                                       \
+        _iov[1].iov_len = (r_cookie_len);                                                          \
                                                                                                         \
         mpi_errno = MPIDI_CH3_iStartMsgv((vc), _iov, (r_cookie_len) ? 2 : 1, &_cts_req);                \
         MPIR_ERR_CHKANDJUMP(mpi_errno, mpi_errno, MPI_ERR_OTHER, "**ctspkt");                           \
@@ -202,17 +198,13 @@ typedef union MPIDI_CH3_nem_pkt
         }                                                                                               \
     } while (0)
         
-#undef FUNCNAME
-#define FUNCNAME MPID_nem_lmt_send_COOKIE
-#undef FCNAME
-#define FCNAME MPL_QUOTE(FUNCNAME)
 static inline int MPID_nem_lmt_send_COOKIE(MPIDI_VC_t *vc, MPIR_Request *req,
                                            void *cookie_buf, MPI_Aint cookie_len)
 {
     int mpi_errno = MPI_SUCCESS;
     MPID_PKT_DECL_CAST(_upkt, MPID_nem_pkt_lmt_cookie_t, cookie_pkt);
     MPIR_Request *cookie_req;
-    MPL_IOV iov[2];
+    struct iovec iov[2];
 
     MPL_DBG_MSG(MPIDI_CH3_DBG_OTHER,VERBOSE,"sending rndv COOKIE packet");
     MPIDI_Pkt_init(cookie_pkt, MPIDI_NEM_PKT_LMT_COOKIE);
@@ -239,10 +231,10 @@ static inline int MPID_nem_lmt_send_COOKIE(MPIDI_VC_t *vc, MPIR_Request *req,
             break;
     }
 
-    iov[0].MPL_IOV_BUF = (MPL_IOV_BUF_CAST) cookie_pkt;
-    iov[0].MPL_IOV_LEN = sizeof(*cookie_pkt);
-    iov[1].MPL_IOV_BUF = (MPL_IOV_BUF_CAST) cookie_buf;
-    iov[1].MPL_IOV_LEN = cookie_len;
+    iov[0].iov_base = (void *) cookie_pkt;
+    iov[0].iov_len = sizeof(*cookie_pkt);
+    iov[1].iov_base = (void *) cookie_buf;
+    iov[1].iov_len = cookie_len;
 
     mpi_errno = MPIDI_CH3_iStartMsgv(vc, iov, (cookie_len ? 2 : 1), &cookie_req);
     MPIR_ERR_CHKANDJUMP(mpi_errno, mpi_errno, MPI_ERR_OTHER, "**cookiepkt");
