@@ -7,11 +7,6 @@
 
 #include "tcp_impl.h"
 #include "socksm.h"
-#ifdef USE_PMI2_API
-#include "pmi2.h"
-#else
-#include "pmi.h"
-#endif
 
 /* FIXME trace/log all the state transitions */
 
@@ -833,16 +828,10 @@ int MPID_nem_tcp_connect(struct MPIDI_VC *const vc)
          */
         if (vc->pg != NULL) {   /* VC is not a temporary one */
             char *bc;
-            int pmi_errno;
             int val_max_sz;
 
-#ifdef USE_PMI2_API
-            val_max_sz = PMI2_MAX_VALLEN;
-#else
-            pmi_errno = PMI_KVS_Get_value_length_max(&val_max_sz);
-            MPIR_ERR_CHKANDJUMP1(pmi_errno, mpi_errno, MPI_ERR_OTHER, "**fail", "**fail %d",
-                                 pmi_errno);
-#endif
+            val_max_sz = MPIR_pmi_max_val_size();
+
             MPIR_CHKLMEM_MALLOC(bc, char *, val_max_sz, mpi_errno, "bc", MPL_MEM_OTHER);
 
             sc->is_tmpvc = FALSE;
