@@ -29,8 +29,9 @@
  */
 
 /* Initializes hint structure based MPIDI_OFI_global.settings (or config macros) */
-void MPIDI_OFI_init_hints(struct fi_info *hints)
+int MPIDI_OFI_init_hints(struct fi_info *hints)
 {
+    int mpi_errno = MPI_SUCCESS;
     MPIR_Assert(hints != NULL);
 
     /* ------------------------------------------------------------------------ */
@@ -86,12 +87,15 @@ void MPIDI_OFI_init_hints(struct fi_info *hints)
         hints->caps |= FI_ATOMICS;      /* Atomics capabilities    */
     }
 
-    if (MPIDI_OFI_ENABLE_TAGGED) {
-        hints->caps |= FI_TAGGED;       /* Tag matching interface  */
+    if (MPIDI_OFI_ENABLE_DATA) {
         hints->caps |= FI_DIRECTED_RECV;        /* Match source address    */
         hints->domain_attr->cq_data_size = MPIDI_OFI_MIN_CQ_DATA_SIZE;  /* Minimum size for completion data entry */
     }
 
+    if (MPIDI_OFI_ENABLE_TAGGED) {
+        hints->caps |= FI_TAGGED;       /* Tag matching interface  */
+    }
+
     if (MPIDI_OFI_ENABLE_AM) {
         hints->caps |= FI_MSG;  /* Message Queue apis      */
         hints->caps |= FI_MULTI_RECV;   /* Shared receive buffer   */
@@ -159,6 +163,16 @@ void MPIDI_OFI_init_hints(struct fi_info *hints)
             hints->domain_attr->mr_mode |= FI_MR_PROV_KEY;
         }
 #endif
+
+#ifdef FI_MR_HMEM
+        if (MPIDI_OFI_ENABLE_HMEM) {
+            hints->domain_attr->mr_mode |= FI_MR_HMEM;
+        }
+#endif
+
+#ifdef FI_MR_ENDPOINT
+        hints->domain_attr->mr_mode |= FI_MR_ENDPOINT;
+#endif
     } else {
         /* In old versions FI_MR_BASIC is equivallent to set
          * FI_MR_VIRT_ADDR, FI_MR_PROV_KEY, and FI_MR_ALLOCATED on.
@@ -173,7 +187,9 @@ void MPIDI_OFI_init_hints(struct fi_info *hints)
         }
     }
     hints->tx_attr->op_flags = FI_COMPLETION;
+    /* apply SAS hint to both tx and rx or else ordering may not be guaranteed */
     hints->tx_attr->msg_order = FI_ORDER_SAS;
+    hints->rx_attr->msg_order = FI_ORDER_SAS;
     /* direct RMA operations supported only with delivery complete mode,
      * else (AM mode) delivery complete is not required */
     if (MPIDI_OFI_ENABLE_RMA || MPIDI_OFI_ENABLE_ATOMICS) {
@@ -182,6 +198,23 @@ void MPIDI_OFI_init_hints(struct fi_info *hints)
         if (MPIDI_OFI_ENABLE_ATOMICS)
             hints->tx_attr->msg_order |= MPIDI_OFI_ATOMIC_ORDER_FLAGS;
     }
+
+    if (MPIDI_OFI_ENABLE_TRIGGERED) {
+        if (MPIDI_OFI_ENABLE_ATOMICS && MPIDI_OFI_ENABLE_TAGGED) {
+            /* needs FI_TAGGED FI_DIRECTED_RECV ... */
+            hints->caps |=
+                FI_TRIGGER | FI_DIRECTED_RECV | FI_RMA_EVENT | FI_READ | FI_WRITE | FI_RECV |
+                FI_SEND | FI_REMOTE_READ | FI_REMOTE_WRITE | FI_ATOMICS | FI_RMA;
+            hints->domain_attr->mr_mode |= FI_MR_RMA_EVENT;
+            hints->domain_attr->data_progress = FI_PROGRESS_AUTO;
+        } else {
+            MPL_DBG_MSG_FMT(MPIDI_CH4_DBG_GENERAL, VERBOSE,
+                            (MPL_DBG_FDEST,
+                             "Triggered ops cannot be enabled if atomics or tagged is disabled"));
+            MPIR_ERR_CHKANDJUMP(1, mpi_errno, MPI_ERR_OTHER, "**ofid_enable_trigger");
+        }
+    }
+
     hints->tx_attr->comp_order = FI_ORDER_NONE;
     hints->rx_attr->op_flags = FI_COMPLETION;
     hints->rx_attr->total_buffered_recv = 0;    /* FI_RM_ENABLED ensures buffering of unexpected messages */
@@ -190,6 +223,15 @@ void MPIDI_OFI_init_hints(struct fi_info *hints)
         /*     PROTOCOL         |  CONTEXT  |        SOURCE         |       TAG          */
         MPIDI_OFI_PROTOCOL_MASK | 0 | MPIDI_OFI_SOURCE_MASK | 0 /* With source bits */ :
         MPIDI_OFI_PROTOCOL_MASK | 0 | 0 | MPIDI_OFI_TAG_MASK /* No source bits */ ;
+
+#ifdef FI_HMEM
+    if (MPIDI_OFI_ENABLE_HMEM && MPIR_CVAR_ENABLE_GPU) {
+        hints->caps |= FI_HMEM;
+    }
+#endif
+
+  fn_fail:
+    return mpi_errno;
 }
 
 void MPIDI_OFI_set_auto_progress(struct fi_info *hints)
@@ -217,6 +259,7 @@ void MPIDI_OFI_init_settings(MPIDI_OFI_capabilities_t * p_settings, const char *
     memset(p_settings, 0, sizeof(MPIDI_OFI_capabilities_t));
 
     /* Seed the settings values for cases where we are using runtime sets */
+    UPDATE_SETTING_BY_CAP(enable_data, MPIR_CVAR_CH4_OFI_ENABLE_DATA);
     UPDATE_SETTING_BY_CAP(enable_av_table, MPIR_CVAR_CH4_OFI_ENABLE_AV_TABLE);
     UPDATE_SETTING_BY_CAP(enable_scalable_endpoints, MPIR_CVAR_CH4_OFI_ENABLE_SCALABLE_ENDPOINTS);
     /* If the user specifies -1 (=don't care) and the provider supports it, then try to use STX
@@ -226,6 +269,7 @@ void MPIDI_OFI_init_settings(MPIDI_OFI_capabilities_t * p_settings, const char *
     UPDATE_SETTING_BY_CAP(enable_mr_virt_address, MPIR_CVAR_CH4_OFI_ENABLE_MR_VIRT_ADDRESS);
     UPDATE_SETTING_BY_CAP(enable_mr_prov_key, MPIR_CVAR_CH4_OFI_ENABLE_MR_PROV_KEY);
     UPDATE_SETTING_BY_CAP(enable_mr_allocated, MPIR_CVAR_CH4_OFI_ENABLE_MR_ALLOCATED);
+    UPDATE_SETTING_BY_CAP(enable_mr_register_null, MPIR_CVAR_CH4_OFI_ENABLE_MR_REGISTER_NULL);
 
     UPDATE_SETTING_BY_CAP(enable_tagged, MPIR_CVAR_CH4_OFI_ENABLE_TAGGED);
     UPDATE_SETTING_BY_CAP(enable_am, MPIR_CVAR_CH4_OFI_ENABLE_AM);
@@ -247,6 +291,9 @@ void MPIDI_OFI_init_settings(MPIDI_OFI_capabilities_t * p_settings, const char *
     UPDATE_SETTING_BY_CAP(major_version, MPIR_CVAR_CH4_OFI_MAJOR_VERSION);
     UPDATE_SETTING_BY_CAP(minor_version, MPIR_CVAR_CH4_OFI_MINOR_VERSION);
     UPDATE_SETTING_BY_CAP(num_am_buffers, MPIR_CVAR_CH4_OFI_NUM_AM_BUFFERS);
+    if (!strcmp(prov_name, "sockets")) {
+        UPDATE_SETTING_BY_CAP(enable_triggered, MPIR_CVAR_CH4_OFI_ENABLE_TRIGGERED);
+    }
     if (p_settings->num_am_buffers < 0) {
         p_settings->num_am_buffers = 0;
     }
@@ -254,8 +301,13 @@ void MPIDI_OFI_init_settings(MPIDI_OFI_capabilities_t * p_settings, const char *
         p_settings->num_am_buffers = MPIDI_OFI_MAX_NUM_AM_BUFFERS;
     }
 
+
     /* Always required settings */
     MPIDI_OFI_global.settings.require_rdm = 1;
+    UPDATE_SETTING_BY_CAP(num_optimized_memory_regions,
+                          MPIR_CVAR_CH4_OFI_NUM_OPTIMIZED_MEMORY_REGIONS);
+    UPDATE_SETTING_BY_CAP(enable_hmem, MPIR_CVAR_CH4_OFI_ENABLE_HMEM);
+    UPDATE_SETTING_BY_CAP(enable_mr_hmem, MPIR_CVAR_CH4_OFI_ENABLE_MR_HMEM);
 }
 
 #define CHECK_CAP(setting, cond_bad) \
@@ -276,23 +328,13 @@ int MPIDI_OFI_match_provider(struct fi_info *prov,
 {
     int score = 0;
 
-    MPL_DBG_MSG_FMT(MPIDI_CH4_DBG_GENERAL, VERBOSE, (MPL_DBG_FDEST, "Provider name: %s",
-                                                     prov->fabric_attr->prov_name));
-
-    CHECK_CAP(enable_scalable_endpoints,
-              prov->domain_attr->max_ep_tx_ctx <= 1 ||
-              (prov->caps & FI_NAMED_RX_CTX) != FI_NAMED_RX_CTX);
+    CHECK_CAP(enable_data, prov->domain_attr->cq_data_size < MPIDI_OFI_MIN_CQ_DATA_SIZE);
 
     /* From the fi_getinfo manpage: "FI_TAGGED implies the ability to send and receive
      * tagged messages." Therefore no need to specify FI_SEND|FI_RECV.  Moreover FI_SEND
      * and FI_RECV are mutually exclusive, so they should never be set both at the same
      * time. */
-    /* This capability set also requires the ability to receive data in the completion
-     * queue object (at least 32 bits). Previously, this was a separate capability set,
-     * but as more and more providers supported this feature, the decision was made to
-     * require it. */
-    CHECK_CAP(enable_tagged, !(prov->caps & FI_TAGGED) ||
-              prov->domain_attr->cq_data_size < MPIDI_OFI_MIN_CQ_DATA_SIZE);
+    CHECK_CAP(enable_tagged, !(prov->caps & FI_TAGGED));
 
     CHECK_CAP(enable_am, (prov->caps & (FI_MSG | FI_MULTI_RECV)) != (FI_MSG | FI_MULTI_RECV));
 
@@ -307,8 +349,12 @@ int MPIDI_OFI_match_provider(struct fi_info *prov,
     CHECK_CAP(enable_control_auto_progress,
               !(prov->domain_attr->control_progress & FI_PROGRESS_AUTO));
 
-    CHECK_CAP(enable_data_auto_progress, !(prov->domain_attr->data_progress & FI_PROGRESS_AUTO));
+    CHECK_CAP(enable_triggered,
+              !(prov->caps & (FI_ATOMICS | FI_RMA | FI_TRIGGER | FI_DIRECTED_RECV | FI_RMA_EVENT |
+                              FI_READ | FI_WRITE | FI_RECV | FI_SEND | FI_REMOTE_READ |
+                              FI_REMOTE_WRITE)));
 
+    CHECK_CAP(enable_data_auto_progress, !(prov->domain_attr->data_progress & FI_PROGRESS_AUTO));
     CHECK_CAP(require_rdm, prov->ep_attr->type != FI_EP_RDM);
 
     return score;
@@ -336,10 +382,10 @@ void MPIDI_OFI_update_global_settings(struct fi_info *prov)
                                   prov->domain_attr->mr_mode & (FI_MR_PROV_KEY | FI_MR_BASIC));
     UPDATE_SETTING_BY_INFO_DIRECT(enable_mr_allocated,
                                   prov->domain_attr->mr_mode & (FI_MR_ALLOCATED | FI_MR_BASIC));
-    UPDATE_SETTING_BY_INFO(enable_tagged,
-                           (prov->caps & FI_TAGGED) &&
-                           (prov->caps & FI_DIRECTED_RECV) &&
-                           (prov->domain_attr->cq_data_size >= MPIDI_OFI_MIN_CQ_DATA_SIZE));
+    UPDATE_SETTING_BY_INFO_DIRECT(enable_data,
+                                  (prov->caps & FI_DIRECTED_RECV) &&
+                                  prov->domain_attr->cq_data_size >= MPIDI_OFI_MIN_CQ_DATA_SIZE);
+    UPDATE_SETTING_BY_INFO(enable_tagged, prov->caps & FI_TAGGED);
     UPDATE_SETTING_BY_INFO(enable_am,
                            (prov->caps & (FI_MSG | FI_MULTI_RECV)) == (FI_MSG | FI_MULTI_RECV));
     UPDATE_SETTING_BY_INFO(enable_rma, prov->caps & FI_RMA);
@@ -359,4 +405,24 @@ void MPIDI_OFI_update_global_settings(struct fi_info *prov)
         MPIDI_OFI_global.settings.max_endpoints = MPIDI_OFI_MAX_ENDPOINTS_REGULAR;
         MPIDI_OFI_global.settings.max_endpoints_bits = MPIDI_OFI_MAX_ENDPOINTS_BITS_REGULAR;
     }
+
+    /* if CQ data is disabled, make sure we have sufficient source_bits */
+    if (!MPIDI_OFI_global.settings.enable_data) {
+        if (MPIDI_OFI_global.settings.source_bits == 0) {
+            MPIDI_OFI_global.settings.context_bits = MPIDI_OFI_CONTEXT_BITS_b;
+            MPIDI_OFI_global.settings.source_bits = MPIDI_OFI_SOURCE_BITS_b;
+            MPIDI_OFI_global.settings.tag_bits = MPIDI_OFI_TAG_BITS_b;
+        }
+    }
+
+    if (!strcmp(prov->fabric_attr->prov_name, "sockets")) {
+        MPIDI_OFI_global.settings.enable_triggered = MPIDI_OFI_global.settings.enable_triggered &&
+            (prov->caps & (FI_ATOMICS | FI_RMA | FI_TRIGGER | FI_DIRECTED_RECV | FI_RMA_EVENT |
+                           FI_READ | FI_WRITE | FI_RECV | FI_SEND | FI_REMOTE_READ |
+                           FI_REMOTE_WRITE));
+    }
+
+    if (MPIDI_OFI_global.settings.enable_triggered)
+        MPIDI_OFI_global.settings.enable_data_auto_progress = 1;
+
 }
