@@ -17,8 +17,8 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_POSIX_am_release_req_hdr(MPIDI_POSIX_am_reque
     MPIR_FUNC_ENTER;
 
 #ifndef POSIX_AM_REQUEST_INLINE
-    int vsi = (*req_hdr_ptr)->src_vsi;
-    MPIDU_genq_private_pool_free_cell(MPIDI_POSIX_global.per_vsi[vsi].am_hdr_buf_pool,
+    int vci = (*req_hdr_ptr)->src_vci;
+    MPIDU_genq_private_pool_free_cell(MPIDI_POSIX_global.per_vci[vci].am_hdr_buf_pool,
                                       (*req_hdr_ptr));
 #endif
 
@@ -29,7 +29,7 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_POSIX_am_release_req_hdr(MPIDI_POSIX_am_reque
 MPL_STATIC_INLINE_PREFIX int MPIDI_POSIX_am_init_req_hdr(const void *am_hdr,
                                                          size_t am_hdr_sz,
                                                          MPIDI_POSIX_am_request_header_t **
-                                                         req_hdr_ptr, MPIR_Request * sreq, int vsi)
+                                                         req_hdr_ptr, MPIR_Request * sreq, int vci)
 {
     int mpi_errno = MPI_SUCCESS;
     MPIDI_POSIX_am_request_header_t *req_hdr = *req_hdr_ptr;
@@ -42,7 +42,7 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_POSIX_am_init_req_hdr(const void *am_hdr,
 #endif /* POSIX_AM_REQUEST_INLINE */
 
     if (req_hdr == NULL) {
-        MPIDU_genq_private_pool_alloc_cell(MPIDI_POSIX_global.per_vsi[vsi].am_hdr_buf_pool,
+        MPIDU_genq_private_pool_alloc_cell(MPIDI_POSIX_global.per_vci[vci].am_hdr_buf_pool,
                                            (void **) &req_hdr);
         MPIR_ERR_CHKANDJUMP(!req_hdr, mpi_errno, MPI_ERR_OTHER, "**nomem");
 
@@ -53,7 +53,7 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_POSIX_am_init_req_hdr(const void *am_hdr,
     }
 
     if (am_hdr) {
-        MPIR_Typerep_copy(req_hdr->am_hdr, am_hdr, am_hdr_sz);
+        MPIR_Typerep_copy(req_hdr->am_hdr, am_hdr, am_hdr_sz, MPIR_TYPEREP_FLAG_NONE);
     }
 
     *req_hdr_ptr = req_hdr;
