@@ -81,6 +81,19 @@ cvars:
         knomial_2 - knomial_2 tree type (ranks are added in order from the right side)
         	    knomial_2 is only supported with non topology aware trees.
 
+    - name        : MPIR_CVAR_REDUCE_INTRANODE_MSG_SIZE_THRESHOLD
+      category    : COLLECTIVE
+      type        : int
+      default     : 2048
+      class       : none
+      verbosity   : MPI_T_VERBOSITY_USER_BASIC
+      scope       : MPI_T_SCOPE_ALL_EQ
+      description : >-
+        MPIR_CVAR_REDUCE_INTRANODE_TREE_KVAL and MPIR_CVAR_REDUCE_INTRANODE_TREE_TYPE are used
+        when the message size is smaller than or equal to this threshold;
+        MPIR_CVAR_REDUCE_INTRANODE_TREE_KVAL_LARGE and MPIR_CVAR_REDUCE_INTRANODE_TREE_TYPE_LARGE are
+        used when the message size is larger than this threshold.
+
     - name        : MPIR_CVAR_REDUCE_INTRANODE_TREE_KVAL
       category    : COLLECTIVE
       type        : int
@@ -91,6 +104,16 @@ cvars:
       description : >-
         K value for the kary/knomial tree for intra-node reduce
 
+    - name        : MPIR_CVAR_REDUCE_INTRANODE_TREE_KVAL_LARGE
+      category    : COLLECTIVE
+      type        : int
+      default     : 2
+      class       : none
+      verbosity   : MPI_T_VERBOSITY_USER_BASIC
+      scope       : MPI_T_SCOPE_ALL_EQ
+      description : >-
+        K value for the kary/knomial tree for intra-node reduce. Used for large messages.
+
     - name        : MPIR_CVAR_REDUCE_INTRANODE_TREE_TYPE
       category    : COLLECTIVE
       type        : string
@@ -105,6 +128,20 @@ cvars:
         knomial_2 - knomial_2 tree type (ranks are added in order from the right side)
         	    knomial_2 is only supported with non topology aware trees.
 
+    - name        : MPIR_CVAR_REDUCE_INTRANODE_TREE_TYPE_LARGE
+      category    : COLLECTIVE
+      type        : string
+      default     : kary
+      class       : none
+      verbosity   : MPI_T_VERBOSITY_USER_BASIC
+      scope       : MPI_T_SCOPE_ALL_EQ
+      description : >-
+        Tree type for intra-node reduce tree. Used for large messages.
+        kary      - kary tree type
+        knomial_1 - knomial_1 tree type (ranks are added in order from the left side)
+        knomial_2 - knomial_2 tree type (ranks are added in order from the right side)
+        	    knomial_2 is only supported with non topology aware trees.
+
     - name        : MPIR_CVAR_ENABLE_INTRANODE_TOPOLOGY_AWARE_TREES
       category    : COLLECTIVE
       type        : int
@@ -139,20 +176,6 @@ cvars:
 #include "topotree_util.h"
 
 
-MPIDI_POSIX_release_gather_tree_type_t MPIDI_POSIX_Bcast_tree_type, MPIDI_POSIX_Reduce_tree_type;
-
-static int get_tree_type(const char *tree_type_name)
-{
-    if (0 == strcmp(tree_type_name, "kary"))
-        return MPIDI_POSIX_RELEASE_GATHER_TREE_TYPE_KARY;
-    else if (0 == strcmp(tree_type_name, "knomial_1"))
-        return MPIDI_POSIX_RELEASE_GATHER_TREE_TYPE_KNOMIAL_1;
-    else if (0 == strcmp(tree_type_name, "knomial_2"))
-        return MPIDI_POSIX_RELEASE_GATHER_TREE_TYPE_KNOMIAL_2;
-    else
-        return MPIDI_POSIX_RELEASE_GATHER_TREE_TYPE_KARY;
-}
-
 /* Initialize the release_gather struct to NULL */
 int MPIDI_POSIX_mpi_release_gather_comm_init_null(MPIR_Comm * comm_ptr)
 {
@@ -177,7 +200,6 @@ int MPIDI_POSIX_mpi_release_gather_comm_init(MPIR_Comm * comm_ptr,
     MPIR_FUNC_ENTER;
 
     int mpi_errno = MPI_SUCCESS;
-    int mpi_errno_ret = MPI_SUCCESS;
     int rank, num_ranks;
     bool initialize_flags = false, initialize_bcast_buf = false, initialize_reduce_buf = false;
     int flags_num_pages, fallback = 0;
@@ -197,17 +219,21 @@ int MPIDI_POSIX_mpi_release_gather_comm_init(MPIR_Comm * comm_ptr,
      * reduce buffer (divided into multiple cells) per rank. */
 
     if (RELEASE_GATHER_FIELD(comm_ptr, is_initialized) == 0) {
-        RELEASE_GATHER_FIELD(comm_ptr, is_initialized) = 1;
         /* CVARs may get updated. Turn them into per-comm settings */
         RELEASE_GATHER_FIELD(comm_ptr, bcast_tree_type) =
-            get_tree_type(MPIR_CVAR_BCAST_INTRANODE_TREE_TYPE);
+            MPIDI_POSIX_mpi_release_gather_get_tree_type(MPIR_CVAR_BCAST_INTRANODE_TREE_TYPE);
         RELEASE_GATHER_FIELD(comm_ptr, bcast_tree_kval) = MPIR_CVAR_BCAST_INTRANODE_TREE_KVAL;
         RELEASE_GATHER_FIELD(comm_ptr, bcast_shm_size) =
             MPIR_CVAR_BCAST_INTRANODE_BUFFER_TOTAL_SIZE;
         RELEASE_GATHER_FIELD(comm_ptr, bcast_num_cells) = MPIR_CVAR_BCAST_INTRANODE_NUM_CELLS;
         RELEASE_GATHER_FIELD(comm_ptr, reduce_tree_type) =
-            get_tree_type(MPIR_CVAR_REDUCE_INTRANODE_TREE_TYPE);
+            MPIDI_POSIX_mpi_release_gather_get_tree_type(MPIR_CVAR_REDUCE_INTRANODE_TREE_TYPE);
         RELEASE_GATHER_FIELD(comm_ptr, reduce_tree_kval) = MPIR_CVAR_REDUCE_INTRANODE_TREE_KVAL;
+        RELEASE_GATHER_FIELD(comm_ptr, reduce_tree_type_large) =
+            MPIDI_POSIX_mpi_release_gather_get_tree_type
+            (MPIR_CVAR_REDUCE_INTRANODE_TREE_TYPE_LARGE);
+        RELEASE_GATHER_FIELD(comm_ptr, reduce_tree_kval_large) =
+            MPIR_CVAR_REDUCE_INTRANODE_TREE_KVAL_LARGE;
         RELEASE_GATHER_FIELD(comm_ptr, reduce_shm_size) =
             MPIR_CVAR_REDUCE_INTRANODE_BUFFER_TOTAL_SIZE;
         RELEASE_GATHER_FIELD(comm_ptr, reduce_num_cells) = MPIR_CVAR_REDUCE_INTRANODE_NUM_CELLS;
@@ -278,19 +304,19 @@ int MPIDI_POSIX_mpi_release_gather_comm_init(MPIR_Comm * comm_ptr,
 
                 fallback = 1;
                 MPIR_Bcast_impl(&fallback, 1, MPI_INT, 0, comm_ptr, errflag);
-                MPIR_ERR_SETANDJUMP(mpi_errno_ret, MPI_ERR_NO_MEM, "**nomem");
+                MPIR_ERR_SETANDJUMP(mpi_errno, MPI_ERR_NO_MEM, "**nomem");
             } else {
                 /* More shm can be created, update the shared counter */
                 MPL_atomic_fetch_add_uint64(MPIDI_POSIX_shm_limit_counter, memory_to_be_allocated);
                 fallback = 0;
                 mpi_errno = MPIR_Bcast_impl(&fallback, 1, MPI_INT, 0, comm_ptr, errflag);
-                MPIR_ERR_COLL_CHECKANDCONT(mpi_errno, errflag, mpi_errno_ret);
+                MPIR_ERR_CHECK(mpi_errno);
             }
         } else {
             mpi_errno = MPIR_Bcast_impl(&fallback, 1, MPI_INT, 0, comm_ptr, errflag);
-            MPIR_ERR_COLL_CHECKANDCONT(mpi_errno, errflag, mpi_errno_ret);
+            MPIR_ERR_CHECK(mpi_errno);
             if (fallback) {
-                MPIR_ERR_SETANDJUMP(mpi_errno_ret, MPI_ERR_NO_MEM, "**nomem");
+                MPIR_ERR_SETANDJUMP(mpi_errno, MPI_ERR_NO_MEM, "**nomem");
             }
         }
     }
@@ -318,7 +344,15 @@ int MPIDI_POSIX_mpi_release_gather_comm_init(MPIR_Comm * comm_ptr,
                                                  RELEASE_GATHER_FIELD(comm_ptr, reduce_tree_type),
                                                  &release_gather_info_ptr->reduce_tree,
                                                  &topotree_fail[1]);
-                MPIR_ERR_COLL_CHECKANDCONT(mpi_errno, errflag, mpi_errno_ret);
+                MPIR_ERR_CHECK(mpi_errno);
+                mpi_errno =
+                    MPIR_Treealgo_tree_create(rank, num_ranks,
+                                              RELEASE_GATHER_FIELD(comm_ptr,
+                                                                   reduce_tree_type_large),
+                                              RELEASE_GATHER_FIELD(comm_ptr,
+                                                                   reduce_tree_kval_large), 0,
+                                              &release_gather_info_ptr->reduce_tree_large);
+                MPIR_ERR_CHECK(mpi_errno);
             } else {
                 /* Finalize was already called and MPIR_Process.hwloc_topology has been destroyed */
                 topotree_fail[0] = -1;
@@ -326,7 +360,7 @@ int MPIDI_POSIX_mpi_release_gather_comm_init(MPIR_Comm * comm_ptr,
             }
             mpi_errno = MPIR_Allreduce_impl(MPI_IN_PLACE, topotree_fail, 2, MPI_INT,
                                             MPI_MAX, comm_ptr, errflag);
-            MPIR_ERR_COLL_CHECKANDCONT(mpi_errno, errflag, mpi_errno_ret);
+            MPIR_ERR_CHECK(mpi_errno);
         } else {
             topotree_fail[0] = -1;
             topotree_fail[1] = -1;
@@ -341,18 +375,26 @@ int MPIDI_POSIX_mpi_release_gather_comm_init(MPIR_Comm * comm_ptr,
                                           RELEASE_GATHER_FIELD(comm_ptr, bcast_tree_type),
                                           RELEASE_GATHER_FIELD(comm_ptr, bcast_tree_kval), 0,
                                           &release_gather_info_ptr->bcast_tree);
-            MPIR_ERR_COLL_CHECKANDCONT(mpi_errno, errflag, mpi_errno_ret);
+            MPIR_ERR_CHECK(mpi_errno);
         }
 
         if (topotree_fail[1] != 0) {
-            if (topotree_fail[1] == 1)
+            if (topotree_fail[1] == 1) {
                 MPIR_Treealgo_tree_free(&release_gather_info_ptr->reduce_tree);
+                MPIR_Treealgo_tree_free(&release_gather_info_ptr->reduce_tree_large);
+            }
             mpi_errno =
                 MPIR_Treealgo_tree_create(rank, num_ranks,
                                           RELEASE_GATHER_FIELD(comm_ptr, reduce_tree_type),
                                           RELEASE_GATHER_FIELD(comm_ptr, reduce_tree_kval), 0,
                                           &release_gather_info_ptr->reduce_tree);
-            MPIR_ERR_COLL_CHECKANDCONT(mpi_errno, errflag, mpi_errno_ret);
+            MPIR_ERR_CHECK(mpi_errno);
+            mpi_errno =
+                MPIR_Treealgo_tree_create(rank, num_ranks,
+                                          RELEASE_GATHER_FIELD(comm_ptr, reduce_tree_type_large),
+                                          RELEASE_GATHER_FIELD(comm_ptr, reduce_tree_kval_large), 0,
+                                          &release_gather_info_ptr->reduce_tree_large);
+            MPIR_ERR_CHECK(mpi_errno);
 
         }
 
@@ -363,17 +405,12 @@ int MPIDI_POSIX_mpi_release_gather_comm_init(MPIR_Comm * comm_ptr,
         release_gather_info_ptr->bcast_buf_addr = NULL;
         release_gather_info_ptr->reduce_buf_addr = NULL;
         release_gather_info_ptr->child_reduce_buf_addr = NULL;
+        release_gather_info_ptr->child_reduce_buf_addr_large = NULL;
 
         mpi_errno =
             MPIDU_shm_alloc(comm_ptr, flags_shm_size,
                             (void **) &(release_gather_info_ptr->flags_addr));
-        if (mpi_errno) {
-            /* for communication errors, just record the error but continue */
-            MPIR_ERR_ADD(mpi_errno_ret, MPIR_ERR_OTHER);
-        }
-
-        MPIR_ERR_COLL_CHECKANDCONT(mpi_errno_ret, errflag, mpi_errno_ret);
-        MPIR_ERR_CHECK(mpi_errno_ret);
+        MPIR_ERR_CHECK(mpi_errno);
 
         /* Calculate gather and release flag address and initialize to the gather and release states */
         release_gather_info_ptr->gather_flag_addr =
@@ -387,7 +424,7 @@ int MPIDI_POSIX_mpi_release_gather_comm_init(MPIR_Comm * comm_ptr,
 
         /* Make sure all the flags are set before ranks start reading each other's flags from shm */
         mpi_errno = MPIR_Barrier_impl(comm_ptr, errflag);
-        MPIR_ERR_COLL_CHECKANDCONT(mpi_errno, errflag, mpi_errno_ret);
+        MPIR_ERR_CHECK(mpi_errno);
     }
 
     if (initialize_bcast_buf) {
@@ -399,10 +436,7 @@ int MPIDI_POSIX_mpi_release_gather_comm_init(MPIR_Comm * comm_ptr,
         mpi_errno =
             MPIDU_shm_alloc(comm_ptr, RELEASE_GATHER_FIELD(comm_ptr, bcast_shm_size),
                             (void **) &(RELEASE_GATHER_FIELD(comm_ptr, bcast_buf_addr)));
-        if (mpi_errno) {
-            MPIR_ERR_ADD(mpi_errno_ret, MPIR_ERR_OTHER);
-        }
-        MPIR_ERR_COLL_CHECKANDCONT(mpi_errno, errflag, mpi_errno_ret);
+        MPIR_ERR_CHECK(mpi_errno);
     }
 
     if (initialize_reduce_buf) {
@@ -410,6 +444,8 @@ int MPIDI_POSIX_mpi_release_gather_comm_init(MPIR_Comm * comm_ptr,
         int i;
         RELEASE_GATHER_FIELD(comm_ptr, child_reduce_buf_addr) =
             MPL_malloc(num_ranks * sizeof(void *), MPL_MEM_COLL);
+        RELEASE_GATHER_FIELD(comm_ptr, child_reduce_buf_addr_large) =
+            MPL_malloc(num_ranks * sizeof(void *), MPL_MEM_COLL);
 
         if (rank == 0)
             MPL_atomic_fetch_add_uint64(MPIDI_POSIX_shm_limit_counter,
@@ -418,12 +454,7 @@ int MPIDI_POSIX_mpi_release_gather_comm_init(MPIR_Comm * comm_ptr,
         mpi_errno =
             MPIDU_shm_alloc(comm_ptr, num_ranks * RELEASE_GATHER_FIELD(comm_ptr, reduce_shm_size),
                             (void **) &(RELEASE_GATHER_FIELD(comm_ptr, reduce_buf_addr)));
-        if (mpi_errno) {
-            /* for communication errors, just record the error but continue */
-            MPIR_ERR_ADD(mpi_errno_ret, MPIR_ERR_OTHER);
-        }
-        MPIR_ERR_COLL_CHECKANDCONT(mpi_errno_ret, errflag, mpi_errno_ret);
-        MPIR_ERR_CHECK(mpi_errno_ret);
+        MPIR_ERR_CHECK(mpi_errno);
 
         /* Store address of each of the children's reduce buffer */
         char *addr;
@@ -434,13 +465,22 @@ int MPIDI_POSIX_mpi_release_gather_comm_init(MPIR_Comm * comm_ptr,
             RELEASE_GATHER_FIELD(comm_ptr, child_reduce_buf_addr[i]) =
                 addr + (child_rank * RELEASE_GATHER_FIELD(comm_ptr, reduce_shm_size));
         }
+        for (i = 0; i < RELEASE_GATHER_FIELD(comm_ptr, reduce_tree_large.num_children); i++) {
+            int child_rank =
+                *(int *) utarray_eltptr(RELEASE_GATHER_FIELD(comm_ptr, reduce_tree_large).children,
+                                        i);
+            RELEASE_GATHER_FIELD(comm_ptr, child_reduce_buf_addr_large[i]) =
+                addr + (child_rank * RELEASE_GATHER_FIELD(comm_ptr, reduce_shm_size));
+        }
     }
 
   fn_exit:
     MPIR_FUNC_EXIT;
-    if (mpi_errno_ret != MPI_SUCCESS)
+    if (mpi_errno != MPI_SUCCESS) {
+        MPIDI_POSIX_mpi_release_gather_comm_free(comm_ptr);
         RELEASE_GATHER_FIELD(comm_ptr, is_initialized) = 0;
-    return mpi_errno_ret;
+    }
+    return mpi_errno;
   fn_fail:
     goto fn_exit;
 }
@@ -451,7 +491,6 @@ int MPIDI_POSIX_mpi_release_gather_comm_free(MPIR_Comm * comm_ptr)
     MPIR_FUNC_ENTER;
 
     int mpi_errno = MPI_SUCCESS;
-    int mpi_errno_ret = MPI_SUCCESS;
     MPIR_Errflag_t errflag ATTRIBUTE((unused)) = MPIR_ERR_NONE;
 
     /* Clean up is not required for NULL struct */
@@ -466,7 +505,7 @@ int MPIDI_POSIX_mpi_release_gather_comm_free(MPIR_Comm * comm_ptr)
 
     /* destroy and detach shared memory used for flags */
     mpi_errno = MPIDU_shm_free(RELEASE_GATHER_FIELD(comm_ptr, flags_addr));
-    MPIR_ERR_COLL_CHECKANDCONT(mpi_errno, errflag, mpi_errno_ret);
+    MPIR_ERR_CHECK(mpi_errno);
 
     if (RELEASE_GATHER_FIELD(comm_ptr, bcast_buf_addr) != NULL) {
         if (comm_ptr->rank == 0)
@@ -474,7 +513,7 @@ int MPIDI_POSIX_mpi_release_gather_comm_free(MPIR_Comm * comm_ptr)
                                         RELEASE_GATHER_FIELD(comm_ptr, bcast_shm_size));
         /* destroy and detach shared memory used for bcast buffer */
         mpi_errno = MPIDU_shm_free(RELEASE_GATHER_FIELD(comm_ptr, bcast_buf_addr));
-        MPIR_ERR_COLL_CHECKANDCONT(mpi_errno, errflag, mpi_errno_ret);
+        MPIR_ERR_CHECK(mpi_errno);
     }
 
     if (RELEASE_GATHER_FIELD(comm_ptr, reduce_buf_addr) != NULL) {
@@ -483,15 +522,19 @@ int MPIDI_POSIX_mpi_release_gather_comm_free(MPIR_Comm * comm_ptr)
                                         RELEASE_GATHER_FIELD(comm_ptr, reduce_shm_size));
         /* destroy and detach shared memory used for reduce buffers */
         mpi_errno = MPIDU_shm_free(RELEASE_GATHER_FIELD(comm_ptr, reduce_buf_addr));
-        MPIR_ERR_COLL_CHECKANDCONT(mpi_errno, errflag, mpi_errno_ret);
+        MPIR_ERR_CHECK(mpi_errno);
 
         MPL_free(RELEASE_GATHER_FIELD(comm_ptr, child_reduce_buf_addr));
+        MPL_free(RELEASE_GATHER_FIELD(comm_ptr, child_reduce_buf_addr_large));
     }
 
     MPIR_Treealgo_tree_free(&(RELEASE_GATHER_FIELD(comm_ptr, bcast_tree)));
     MPIR_Treealgo_tree_free(&(RELEASE_GATHER_FIELD(comm_ptr, reduce_tree)));
+    MPIR_Treealgo_tree_free(&(RELEASE_GATHER_FIELD(comm_ptr, reduce_tree_large)));
 
   fn_exit:
     MPIR_FUNC_EXIT;
-    return mpi_errno_ret;
+    return mpi_errno;
+  fn_fail:
+    goto fn_exit;
 }
