@@ -60,7 +60,7 @@ int main(int argc, char *argv[])
         for (i = 0; i < size; i++) {
             for (j = 0; j < COUNT; j++) {
                 if (recvbuf[i * COUNT + j] != i * VERIFY_CONST + j) {
-                    printf("PE 0: mis-match error");
+                    printf("PE 0: mismatch error");
                     printf("  recbuf[%d * %d + %d] = ", i, COUNT, j);
                     printf("  %ld,", recvbuf[i * COUNT + j]);
                     printf("  should be %ld\n", i * VERIFY_CONST + j);
