@@ -67,9 +67,9 @@ int builtin_float_test(void)
 {
     int nints, nadds, ntypes, combiner;
 
-    int err, errs = 0;
+    int errs = 0;
 
-    err = MPI_Type_get_envelope(MPI_FLOAT, &nints, &nadds, &ntypes, &combiner);
+    MPI_Type_get_envelope(MPI_FLOAT, &nints, &nadds, &ntypes, &combiner);
 
     if (combiner != MPI_COMBINER_NAMED)
         errs++;
@@ -200,7 +200,7 @@ int optimizable_vector_of_basics_test(void)
     int i;
     MPI_Aint sizeofint, sizeoftype, position;
 
-    int err, errs = 0;
+    int errs = 0;
 
     MPI_Pack_external_size((char *) "external32", 1, MPI_INT, &sizeofint);
 
@@ -211,7 +211,7 @@ int optimizable_vector_of_basics_test(void)
     }
 
     /* set up type */
-    err = MPI_Type_vector(10, 2, 2, MPI_INT, &parent_type);
+    MPI_Type_vector(10, 2, 2, MPI_INT, &parent_type);
 
     MPI_Type_commit(&parent_type);
 
@@ -228,8 +228,7 @@ int optimizable_vector_of_basics_test(void)
     buf = (char *) malloc(sizeoftype);
 
     position = 0;
-    err = MPI_Pack_external((char *) "external32",
-                            array, 1, parent_type, buf, sizeoftype, &position);
+    MPI_Pack_external((char *) "external32", array, 1, parent_type, buf, sizeoftype, &position);
 
     if (position != sizeoftype) {
         errs++;
@@ -240,8 +239,7 @@ int optimizable_vector_of_basics_test(void)
 
     memset(array, 0, 20 * sizeof(int));
     position = 0;
-    err = MPI_Unpack_external((char *) "external32",
-                              buf, sizeoftype, &position, array, 1, parent_type);
+    MPI_Unpack_external((char *) "external32", buf, sizeoftype, &position, array, 1, parent_type);
 
     if (position != sizeoftype) {
         errs++;
@@ -283,7 +281,7 @@ int struct_of_basics_test(void)
     MPI_Aint indices[10];
     MPI_Datatype types[10];
 
-    int err, errs = 0;
+    int errs = 0;
 
     MPI_Pack_external_size((char *) "external32", 1, MPI_INT, &sizeofint);
 
@@ -302,7 +300,7 @@ int struct_of_basics_test(void)
     }
 
     /* set up type */
-    err = MPI_Type_create_struct(10, blocks, indices, types, &parent_type);
+    MPI_Type_create_struct(10, blocks, indices, types, &parent_type);
 
     MPI_Type_commit(&parent_type);
 
@@ -318,8 +316,7 @@ int struct_of_basics_test(void)
     buf = (char *) malloc(sizeoftype);
 
     position = 0;
-    err = MPI_Pack_external((char *) "external32",
-                            array, 1, parent_type, buf, sizeoftype, &position);
+    MPI_Pack_external((char *) "external32", array, 1, parent_type, buf, sizeoftype, &position);
 
     if (position != sizeoftype) {
         errs++;
@@ -330,8 +327,7 @@ int struct_of_basics_test(void)
 
     memset(array, 0, 20 * sizeof(int));
     position = 0;
-    err = MPI_Unpack_external((char *) "external32",
-                              buf, sizeoftype, &position, array, 1, parent_type);
+    MPI_Unpack_external((char *) "external32", buf, sizeoftype, &position, array, 1, parent_type);
 
     if (position != sizeoftype) {
         errs++;
