@@ -83,10 +83,14 @@ static mpi_names_t mpi_names[] = {
      * compiler does not support it), the value may be MPI_DATATYPE_NULL */
     {MPI_REAL4, "MPI_REAL4"},
     {MPI_REAL8, "MPI_REAL8"},
+#ifdef HAVE_MPI_REAL16
     {MPI_REAL16, "MPI_REAL16"},
+#endif
     {MPI_COMPLEX8, "MPI_COMPLEX8"},
     {MPI_COMPLEX16, "MPI_COMPLEX16"},
+#ifdef HAVE_MPI_COMPLEX32
     {MPI_COMPLEX32, "MPI_COMPLEX32"},
+#endif
     {MPI_INTEGER1, "MPI_INTEGER1"},
     {MPI_INTEGER2, "MPI_INTEGER2"},
     {MPI_INTEGER4, "MPI_INTEGER4"},
@@ -120,6 +124,9 @@ static mpi_names_t mpi_names[] = {
 #if MTEST_HAVE_MIN_MPI_VERSION(3,0)
     /* added in MPI 3 */
     {MPI_COUNT, "MPI_COUNT"},
+#endif
+#if MTEST_HAVE_MIN_MPI_VERSION(4,1)
+    {MPI_DATATYPE_NULL, "MPI_DATATYPE_NULL"},
 #endif
     {0, (char *) 0},    /* Sentinel used to indicate the last element */
 };
@@ -152,8 +159,14 @@ int main(int argc, char **argv)
     isSynonymName = 0;
     for (i = 0; mpi_names[i].name != 0; i++) {
         /* Are we in the optional types? */
-        if (strcmp(mpi_names[i].name, "MPI_REAL4") == 0)
+        if (strcmp(mpi_names[i].name, "MPI_REAL4") == 0) {
             inOptional = 1;
+#ifdef ENABLE_STRICTMPI
+            /* strictly we can't use MPI_DATATYPE_NULL to test datatype availability,
+             * just skip the optional datatypes */
+            break;
+#endif
+        }
         /* If this optional type is not supported, skip it */
         if (inOptional && mpi_names[i].dtype == MPI_DATATYPE_NULL)
             continue;
